Package: cuda.ml
Type: Package
Title: R Interface for the RAPIDS cuML Suite of Libraries
Version: 0.3.3
Authors@R: 
    c(person(given = "Yitao",
           family = "Li",
           role = c("aut", "cph"),
           email = "yitaoli1990@gmail.com",
           comment = c(ORCID = "0000-0002-1261-905X")),
      person(given = "Tomasz",
             family = "Kalinowski",
             role = c("aut", "cre", "cph"),
             email = "tomasz@posit.co"),
      person(given = "Daniel",
             family = "Falbel",
             role = c("aut", "cph"),
             email = "daniel@posit.co"),
      person(given = "RStudio",
           role = c("cph", "fnd")))
Description: R interface for RAPIDS cuML (<https://github.com/rapidsai/cuml>),
    a suite of GPU-accelerated machine learning libraries powered by CUDA
    (<https://en.wikipedia.org/wiki/CUDA>).
License: MIT + file LICENSE
URL: https://mlverse.github.io/cuda.ml/
BugReports: https://github.com/mlverse/cuda.ml/issues
Depends: R (>= 3.2)
Imports: ellipsis, hardhat, parsnip, Rcpp (>= 1.0.6), rlang (>= 0.1.4)
Suggests: callr, glmnet, MASS, magrittr, mlbench, purrr, reticulate,
        testthat, xgboost
LinkingTo: Rcpp
Encoding: UTF-8
RoxygenNote: 7.3.3
OS_type: unix
SystemRequirements: RAPIDS cuML (see https://rapids.ai/start.html)
NeedsCompilation: yes
Packaged: 2026-04-29 15:32:47 UTC; dfalbel
Author: Yitao Li [aut, cph] (ORCID: <https://orcid.org/0000-0002-1261-905X>),
  Tomasz Kalinowski [aut, cre, cph],
  Daniel Falbel [aut, cph],
  RStudio [cph, fnd]
Maintainer: Tomasz Kalinowski <tomasz@posit.co>
Repository: CRAN
Date/Publication: 2026-04-29 16:10:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-04-29 20:35:36 UTC; unix
Archs: cuda.ml.so.dSYM

Package: cuda.ml
Type: Package
Title: R Interface for the RAPIDS cuML Suite of Libraries
Version: 0.3.3
Authors@R: 
    c(person(given = "Yitao",
           family = "Li",
           role = c("aut", "cph"),
           email = "yitaoli1990@gmail.com",
           comment = c(ORCID = "0000-0002-1261-905X")),
      person(given = "Tomasz",
             family = "Kalinowski",
             role = c("aut", "cre", "cph"),
             email = "tomasz@posit.co"),
      person(given = "Daniel",
             family = "Falbel",
             role = c("aut", "cph"),
             email = "daniel@posit.co"),
      person(given = "RStudio",
           role = c("cph", "fnd")))
Description: R interface for RAPIDS cuML (<https://github.com/rapidsai/cuml>),
    a suite of GPU-accelerated machine learning libraries powered by CUDA
    (<https://en.wikipedia.org/wiki/CUDA>).
License: MIT + file LICENSE
URL: https://mlverse.github.io/cuda.ml/
BugReports: https://github.com/mlverse/cuda.ml/issues
Depends: R (>= 3.2)
Imports: ellipsis, hardhat, parsnip, Rcpp (>= 1.0.6), rlang (>= 0.1.4)
Suggests: callr, glmnet, MASS, magrittr, mlbench, purrr, reticulate,
        testthat, xgboost
LinkingTo: Rcpp
Encoding: UTF-8
RoxygenNote: 7.3.3
OS_type: unix
SystemRequirements: RAPIDS cuML (see https://rapids.ai/start.html)
NeedsCompilation: yes
Packaged: 2026-04-29 15:32:47 UTC; dfalbel
Author: Yitao Li [aut, cph] (ORCID: <https://orcid.org/0000-0002-1261-905X>),
  Tomasz Kalinowski [aut, cre, cph],
  Daniel Falbel [aut, cph],
  RStudio [cph, fnd]
Maintainer: Tomasz Kalinowski <tomasz@posit.co>
Repository: CRAN
Date/Publication: 2026-04-29 16:10:02 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-05-01 18:38:01 UTC; unix
Archs: cuda.ml.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(cuda_ml_can_predict_class_probabilities,cuda_ml_fil)
S3method(cuda_ml_can_predict_class_probabilities,cuda_ml_knn)
S3method(cuda_ml_can_predict_class_probabilities,cuda_ml_model)
S3method(cuda_ml_can_predict_class_probabilities,cuda_ml_rand_forest)
S3method(cuda_ml_can_predict_class_probabilities,default)
S3method(cuda_ml_elastic_net,data.frame)
S3method(cuda_ml_elastic_net,default)
S3method(cuda_ml_elastic_net,formula)
S3method(cuda_ml_elastic_net,matrix)
S3method(cuda_ml_elastic_net,recipe)
S3method(cuda_ml_get_state,cuda_ml_model)
S3method(cuda_ml_get_state,cuda_ml_pca)
S3method(cuda_ml_get_state,cuda_ml_rand_forest)
S3method(cuda_ml_get_state,cuda_ml_rand_proj_model)
S3method(cuda_ml_get_state,cuda_ml_svc)
S3method(cuda_ml_get_state,cuda_ml_svc_ovr)
S3method(cuda_ml_get_state,cuda_ml_svr)
S3method(cuda_ml_get_state,cuda_ml_umap)
S3method(cuda_ml_get_state,default)
S3method(cuda_ml_inverse_transform,cuda_ml_pca)
S3method(cuda_ml_inverse_transform,cuda_ml_tsvd)
S3method(cuda_ml_is_classifier,cuda_ml_model)
S3method(cuda_ml_is_classifier,default)
S3method(cuda_ml_knn,data.frame)
S3method(cuda_ml_knn,default)
S3method(cuda_ml_knn,formula)
S3method(cuda_ml_knn,matrix)
S3method(cuda_ml_knn,recipe)
S3method(cuda_ml_lasso,data.frame)
S3method(cuda_ml_lasso,default)
S3method(cuda_ml_lasso,formula)
S3method(cuda_ml_lasso,matrix)
S3method(cuda_ml_lasso,recipe)
S3method(cuda_ml_logistic_reg,data.frame)
S3method(cuda_ml_logistic_reg,default)
S3method(cuda_ml_logistic_reg,formula)
S3method(cuda_ml_logistic_reg,matrix)
S3method(cuda_ml_logistic_reg,recipe)
S3method(cuda_ml_ols,data.frame)
S3method(cuda_ml_ols,default)
S3method(cuda_ml_ols,formula)
S3method(cuda_ml_ols,matrix)
S3method(cuda_ml_ols,recipe)
S3method(cuda_ml_rand_forest,data.frame)
S3method(cuda_ml_rand_forest,default)
S3method(cuda_ml_rand_forest,formula)
S3method(cuda_ml_rand_forest,matrix)
S3method(cuda_ml_rand_forest,recipe)
S3method(cuda_ml_ridge,data.frame)
S3method(cuda_ml_ridge,default)
S3method(cuda_ml_ridge,formula)
S3method(cuda_ml_ridge,matrix)
S3method(cuda_ml_ridge,recipe)
S3method(cuda_ml_serialize,cuda_ml_model)
S3method(cuda_ml_serialize,default)
S3method(cuda_ml_set_state,cuda_ml_model_state)
S3method(cuda_ml_set_state,cuda_ml_pca_model_state)
S3method(cuda_ml_set_state,cuda_ml_rand_forest_model_state)
S3method(cuda_ml_set_state,cuda_ml_rand_proj_model_state)
S3method(cuda_ml_set_state,cuda_ml_svc_model_state)
S3method(cuda_ml_set_state,cuda_ml_svc_ovr_model_state)
S3method(cuda_ml_set_state,cuda_ml_svr_model_state)
S3method(cuda_ml_set_state,cuda_ml_umap_model_state)
S3method(cuda_ml_set_state,default)
S3method(cuda_ml_sgd,data.frame)
S3method(cuda_ml_sgd,default)
S3method(cuda_ml_sgd,formula)
S3method(cuda_ml_sgd,matrix)
S3method(cuda_ml_sgd,recipe)
S3method(cuda_ml_svm,data.frame)
S3method(cuda_ml_svm,default)
S3method(cuda_ml_svm,formula)
S3method(cuda_ml_svm,matrix)
S3method(cuda_ml_svm,recipe)
S3method(cuda_ml_transform,cuda_ml_rand_proj_model)
S3method(cuda_ml_transform,cuda_ml_tsvd)
S3method(cuda_ml_transform,cuda_ml_umap)
S3method(predict,cuda_ml_fil)
S3method(predict,cuda_ml_knn)
S3method(predict,cuda_ml_linear_model)
S3method(predict,cuda_ml_logistic_reg)
S3method(predict,cuda_ml_rand_forest)
S3method(predict,cuda_ml_svm)
export(cuML_major_version)
export(cuML_minor_version)
export(cuda_ml_agglomerative_clustering)
export(cuda_ml_can_predict_class_probabilities)
export(cuda_ml_dbscan)
export(cuda_ml_elastic_net)
export(cuda_ml_fil_enabled)
export(cuda_ml_fil_load_model)
export(cuda_ml_inverse_transform)
export(cuda_ml_is_classifier)
export(cuda_ml_kmeans)
export(cuda_ml_knn)
export(cuda_ml_knn_algo_ivfflat)
export(cuda_ml_knn_algo_ivfpq)
export(cuda_ml_knn_algo_ivfsq)
export(cuda_ml_lasso)
export(cuda_ml_logistic_reg)
export(cuda_ml_ols)
export(cuda_ml_pca)
export(cuda_ml_rand_forest)
export(cuda_ml_rand_proj)
export(cuda_ml_ridge)
export(cuda_ml_serialise)
export(cuda_ml_serialize)
export(cuda_ml_sgd)
export(cuda_ml_svm)
export(cuda_ml_transform)
export(cuda_ml_tsne)
export(cuda_ml_tsvd)
export(cuda_ml_umap)
export(cuda_ml_unserialise)
export(cuda_ml_unserialize)
export(has_cuML)
import(Rcpp)
importFrom(ellipsis,check_dots_used)
importFrom(rlang,"%||%")
importFrom(stats,predict)
useDynLib(cuda.ml, .registration = TRUE)

new_pca_model <- function(model) {
  class(model) <- c("cuda_ml_pca", "cuda_ml_model", class(model))

  model
}

#' Perform principal component analysis.
#'
#' Compute principal component(s) of the input data. Each feature from the input
#' will be mean-centered (but not scaled) before the SVD computation takes
#' place.
#'
#' @template model-with-numeric-input
#' @template eigen-decomposition
#' @template transform-input
#' @template cuML-log-level
#' @param n_components Number of principal component(s) to keep. Default:
#'   min(nrow(x), ncol(x)).
#' @param whiten If TRUE, then de-correlate all components, making each
#'   component have unit variance  and removing multi-collinearity.
#'   Default: FALSE.
#'
#' @return A PCA model object with the following attributes:
#'    - "components": a matrix of \code{n_components} rows containing the top
#'      principal components.
#'    - "explained_variance": amount of variance within the input data explained
#'      by each component.
#'    - "explained_variance_ratio": fraction of variance within the input data
#'      explained by each component.
#'    - "singular_values": singular values (non-negative) corresponding to the
#'      top principal components.
#'    - "mean": the column wise mean of \code{x} which was used to mean-center
#'      \code{x} first.
#'    - "transformed_data": (only present if "transform_input" is set to TRUE)
#'      an approximate representation of input data based on principal
#'      components.
#'    - "pca_params": opaque pointer to PCA parameters which will be used for
#'      performing inverse transforms.
#'
#'  The model object can be used as input to the inverse_transform() function to
#'  map a representation based on principal components back to the original
#'  feature space.
#'
#' @examples
#'
#' library(cuda.ml)
#'
#' iris.pca <- cuda_ml_pca(iris[1:4], n_components = 3)
#' print(iris.pca)
#' @export
cuda_ml_pca <- function(x,
                        n_components = NULL,
                        eig_algo = c("dq", "jacobi"),
                        tol = 1e-7, n_iters = 15L,
                        whiten = FALSE,
                        transform_input = TRUE,
                        cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace")) {
  n_components <- n_components %||% min(nrow(x), ncol(x))
  eig_algo <- match_eig_algo(eig_algo)
  cuML_log_level <- match_cuML_log_level(cuML_log_level)

  model_obj <- .pca_fit_transform(
    x = as.matrix(x),
    n_components = as.integer(n_components),
    algo = eig_algo,
    tol = as.numeric(tol),
    n_iters = as.integer(n_iters),
    whiten = whiten,
    transform_input = transform_input,
    verbosity = cuML_log_level
  )

  new_pca_model(model_obj)
}

#' @export
cuda_ml_inverse_transform.cuda_ml_pca <- function(model, x, ...) {
  .pca_inverse_transform(model = model, x = as.matrix(x))
}

cuda_ml_get_state.cuda_ml_pca <- function(model) {
  model_state <- .pca_get_state(model)

  new_model_state(model_state, "cuda_ml_pca_model_state")
}

cuda_ml_set_state.cuda_ml_pca_model_state <- function(model_state) {
  model_state <- .pca_set_state(model_state)

  new_pca_model(model_state)
}

context("Principal Component Analysis")

cuda_ml_pca_model <- cuda_ml_pca(
  iris[, which(names(iris) != "Species")],
  n_components = 3
)

sklearn_pca_model <- sklearn$decomposition$PCA(n_components = 3L, copy = TRUE)
sklearn_pca_model$fit(sklearn_iris_dataset$data)

test_that("cuda_ml_pca() works as expected", {
  expect_equal(
    cuda_ml_pca_model$components, as.matrix(sklearn_pca_model$components_),
    tolerance = 1e-8, scale = 1
  )
  expect_equal(
    cuda_ml_pca_model$explained_variance,
    as.numeric(sklearn_pca_model$explained_variance_),
    tolerance = 1e-8, scale = 1
  )
  expect_equal(
    cuda_ml_pca_model$explained_variance_ratio,
    as.numeric(sklearn_pca_model$explained_variance_ratio_),
    tolerance = 1e-8, scale = 1
  )
  expect_equal(
    cuda_ml_pca_model$singular_values,
    as.numeric(sklearn_pca_model$singular_values_),
    tolerance = 1e-8, scale = 1
  )
  expect_equal(
    cuda_ml_pca_model$mean, as.numeric(sklearn_pca_model$mean_),
    tolerance = 1e-8, scale = 1
  )
})

test_that("cuda_ml_inverse_transform() works as expected for PCA models", {
  expect_equal(
    cuda_ml_pca_model$transformed_data %>%
      sklearn_pca_model$inverse_transform() %>%
      as.matrix(),
    cuda_ml_pca_model %>%
      cuda_ml_inverse_transform(cuda_ml_pca_model$transformed_data),
    tolerance = 1e-8, scale = 1
  )
})

#' utility functions that may potentially become platform-specific in future

nproc <- function() {
  # Try to run `nproc` to detect number of cores available
  tryCatch(
    as.integer(system2("nproc", stdout = TRUE, stderr = NULL)),
    error = function(e) 2L
  )
}

elastic_net_validate_alpha <- function(alpha) {
  if (alpha <= 0) {
    stop("`alpha` (multiplier of the elastic penalty term) must be positive!")
  }
}

#' Train a linear model using elastic regression.
#'
#' Train a linear model with combined L1 and L2 priors as the regularizer.
#'
#' @template supervised-model-inputs
#' @template supervised-model-output
#' @template ellipsis-unused
#' @template fit-intercept
#' @template normalize-input
#' @template coordinate-descend
#' @template l1_ratio
#' @param alpha Multiplier of the penalty term (i.e., the result would become
#'   and Ordinary Least Square model if \code{alpha} were set to 0). Default: 1.
#'   For numerical reasons, running elastic regression with \code{alpha} set to
#'   0 is not advised. For the \code{alpha}-equals-to-0 scenario, one should use
#'   \code{cuda_ml_ols} to train an OLS model instead.
#'   Default: 1.
#'
#' @return An elastic net regressor that can be used with the 'predict' S3
#'   generic to make predictions on new data points.
#'
#' @examples
#'
#' library(cuda.ml)
#'
#' model <- cuda_ml_elastic_net(
#'   formula = mpg ~ ., data = mtcars, alpha = 1e-3, l1_ratio = 0.6
#' )
#' cuda_ml_predictions <- predict(model, mtcars)
#'
#' # predictions will be comparable to those from a `glmnet` model with `lambda`
#' # set to 1e-3 and `alpha` set to 0.6
#' # (in `glmnet`, `lambda` is the weight of the penalty term, and `alpha` is
#' #  the elastic mixing parameter between L1 and L2 penalties.
#'
#' library(glmnet)
#'
#' glmnet_model <- glmnet(
#'   x = as.matrix(mtcars[names(mtcars) != "mpg"]), y = mtcars$mpg,
#'   alpha = 0.6, lambda = 1e-3, nlambda = 1, standardize = FALSE
#' )
#'
#' glm_predictions <- predict(
#'   glmnet_model, as.matrix(mtcars[names(mtcars) != "mpg"]),
#'   s = 0
#' )
#'
#' print(
#'   all.equal(
#'     as.numeric(glm_predictions),
#'     cuda_ml_predictions$.pred,
#'     tolerance = 1e-2
#'   )
#' )
#' @importFrom ellipsis check_dots_used
#' @export
cuda_ml_elastic_net <- function(x, ...) {
  check_dots_used()
  UseMethod("cuda_ml_elastic_net")
}

#' @rdname cuda_ml_elastic_net
#' @export
cuda_ml_elastic_net.default <- function(x, ...) {
  report_undefined_fn("cuda_ml_elastic_net", x)
}

#' @rdname cuda_ml_elastic_net
#' @export
cuda_ml_elastic_net.data.frame <- function(x, y,
                                           alpha = 1, l1_ratio = 0.5,
                                           max_iter = 1000L, tol = 1e-3,
                                           fit_intercept = TRUE,
                                           normalize_input = FALSE,
                                           selection = c("cyclic", "random"),
                                           ...) {
  processed <- hardhat::mold(x, y)

  cuda_ml_elastic_net_bridge(
    processed = processed,
    alpha = alpha,
    l1_ratio = l1_ratio,
    max_iter = max_iter,
    tol = tol,
    fit_intercept = fit_intercept,
    normalize_input = normalize_input,
    selection = selection
  )
}

#' @rdname cuda_ml_elastic_net
#' @export
cuda_ml_elastic_net.matrix <- function(x, y,
                                       alpha = 1, l1_ratio = 0.5,
                                       max_iter = 1000L, tol = 1e-3,
                                       fit_intercept = TRUE,
                                       normalize_input = FALSE,
                                       selection = c("cyclic", "random"),
                                       ...) {
  processed <- hardhat::mold(x, y)

  cuda_ml_elastic_net_bridge(
    processed = processed,
    alpha = alpha,
    l1_ratio = l1_ratio,
    max_iter = max_iter,
    tol = tol,
    fit_intercept = fit_intercept,
    normalize_input = normalize_input,
    selection = selection
  )
}

#' @rdname cuda_ml_elastic_net
#' @export
cuda_ml_elastic_net.formula <- function(formula, data,
                                        alpha = 1, l1_ratio = 0.5,
                                        max_iter = 1000L, tol = 1e-3,
                                        fit_intercept = TRUE,
                                        normalize_input = FALSE,
                                        selection = c("cyclic", "random"),
                                        ...) {
  processed <- hardhat::mold(formula, data)

  cuda_ml_elastic_net_bridge(
    processed = processed,
    alpha = alpha,
    l1_ratio = l1_ratio,
    max_iter = max_iter,
    tol = tol,
    fit_intercept = fit_intercept,
    normalize_input = normalize_input,
    selection = selection
  )
}

#' @rdname cuda_ml_elastic_net
#' @export
cuda_ml_elastic_net.recipe <- function(x, data,
                                       alpha = 1, l1_ratio = 0.5,
                                       max_iter = 1000L, tol = 1e-3,
                                       fit_intercept = TRUE,
                                       normalize_input = FALSE,
                                       selection = c("cyclic", "random"),
                                       ...) {
  processed <- hardhat::mold(x, data)

  cuda_ml_elastic_net_bridge(
    processed = processed,
    alpha = alpha,
    l1_ratio = l1_ratio,
    max_iter = max_iter,
    tol = tol,
    fit_intercept = fit_intercept,
    normalize_input = normalize_input,
    selection = selection
  )
}

cuda_ml_elastic_net_bridge <- function(processed,
                                       alpha, l1_ratio,
                                       max_iter, tol,
                                       fit_intercept,
                                       normalize_input,
                                       selection = c("cyclic", "random")) {
  validate_lm_input(processed)
  elastic_net_validate_alpha(alpha)
  selection <- match.arg(selection)
  if (!fit_intercept && normalize_input) {
    stop(
      "fit_intercept=FALSE, normalize_input=TRUE is unsupported for elastic ",
      "net"
    )
  }

  x <- as.matrix(processed$predictors)
  y <- processed$outcomes[[1]]

  model_xptr <- .cd_fit(
    x = x,
    y = y,
    fit_intercept = fit_intercept,
    normalize_input = normalize_input,
    epochs = as.integer(max_iter),
    loss = 0L, # squared loss
    alpha = as.numeric(alpha),
    l1_ratio = as.numeric(l1_ratio),
    shuffle = identical(selection, "random"),
    tol = as.numeric(tol)
  )

  new_linear_model(
    cls = "cuda_ml_elastic_net",
    xptr = model_xptr,
    blueprint = processed$blueprint
  )
}

context("Elastic Net")

test_that("Elastic net regressor works as expected", {
  for (normalize_input in c(FALSE, TRUE)) {
    for (fit_intercept in c(FALSE, TRUE)) {
      for (l1_ratio in c(0.4, 0.5, 0.6)) {
        if (!fit_intercept && normalize_input) {
          next
        }

        if (!fit_intercept) {
          input <- mtcars
          input[, which(names(mtcars) != "mpg")] <- scale(
            input[, which(names(mtcars) != "mpg")],
            scale = FALSE
          )
          input <- as.matrix(input)
        } else {
          input <- as.matrix(mtcars)
        }

        if (normalize_input) {
          sklearn_scaler <- sklearn$preprocessing$StandardScaler(
            copy = TRUE, with_mean = TRUE, with_std = TRUE
          )
          sklearn_scaler$fit(as.matrix(mtcars[, which(names(mtcars) != "mpg")]))
          sklearn_predictors <- sklearn_scaler$transform(
            as.matrix(mtcars[, which(names(mtcars) != "mpg")])
          )
        } else {
          sklearn_predictors <- as.matrix(input[, which(names(mtcars) != "mpg")])
        }

        sklearn_elastic_net_regressor <- sklearn$linear_model$ElasticNet(
          alpha = 1e-3,
          max_iter = 10000L,
          tol = 1e-4,
          fit_intercept = fit_intercept,
          l1_ratio = l1_ratio
        )
        sklearn_elastic_net_regressor$fit(
          X = sklearn_predictors,
          y = mtcars$mpg
        )
        sklearn_elastic_net_regressor_preds <- sklearn_elastic_net_regressor$predict(
          sklearn_predictors
        )

        cuda_ml_elastic_net_regressor <- cuda_ml_elastic_net(
          mpg ~ ., data = input,
          alpha = 1e-3,
          l1_ratio = l1_ratio,
          max_iter = 10000,
          tol = 1e-4,
          fit_intercept = fit_intercept,
          normalize_input = normalize_input
        )
        cuda_ml_elastic_net_regressor_preds <- predict(
          cuda_ml_elastic_net_regressor, input
        )

        expect_equal(
          cuda_ml_elastic_net_regressor_preds$.pred,
          as.numeric(sklearn_elastic_net_regressor_preds),
          tol = 0.3,
          scale = 1
        )
      }
    }
  }
})

decision_tree_match_split_criterion <- function(criterion = c("gini", "entropy", "mse", "mae"),
                                                classification = TRUE) {
  criterion <- criterion %||% ifelse(classification, "gini", "mse")
  criterion <- match.arg(criterion)

  if (classification && criterion %in% c("mse", "mae")) {
    stop("'", criterion, "' is not a valid criterion for classification.")
  }

  switch(criterion,
    gini = 0L,
    entropy = 1L,
    mse = 2L,
    mae = 3L
  )
}

sgd_match_loss <- function(loss = c("squared_loss", "log", "hinge")) {
  loss <- match.arg(loss)

  switch(loss,
    squared_loss = 0L,
    log = 1L,
    hinge = 2L
  )
}

sgd_match_penalty <- function(penalty = c("none", "l1", "l2", "elasticnet")) {
  penalty <- match.arg(penalty)

  switch(penalty,
    none = 0L,
    l1 = 1L,
    l2 = 2L,
    elasticnet = 3L
  )
}

sgd_match_learning_rate <- function(learning_rate = c("constant", "invscaling", "adaptive")) {
  learning_rate <- match.arg(learning_rate)

  switch(learning_rate,
    constant = 1L,
    invscaling = 2L,
    adaptive = 3L
  )
}

#' Train a MBSGD linear model.
#'
#' Train a linear model using mini-batch stochastic gradient descent.
#'
#' @template supervised-model-inputs
#' @template supervised-model-output
#' @template ellipsis-unused
#' @template fit-intercept
#' @template l1_ratio
#' @param loss Loss function, must be one of {"squared_loss", "log", "hinge"}.
#' @param penalty Type of regularization to perform, must be one of
#'   {"none", "l1", "l2", "elasticnet"}.
#'
#'   - "none": no regularization.
#'   - "l1": perform regularization based on the L1-norm (LASSO) which tries to
#'           minimize the sum of the absolute values of the coefficients.
#'   - "l2": perform regularization based on the L2 norm (Ridge) which tries to
#'           minimize the sum of the square of the coefficients.
#'   - "elasticnet": perform the Elastic Net regularization which is based on
#'                   the weighted averable of L1 and L2 norms.
#'   Default: "none".
#' @param alpha Multiplier of the penalty term. Default: 1e-4.
#' @param batch_size The number of samples that will be included in each batch.
#'   Default: 32L.
#' @param epochs The number of times the model should iterate through the entire
#'   dataset during training. Default: 1000L.
#' @param tol Threshold for stopping training. Training will stop if
#'   (loss in current epoch) > (loss in previous epoch) - \code{tol}.
#'   Default: 1e-3.
#' @param shuffle Whether to shuffles the training data after each epoch.
#'   Default: True.
#' @param eta0 The initial learning rate. Default: 1e-3.
#' @param power_t The exponent used for calculating the invscaling learning
#'   rate. Default: 0.5.
#' @param learning_rate Must be one of {"constant", "invscaling", "adaptive"}.
#'
#'   - "constant": the learning rate will be kept constant.
#'   - "invscaling": (learning rate) = (initial learning rate) / pow(t, power_t)
#'                   where \code{t} is the number of epochs and
#'                   \code{power_t} is a tunable parameter of this model.
#'   - "adaptive": (learning rate) = (initial learning rate) as long as the
#'                 training loss keeps decreasing. Each time the last
#'                 \code{n_iter_no_change} consecutive epochs fail to decrease
#'                 the training loss by \code{tol}, the current learning rate is
#'                 divided by 5.
#'   Default: "constant".
#' @param eta0 The initial learning rate. Default: 1e-3.
#' @param power_t The exponent used in the invscaling learning rate
#'   calculations.
#' @param n_iters_no_change The maximum number of epochs to train if there is no
#'   imporvement in the model. Default: 5.
#'
#' @return A linear model that can be used with the 'predict' S3 generic to make
#'   predictions on new data points.
#'
#' @examples
#'
#' library(cuda.ml)
#'
#' model <- cuda_ml_sgd(
#'   mpg ~ ., mtcars,
#'   batch_size = 4L, epochs = 50000L,
#'   learning_rate = "adaptive", eta0 = 1e-5,
#'   penalty = "l2", alpha = 1e-5, tol = 1e-6,
#'   n_iters_no_change = 10L
#' )
#'
#' preds <- predict(model, mtcars[names(mtcars) != "mpg"])
#' print(all.equal(preds$.pred, mtcars$mpg, tolerance = 0.09))
#' @importFrom ellipsis check_dots_used
#' @export
cuda_ml_sgd <- function(x, ...) {
  check_dots_used()
  UseMethod("cuda_ml_sgd")
}

#' @rdname cuda_ml_sgd
#' @export
cuda_ml_sgd.default <- function(x, ...) {
  report_undefined_fn("cuda_ml_sgd", x)
}

#' @rdname cuda_ml_sgd
#' @export
cuda_ml_sgd.data.frame <- function(x, y,
                                   fit_intercept = TRUE,
                                   loss = c("squared_loss", "log", "hinge"),
                                   penalty = c("none", "l1", "l2", "elasticnet"),
                                   alpha = 1e-4, l1_ratio = 0.5,
                                   epochs = 1000L, tol = 1e-3, shuffle = TRUE,
                                   learning_rate = c("constant", "invscaling", "adaptive"),
                                   eta0 = 1e-3, power_t = 0.5, batch_size = 32L,
                                   n_iters_no_change = 5L,
                                   ...) {
  processed <- hardhat::mold(x, y)

  cuda_ml_sgd_bridge(
    processed = processed,
    fit_intercept = fit_intercept,
    loss = loss,
    penalty = penalty,
    alpha = alpha,
    l1_ratio = l1_ratio,
    epochs = epochs,
    tol = tol,
    shuffle = shuffle,
    learning_rate = learning_rate,
    eta0 = eta0,
    power_t = power_t,
    batch_size = batch_size,
    n_iters_no_change = n_iters_no_change
  )
}

#' @rdname cuda_ml_sgd
#' @export
cuda_ml_sgd.matrix <- function(x, y,
                               fit_intercept = TRUE,
                               loss = c("squared_loss", "log", "hinge"),
                               penalty = c("none", "l1", "l2", "elasticnet"),
                               alpha = 1e-4, l1_ratio = 0.5,
                               epochs = 1000L, tol = 1e-3, shuffle = TRUE,
                               learning_rate = c("constant", "invscaling", "adaptive"),
                               eta0 = 1e-3, power_t = 0.5, batch_size = 32L,
                               n_iters_no_change = 5L,
                               ...) {
  processed <- hardhat::mold(x, y)

  cuda_ml_sgd_bridge(
    processed = processed,
    fit_intercept = fit_intercept,
    loss = loss,
    penalty = penalty,
    alpha = alpha,
    l1_ratio = l1_ratio,
    epochs = epochs,
    tol = tol,
    shuffle = shuffle,
    learning_rate = learning_rate,
    eta0 = eta0,
    power_t = power_t,
    batch_size = batch_size,
    n_iters_no_change = n_iters_no_change
  )
}

#' @rdname cuda_ml_sgd
#' @export
cuda_ml_sgd.formula <- function(formula, data,
                                fit_intercept = TRUE,
                                loss = c("squared_loss", "log", "hinge"),
                                penalty = c("none", "l1", "l2", "elasticnet"),
                                alpha = 1e-4, l1_ratio = 0.5,
                                epochs = 1000L, tol = 1e-3, shuffle = TRUE,
                                learning_rate = c("constant", "invscaling", "adaptive"),
                                eta0 = 1e-3, power_t = 0.5, batch_size = 32L,
                                n_iters_no_change = 5L,
                                ...) {
  processed <- hardhat::mold(formula, data)

  cuda_ml_sgd_bridge(
    processed = processed,
    fit_intercept = fit_intercept,
    loss = loss,
    penalty = penalty,
    alpha = alpha,
    l1_ratio = l1_ratio,
    epochs = epochs,
    tol = tol,
    shuffle = shuffle,
    learning_rate = learning_rate,
    eta0 = eta0,
    power_t = power_t,
    batch_size = batch_size,
    n_iters_no_change = n_iters_no_change
  )
}

#' @rdname cuda_ml_sgd
#' @export
cuda_ml_sgd.recipe <- function(x, data,
                               fit_intercept = TRUE,
                               loss = c("squared_loss", "log", "hinge"),
                               penalty = c("none", "l1", "l2", "elasticnet"),
                               alpha = 1e-4, l1_ratio = 0.5,
                               epochs = 1000L, tol = 1e-3, shuffle = TRUE,
                               learning_rate = c("constant", "invscaling", "adaptive"),
                               eta0 = 1e-3, power_t = 0.5, batch_size = 32L,
                               n_iters_no_change = 5L,
                               ...) {
  processed <- hardhat::mold(x, data)

  cuda_ml_sgd_bridge(
    processed = processed,
    fit_intercept = fit_intercept,
    loss = loss,
    penalty = penalty,
    alpha = alpha,
    l1_ratio = l1_ratio,
    epochs = epochs,
    tol = tol,
    shuffle = shuffle,
    learning_rate = learning_rate,
    eta0 = eta0,
    power_t = power_t,
    batch_size = batch_size,
    n_iters_no_change = n_iters_no_change
  )
}

cuda_ml_sgd_bridge <- function(processed,
                               fit_intercept,
                               loss,
                               penalty,
                               alpha, l1_ratio,
                               epochs, tol, shuffle,
                               learning_rate,
                               eta0, power_t, batch_size,
                               n_iters_no_change) {
  validate_lm_input(processed)
  loss <- sgd_match_loss(loss)
  penalty <- sgd_match_penalty(penalty)

  learning_rate <- sgd_match_learning_rate(learning_rate)
  if (eta0 <= 0) {
    stop("eta0 must be > 0")
  }

  x <- as.matrix(processed$predictors)
  y <- processed$outcomes[[1]]

  model_xptr <- .sgd_fit(
    x = x,
    y = y,
    fit_intercept = fit_intercept,
    batch_size = as.integer(batch_size),
    epochs = as.integer(epochs),
    lr_type = learning_rate,
    eta0 = as.numeric(eta0),
    power_t = as.numeric(power_t),
    loss = loss,
    penalty = penalty,
    alpha = as.numeric(alpha),
    l1_ratio = as.numeric(l1_ratio),
    shuffle = shuffle,
    tol = as.numeric(tol),
    n_iter_no_change = as.integer(n_iters_no_change)
  )

  new_linear_model(
    cls = "cuda_ml_sgd",
    xptr = model_xptr,
    blueprint = processed$blueprint
  )
}

context("Mini-batch Stochastic Gradient Descent")

set.seed(0L)

gen_blobs_df <- function(n_samples) {
  centers <- list(c(0, 2, 2), c(2, -2, 0))
  blob_sz <- n_samples / 2
  X <- gen_blobs(blob_sz = blob_sz, centers = centers)
  y <- c(-1, 1) %>%
    sapply(function(x) rep(x, blob_sz)) %>%
    c()
  df <- cbind(X, y) %>% as.data.frame()

  df
}

test_that("MBSGD works as expected for training linear regressors", {
  train_df <- gen_blobs_df(n_samples = 1000)
  test_df <- gen_blobs_df(n_samples = 100)

  for (learning_rate in c("constant", "adaptive", "invscaling")) {
    for (penalty in c("l1", "l2", "elasticnet")) {
      for (loss in c("squared_loss", "log", "hinge")) {
        model <- cuda_ml_sgd(
          y ~ ., train_df,
          fit_intercept = TRUE,
          loss = loss, penalty = penalty,
          batch_size = 16,
          learning_rate = learning_rate, eta0 = 1e-5,
          tol = 1e-5
        )
        preds <- predict(model, test_df[names(test_df) != "y"])

        expect_lte(sum(sign(preds$.pred) != sign(test_df$y)) / nrow(test_df), 0.01)
      }
    }
  }
})

validate_lm_input <- function(processed) {
  hardhat::validate_outcomes_are_univariate(processed$outcomes)
  hardhat::validate_outcomes_are_numeric(processed$outcomes)
  hardhat::validate_predictors_are_numeric(processed$predictors)

  predictors <- as.matrix(processed$predictors)
  if (ncol(predictors) < 1) stop("Predictors must contain at least 1 feature.")
  if (nrow(predictors) < 2) stop("At least 2 samples are required.")
}

new_linear_model <- function(cls, xptr, ...) {
  new_model(
    cls = c(cls, "cuda_ml_linear_model"),
    mode = "regression",
    xptr = xptr,
    ...
  )
}

#' Make predictions on new data points.
#'
#' Make predictions on new data points using a linear model.
#'
#' @template predict
#'
#' @importFrom ellipsis check_dots_used
#' @export
predict.cuda_ml_linear_model <- function(object, x, ...) {
  check_dots_used()

  processed <- hardhat::forge(x, object$blueprint)

  preds <- .lm_predict(
    input = as.matrix(processed$predictors),
    coef = object$xptr$coef,
    intercept = object$xptr$intercept
  )

  postprocess_regression_results(preds)
}

#' Determine whether Forest Inference Library (FIL) functionalities are enabled
#' in the current installation of {cuda.ml}.
#'
#' CuML Forest Inference Library (FIL) functionalities (see
#' https://github.com/rapidsai/cuml/tree/main/python/cuml/fil#readme) will
#' require Treelite C API. If you need FIL to run tree-based model ensemble on
#' GPU, and \code{fil_enabled()} returns FALSE, then please consider installing
#' Treelite and then re-installing {cuda.ml}.
#'
#' @return A logical value indicating whether the Forest Inference Library (FIL)
#'   functionalities are enabled.
#'
#' @examples
#' if (cuda_ml_fil_enabled()) {
#'   # run GPU-accelerated Forest Inference Library (FIL) functionalities
#' } else {
#'   message(
#'     "FIL functionalities are disabled in the current installation of ",
#'     "{cuda.ml}. Please reinstall Treelite C library first, and then re-install",
#'     " {cuda.ml} to enable FIL."
#'   )
#' }
#' @export
cuda_ml_fil_enabled <- .fil_enabled

fil_match_model_type <- function(filename, model_type = c("xgboost", "lightgbm")) {
  model_type <- match.arg(model_type)

  switch(model_type,
    xgboost = ifelse(grepl("\\.json$", filename), 1L, 0L),
    lightgbm = 2L
  )
}

fil_match_algo <- function(algo = c("auto", "naive", "tree_reorg", "batch_tree_reorg")) {
  algo <- match.arg(algo)

  switch(algo,
    auto = 0L,
    naive = 1L,
    tree_reorg = 2L,
    batch_tree_reorg = 3L
  )
}

file_match_storage_type <- function(storage_type = c("auto", "dense", "sparse")) {
  storage_type <- match.arg(storage_type)

  switch(storage_type,
    auto = 0L,
    dense = 1L,
    sparse = 2L
  )
}

#' Load a XGBoost or LightGBM model file.
#'
#' Load a XGBoost or LightGBM model file using Treelite. The resulting model
#' object can be used to perform high-throughput batch inference on new data
#' points using the GPU acceleration functionality from the CuML Forest
#' Inference Library (FIL).
#'
#' @param filename Path to the saved model file.
#' @param mode Type of task to be performed by the model. Must be one of
#'   {"classification", "regression"}.
#' @param model_type Format of the saved model file. Notice if \code{filename}
#'   ends with ".json" and \code{model_type} is "xgboost", then {cuda.ml} will
#'   assume the model file is in XGBoost JSON (instead of binary) format.
#'   Default: "xgboost".
#' @param algo Type of the algorithm for inference, must be one of the
#'   following.
#'     - "auto":
#'         Choose the algorithm automatically. Currently 'batch_tree_reorg' is
#'         used for dense storage, and 'naive' for sparse storage.
#'     - "naive":
#'         Simple inference using shared memory.
#'     - "tree_reorg":
#'         Similar to naive but with trees rearranged to be more coalescing-
#'         friendly.
#'     - "batch_tree_reorg":
#'         Similar to 'tree_reorg' but predicting multiple rows per thread
#'         block.
#'   Default: "auto".
#' @param threshold Class probability threshold for classification. Ignored for
#'   regression tasks. Default: 0.5.
#' @param storage_type In-memory storage format of the FIL model. Must be one of
#'   the following.
#'   - "auto":
#'       Choose the storage type automatically,
#'   - "dense":
#'       Create a dense forest,
#'   - "sparse":
#'       Create a sparse forest. Requires \code{algo} to be 'naive' or 'auto'.
#' @param threads_per_tree If >1, then have multiple (neighboring) threads infer
#'   on the same tree within a block, which will improve memory bandwith near
#'   tree root (but consuming more shared memory). Default: 1L.
#' @param n_items Number of input samples each thread processes. If 0, then
#'   choose (up to 4) that fit into shared memory. Default: 0L.
#' @param blocks_per_sm Indicates how CuML should determine the number of thread
#'   blocks to lauch for the inference kernel.
#'   - 0:
#'     Launches the number of blocks proportional to the number of data points.
#'   - >= 1:
#'     Attempts to lauch \code{blocks_per_sm} blocks for each streaming
#'     multiprocessor.
#'     This will fail if \code{blocks_per_sm} blocks result in more threads than
#'     the maximum supported number of threads per GPU. Even if successful, it
#'     is not guaranteed that \code{blocks_per_sm} blocks will run on an SM
#'     concurrently.
#'
#' @return A GPU-accelerated FIL model that can be used with the 'predict' S3
#'   generic to make predictions on new data points.
#'
#' @examples
#'
#' library(cuda.ml)
#' library(xgboost)
#'
#' model_path <- file.path(tempdir(), "xgboost.model")
#'
#' model <- xgboost(
#'   data = as.matrix(mtcars[names(mtcars) != "mpg"]),
#'   label = as.matrix(mtcars["mpg"]),
#'   max.depth = 6,
#'   eta = 1,
#'   nthread = 2,
#'   nrounds = 20,
#'   objective = "reg:squarederror"
#' )
#'
#' xgb.save(model, model_path)
#'
#' model <- cuda_ml_fil_load_model(
#'   model_path,
#'   mode = "regression",
#'   model_type = "xgboost"
#' )
#'
#' preds <- predict(model, mtcars[names(mtcars) != "mpg"])
#'
#' print(preds)
#' @export
cuda_ml_fil_load_model <- function(filename,
                                   mode = c("classification", "regression"),
                                   model_type = c("xgboost", "lightgbm"),
                                   algo = c("auto", "naive", "tree_reorg", "batch_tree_reorg"),
                                   threshold = 0.5,
                                   storage_type = c("auto", "dense", "sparse"),
                                   threads_per_tree = 1L, n_items = 0L,
                                   blocks_per_sm = 0L) {
  mode <- match.arg(mode)
  model_type <- fil_match_model_type(filename, model_type)
  algo <- fil_match_algo(algo)
  storage_type <- file_match_storage_type(storage_type)

  xptr <- .fil_load_model(
    model_type = model_type,
    filename = filename,
    algo = algo,
    classification = identical(mode, "classification"),
    threshold = as.numeric(threshold),
    storage_type = storage_type,
    threads_per_tree = as.integer(threads_per_tree),
    n_items = as.integer(n_items),
    blocks_per_sm = as.integer(blocks_per_sm)
  )
  model <- list(mode = mode, xptr = xptr)
  class(model) <- c("cuda_ml_fil", "cuda_ml_model", class(model))

  model
}

#' Make predictions on new data points.
#'
#' Make predictions on new data points using a FIL model.
#'
#' @template predict
#' @template output-class-probabilities
#'
#' @importFrom ellipsis check_dots_used
#' @export
predict.cuda_ml_fil <- function(object, x, output_class_probabilities = FALSE, ...) {
  check_dots_used()

  num_classes <- .fil_get_num_classes(model = object$xptr)
  preds <- .fil_predict(
    model = object$xptr,
    x = as.matrix(x),
    output_class_probabilities = output_class_probabilities
  )

  switch(object$mode,
    classification = {
      if (output_class_probabilities) {
        preds <- hardhat::spruce_prob(
          paste0("class_prob_", seq(num_classes) - 1L), preds
        )
      } else {
        preds <- factor(preds)
        preds <- hardhat::spruce_class(preds)
      }
    },
    regression = {
      preds <- hardhat::spruce_numeric(c(preds))
    }
  )

  preds
}

svm_match_kernel_type <- function(kernel = c("rbf", "tanh", "polynomial", "linear")) {
  kernel <- match.arg(kernel)

  switch(kernel,
    linear = 0L,
    polynomial = 1L,
    rbf = 2L,
    tanh = 3L
  )
}

#' Train a SVM model.
#'
#' Train a Support Vector Machine model for classification or regression tasks.
#'
#' @template supervised-model-inputs
#' @template supervised-model-output
#' @template cuML-log-level
#' @template ellipsis-unused
#' @param cost A positive number for the cost of predicting a sample within or
#'   on the wrong side of the margin. Default: 1.
#' @param kernel Type of the SVM kernel function (must be one of "rbf", "tanh",
#'   "polynomial", or "linear"). Default: "rbf".
#' @param gamma The gamma coefficient (only relevant to polynomial, RBF, and
#'   tanh kernel functions, see explanations below).
#'   Default: 1 / (num features).
#'
#'   The following kernels are implemented:
#'     - RBF K(x_1, x_2) = exp(-gamma |x_1-x_2|^2)
#'     - TANH K(x_1, x_2) = tanh(gamma <x_1,x_2> + coef0)
#'     - POLYNOMIAL K(x_1, x_2) = (gamma <x_1,x_2> + coef0)^degree
#'     - LINEAR K(x_1,x_2) = <x_1,x_2>,
#'   where < , > denotes the dot product.
#' @param coef0 The 0th coefficient (only applicable to polynomial and tanh
#'   kernel functions, see explanations below). Default: 0.
#'
#'   The following kernels are implemented:
#'     - RBF K(x_1, x_2) = exp(-gamma |x_1-x_2|^2)
#'     - TANH K(x_1, x_2) = tanh(gamma <x_1,x_2> + coef0)
#'     - POLYNOMIAL K(x_1, x_2) = (gamma <x_1,x_2> + coef0)^degree
#'     - LINEAR K(x_1,x_2) = <x_1,x_2>,
#'   where < , > denotes the dot product.
#' @param degree Degree of the polynomial kernel function (note: not applicable
#'   to other kernel types, see explanations below). Default: 3.
#'
#'   The following kernels are implemented:
#'     - RBF K(x_1, x_2) = exp(-gamma |x_1-x_2|^2)
#'     - TANH K(x_1, x_2) = tanh(gamma <x_1,x_2> + coef0)
#'     - POLYNOMIAL K(x_1, x_2) = (gamma <x_1,x_2> + coef0)^degree
#'     - LINEAR K(x_1,x_2) = <x_1,x_2>,
#'   where < , > denotes the dot product.
#' @param tol Tolerance to stop fitting. Default: 1e-3.
#' @param max_iter Maximum number of outer iterations in SmoSolver.
#'   Default: 100 * (num samples).
#' @param nochange_steps Number of steps with no change w.r.t convergence.
#'   Default: 1000.
#' @param cache_size Size of kernel cache (MiB) in device memory. Default: 1024.
#' @param epsilon Espsilon parameter of the epsilon-SVR model. There is no
#'   penalty for points that are predicted within the epsilon-tube around the
#'   target values. Please note this parameter is only relevant for regression
#'   tasks. Default: 0.1.
#' @param sample_weights Optional weight assigned to each input data point.
#'
#' @return A SVM classifier / regressor object that can be used with the
#'   'predict' S3 generic to make predictions on new data points.
#'
#' @examples
#'
#' library(cuda.ml)
#'
#' # Classification
#'
#' model <- cuda_ml_svm(
#'   formula = Species ~ .,
#'   data = iris,
#'   kernel = "rbf"
#' )
#'
#' predictions <- predict(model, iris[names(iris) != "Species"])
#'
#' # Regression
#'
#' model <- cuda_ml_svm(
#'   formula = mpg ~ .,
#'   data = mtcars,
#'   kernel = "rbf"
#' )
#'
#' predictions <- predict(model, mtcars)
#' @importFrom ellipsis check_dots_used
#' @export
cuda_ml_svm <- function(x, ...) {
  check_dots_used()
  UseMethod("cuda_ml_svm")
}

#' @rdname cuda_ml_svm
#' @export
cuda_ml_svm.default <- function(x, ...) {
  report_undefined_fn("cuda_ml_svm", x)
}

#' @rdname cuda_ml_svm
#' @export
cuda_ml_svm.data.frame <- function(x, y, cost = 1,
                                   kernel = c("rbf", "tanh", "polynomial", "linear"),
                                   gamma = NULL, coef0 = 0, degree = 3L,
                                   tol = 1e-3, max_iter = NULL,
                                   nochange_steps = 1000L, cache_size = 1024,
                                   epsilon = 0.1, sample_weights = NULL,
                                   cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace"),
                                   ...) {
  processed <- hardhat::mold(x, y)

  cuda_ml_svm_bridge(
    processed = processed,
    cost = cost,
    kernel = kernel,
    gamma = gamma,
    coef0 = coef0,
    degree = degree,
    tol = tol,
    max_iter = max_iter,
    nochange_steps = nochange_steps,
    cache_size = cache_size,
    epsilon = epsilon,
    sample_weights = sample_weights,
    cuML_log_level = cuML_log_level
  )
}

#' @rdname cuda_ml_svm
#' @export
cuda_ml_svm.matrix <- function(x, y, cost = 1,
                               kernel = c("rbf", "tanh", "polynomial", "linear"),
                               gamma = NULL, coef0 = 0, degree = 3L, tol = 1e-3,
                               max_iter = NULL, nochange_steps = 1000L,
                               cache_size = 1024, epsilon = 0.1,
                               sample_weights = NULL,
                               cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace"),
                               ...) {
  processed <- hardhat::mold(x, y)

  cuda_ml_svm_bridge(
    processed,
    cost = cost,
    kernel = kernel,
    gamma = gamma,
    coef0 = coef0,
    degree = degree,
    tol = tol,
    max_iter = max_iter,
    nochange_steps = nochange_steps,
    cache_size = cache_size,
    epsilon = epsilon,
    sample_weights = sample_weights,
    cuML_log_level = cuML_log_level
  )
}

#' @rdname cuda_ml_svm
#' @export
cuda_ml_svm.formula <- function(formula, data, cost = 1,
                                kernel = c("rbf", "tanh", "polynomial", "linear"),
                                gamma = NULL, coef0 = 0, degree = 3L, tol = 1e-3,
                                max_iter = NULL, nochange_steps = 1000L,
                                cache_size = 1024, epsilon = 0.1,
                                sample_weights = NULL,
                                cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace"),
                                ...) {
  processed <- hardhat::mold(formula, data)

  cuda_ml_svm_bridge(
    processed,
    cost = cost,
    kernel = kernel,
    gamma = gamma,
    coef0 = coef0,
    degree = degree,
    tol = tol,
    max_iter = max_iter,
    nochange_steps = nochange_steps,
    cache_size = cache_size,
    epsilon = epsilon,
    sample_weights = sample_weights,
    cuML_log_level = cuML_log_level
  )
}

#' @rdname cuda_ml_svm
#' @export
cuda_ml_svm.recipe <- function(x, data, cost = 1,
                               kernel = c("rbf", "tanh", "polynomial", "linear"),
                               gamma = NULL, coef0 = 0, degree = 3L, tol = 1e-3,
                               max_iter = NULL, nochange_steps = 1000L,
                               cache_size = 1024, epsilon = 0.1,
                               sample_weights = NULL,
                               cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace"),
                               ...) {
  processed <- hardhat::mold(x, data)

  cuda_ml_svm_bridge(
    processed,
    cost = cost,
    kernel = kernel,
    gamma = gamma,
    coef0 = coef0,
    degree = degree,
    tol = tol,
    max_iter = max_iter,
    nochange_steps = nochange_steps,
    cache_size = cache_size,
    epsilon = epsilon,
    sample_weights = sample_weights,
    cuML_log_level = cuML_log_level
  )
}

cuda_ml_svm_bridge <- function(processed, cost, kernel, gamma, coef0, degree, tol,
                               max_iter, nochange_steps, cache_size, epsilon,
                               sample_weights, cuML_log_level) {
  hardhat::validate_predictors_are_numeric(processed$predictors)
  hardhat::validate_outcomes_are_univariate(processed$outcomes)
  x <- as.matrix(processed$predictors)
  y <- processed$outcomes[[1]]

  gamma <- gamma %||% 1.0 / ncol(x)
  max_iter <- max_iter %||% 100L * nrow(x)
  kernel <- svm_match_kernel_type(kernel)
  cuML_log_level <- match_cuML_log_level(cuML_log_level)

  svm_fit_impl <- (
    if (is.factor(y)) {
      # classification
      ylevels <- levels(y)
      if (length(ylevels) > 2) {
        cuda_ml_svm_classification_multiclass_impl
      } else {
        cuda_ml_svm_classification_binary_impl
      }
    } else {
      cuda_ml_svm_regression_impl
    })

  svm_fit_impl(
    processed = processed,
    cost = cost,
    kernel = kernel,
    gamma = gamma,
    coef0 = coef0,
    degree = degree,
    tol = tol,
    max_iter = max_iter,
    nochange_steps = nochange_steps,
    cache_size = cache_size,
    epsilon = epsilon,
    sample_weights = sample_weights,
    cuML_log_level = cuML_log_level
  )
}

cuda_ml_svm_classification_multiclass_impl <- function(processed, cost, kernel,
                                                       gamma, coef0, degree, tol,
                                                       max_iter, nochange_steps,
                                                       cache_size, epsilon,
                                                       sample_weights, cuML_log_level) {
  x <- as.matrix(processed$predictors)
  y <- processed$outcomes[[1]]
  ylevels <- levels(y)

  # implement a one-vs-rest strategy for multi-class classification
  models <- list()
  for (idx in seq_along(ylevels)) {
    ovr_labels <- as.integer(y == ylevels[[idx]])

    model <- (
      if (!any(ovr_labels)) {
        # None of the training data points had the current label.
        NULL
      } else {
        model_xptr <- .svc_fit(
          input = x,
          labels = ovr_labels,
          cost = as.numeric(cost),
          kernel = kernel,
          gamma = as.numeric(gamma),
          coef0 = as.numeric(coef0),
          degree = as.integer(degree),
          tol = as.numeric(tol),
          max_iter = as.integer(max_iter),
          nochange_steps = as.integer(nochange_steps),
          cache_size = as.numeric(cache_size),
          sample_weights = as.numeric(sample_weights),
          verbosity = cuML_log_level
        )

        new_model(
          cls = c("cuda_ml_svc", "cuda_ml_svm"),
          mode = "classification",
          xptr <- model_xptr
        )
      })

    models <- append(models, list(model))
  }

  new_model(
    cls = c("cuda_ml_svc_ovr", "cuda_ml_svm"),
    mode = "classification",
    xptr = models,
    multiclass = TRUE,
    blueprint = processed$blueprint
  )
}

cuda_ml_get_state.cuda_ml_svc_ovr <- function(model) {
  model_state <- list(
    ovr_model_states = lapply(model$xptr, function(x) cuda_ml_get_state(x)),
    blueprint = model$blueprint
  )

  new_model_state(model_state, "cuda_ml_svc_ovr_model_state")
}

cuda_ml_set_state.cuda_ml_svc_ovr_model_state <- function(model_state) {
  new_model(
    cls = c("cuda_ml_svc_ovr", "cuda_ml_svm"),
    mode = "classification",
    xptr = lapply(model_state$ovr_model_states, function(x) cuda_ml_set_state(x)),
    multiclass = TRUE,
    blueprint = model_state$blueprint
  )
}

cuda_ml_svm_classification_binary_impl <- function(processed, cost, kernel, gamma,
                                                   coef0, degree, tol, max_iter,
                                                   nochange_steps, cache_size,
                                                   epsilon, sample_weights,
                                                   cuML_log_level) {
  x <- as.matrix(processed$predictors)
  y <- processed$outcomes[[1]]

  model_xptr <- .svc_fit(
    input = x,
    labels = as.integer(y),
    cost = as.numeric(cost),
    kernel = kernel,
    gamma = as.numeric(gamma),
    coef0 = as.numeric(coef0),
    degree = as.integer(degree),
    tol = as.numeric(tol),
    max_iter = as.integer(max_iter),
    nochange_steps = as.integer(nochange_steps),
    cache_size = as.numeric(cache_size),
    sample_weights = as.numeric(sample_weights),
    verbosity = cuML_log_level
  )

  new_model(
    cls = c("cuda_ml_svc", "cuda_ml_svm"),
    mode = "classification",
    xptr = model_xptr,
    multiclass = FALSE,
    blueprint = processed$blueprint
  )
}

cuda_ml_get_state.cuda_ml_svc <- function(model) {
  model_state <- list(
    model_state = .svc_get_state(model$xptr),
    blueprint = model$blueprint
  )

  new_model_state(model_state, "cuda_ml_svc_model_state")
}

cuda_ml_set_state.cuda_ml_svc_model_state <- function(model_state) {
  new_model(
    cls = c("cuda_ml_svc", "cuda_ml_svm"),
    mode = "classification",
    xptr = .svc_set_state(model_state$model_state),
    multiclass = FALSE,
    blueprint = model_state$blueprint
  )
}

cuda_ml_svm_regression_impl <- function(processed, cost, kernel, gamma, coef0,
                                        degree, tol, max_iter, nochange_steps,
                                        cache_size, epsilon, sample_weights,
                                        cuML_log_level) {
  x <- as.matrix(processed$predictors)
  y <- processed$outcomes[[1]]

  model_xptr <- .svr_fit(
    X = x,
    y = as.numeric(y),
    cost = as.numeric(cost),
    kernel = kernel,
    gamma = as.numeric(gamma),
    coef0 = as.numeric(coef0),
    degree = as.integer(degree),
    tol = as.numeric(tol),
    max_iter = as.integer(max_iter),
    nochange_steps = as.integer(nochange_steps),
    cache_size = as.numeric(cache_size),
    epsilon = as.numeric(epsilon),
    sample_weights = as.numeric(sample_weights),
    verbosity = cuML_log_level
  )

  new_model(
    cls = c("cuda_ml_svr", "cuda_ml_svm"),
    mode = "regression",
    xptr = model_xptr,
    blueprint = processed$blueprint
  )
}

cuda_ml_get_state.cuda_ml_svr <- function(model) {
  model_state <- list(
    model_state = .svr_get_state(model$xptr),
    blueprint = model$blueprint
  )

  new_model_state(model_state, "cuda_ml_svr_model_state")
}

cuda_ml_set_state.cuda_ml_svr_model_state <- function(model_state) {
  new_model(
    cls = c("cuda_ml_svr", "cuda_ml_svm"),
    mode = "regression",
    xptr = .svr_set_state(model_state$model_state),
    blueprint = model_state$blueprint
  )
}

#' Make predictions on new data points.
#'
#' Make predictions on new data points using a CuML SVM model.
#'
#' @template predict
#'
#' @importFrom ellipsis check_dots_used
#' @export
predict.cuda_ml_svm <- function(object, x, ...) {
  check_dots_used()

  processed <- hardhat::forge(x, object$blueprint)

  predict_cuda_ml_svm_bridge(model = object, processed = processed)
}

predict_cuda_ml_svm_bridge <- function(model, processed) {
  svm_predict_impl <- switch(model$mode,
    classification = (
      if (model$multiclass) {
        predict_cuda_ml_svm_classification_multiclass_impl
      } else {
        predict_cuda_ml_svm_classification_binary_impl
      }),
    regression = (
      predict_cuda_ml_svm_regression_impl
    )
  )

  out <- svm_predict_impl(model = model, processed = processed)
  hardhat::validate_prediction_size(out, processed$predictors)

  out
}

predict_cuda_ml_svm_classification_multiclass_impl <- function(model, processed) {
  pred_levels <- get_pred_levels(model)

  scores <- lapply(
    seq_along(pred_levels),
    function(label_idx) {
      if (is.null(model$xptr[[label_idx]])) {
        # None of the training data points had the current label.
        rep(-Inf, nrow(processed$predictors))
      } else {
        .svc_predict(
          model_xptr = model$xptr[[label_idx]]$xptr,
          input = as.matrix(processed$predictors),
          predict_class = FALSE
        )
      }
    }
  )

  preds <- sapply(
    seq_len(nrow(processed$predictors)),
    function(row_idx) {
      row_scores <- lapply(
        seq_along(pred_levels),
        function(label_idx) scores[[label_idx]][[row_idx]]
      )

      which.max(row_scores)
    }
  )

  postprocess_classification_results(preds, model)
}

predict_cuda_ml_svm_classification_binary_impl <- function(model, processed) {
  preds <- .svc_predict(
    model_xptr = model$xptr,
    input = as.matrix(processed$predictors),
    predict_class = TRUE
  )

  postprocess_classification_results(preds, model)
}

predict_cuda_ml_svm_regression_impl <- function(model, processed) {
  preds <- .svr_predict(
    svr_xptr = model$xptr,
    X = as.matrix(processed$predictors)
  )

  postprocess_regression_results(preds)
}

# register the CuML-based rand_forest model for parsnip
register_svm_model <- function(pkgname) {
  for (model in c(paste0("svm_", c("rbf", "poly", "linear")))) {
    for (mode in c("classification", "regression")) {
      parsnip::set_model_engine(model = model, mode = mode, eng = pkgname)
    }
    parsnip::set_dependency(model = model, eng = pkgname, pkg = pkgname)

    parsnip::set_model_arg(
      model = model,
      eng = pkgname,
      parsnip = "cost",
      original = "cost",
      func = list(pkg = "dials", fun = "cost", range = c(-10, 5)),
      has_submodel = FALSE
    )

    parsnip::set_model_arg(
      model = model,
      eng = pkgname,
      parsnip = "margin",
      original = "epsilon",
      func = list(pkg = "dials", fun = "svm_margin"),
      has_submodel = FALSE
    )
  }

  parsnip::set_model_arg(
    model = "svm_rbf",
    eng = pkgname,
    parsnip = "rbf_sigma",
    original = "gamma",
    func = list(pkg = "dials", fun = "rbf_sigma"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "svm_poly",
    eng = pkgname,
    parsnip = "degree",
    original = "degree",
    func = list(pkg = "dials", fun = "degree"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "svm_poly",
    eng = pkgname,
    parsnip = "scale_factor",
    original = "gamma",
    func = list(pkg = "dials", fun = "scale_factor"),
    has_submodel = FALSE
  )

  for (kernel in c("rbf", "poly", "linear")) {
    model <- paste0("svm_", kernel)

    for (mode in c("classification", "regression")) {
      parsnip::set_fit(
        model = model,
        eng = pkgname,
        mode = mode,
        value = list(
          interface = "formula",
          protect = c("formula", "data"),
          func = c(pkg = pkgname, fun = "cuda_ml_svm"),
          defaults = list(kernel = kernel)
        )
      )

      parsnip::set_encoding(
        model = model,
        eng = pkgname,
        mode = mode,
        options = list(
          predictor_indicators = "none",
          compute_intercept = FALSE,
          remove_intercept = FALSE,
          allow_sparse_x = FALSE
        )
      )
    }

    parsnip::set_pred(
      model = model,
      eng = pkgname,
      mode = "classification",
      type = "class",
      value = list(
        pre = NULL,
        post = NULL,
        func = c(fun = "predict"),
        args = list(
          quote(object$fit),
          quote(new_data)
        )
      )
    )

    parsnip::set_pred(
      model = model,
      eng = pkgname,
      mode = "regression",
      type = "numeric",
      value = list(
        pre = NULL,
        post = NULL,
        func = c(fun = "predict"),
        args = list(
          quote(object$fit),
          quote(new_data)
        )
      )
    )
  }
}

context("(de)serialization of SVM models")

test_that("SVM regressor can be serialized and unserialized correctly", {
  model <- cuda_ml_svm(formula = mpg ~ ., data = mtcars, kernel = "rbf")
  model_state <- cuda_ml_serialize(model)

  expected_preds <- predict(model, mtcars)
  actual_preds <- predict_in_sub_proc(
    model_state,
    data = mtcars,
    expected_mode = "regression",
    expected_model_cls = "cuda_ml_svr"
  )

  expect_equal(expected_preds, actual_preds)
})

context("Support Vector Machine")

test_that("cuda_ml_svm() works as expected for binary classification tasks", {
  gen_cuda_ml_binary_svc_input <- function() {
    data <- iris
    data[, "is_versicolor"] <- factor(data[, "Species"] == "versicolor")

    data[, names(data) != "Species"]
  }
  cuda_ml_binary_svc_input <- gen_cuda_ml_binary_svc_input()

  gen_sklearn_binary_svc_input <- function() {
    ds <- sklearn_iris_dataset
    ds$target <- (ds$target == which(levels(iris$Species) == "versicolor"))

    ds
  }
  sklearn_binary_svc_input <- gen_sklearn_binary_svc_input()

  cuda_ml_binary_svc_model <- cuda_ml_svm(
    formula = is_versicolor ~ .,
    data = cuda_ml_binary_svc_input,
    kernel = "rbf"
  )
  cuda_ml_binary_svc_preds <- predict(
    cuda_ml_binary_svc_model,
    cuda_ml_binary_svc_input[, names(cuda_ml_binary_svc_input) != "is_versicolor"]
  )

  sklearn_binary_svc_model <- sklearn$svm$SVC(kernel = "rbf", gamma = "auto")
  sklearn_binary_svc_model$fit(
    sklearn_binary_svc_input$data, sklearn_binary_svc_input$target
  )
  sklearn_binary_svc_preds <- sklearn_binary_svc_model$predict(
    sklearn_binary_svc_input$data
  )

  expect_equal(
    as.logical(cuda_ml_binary_svc_preds$.pred_class),
    as.logical(sklearn_binary_svc_preds)
  )
})

test_that("cuda_ml_svm() works as expected for multi-class classification tasks", {
  cuda_ml_multiclass_svc_input <- iris[, names(iris) != "Species"]

  cuda_ml_multiclass_svc_model <- cuda_ml_svm(
    formula = Species ~ ., data = iris, kernel = "rbf"
  )
  cuda_ml_multiclass_svc_preds <- predict(
    cuda_ml_multiclass_svc_model, cuda_ml_multiclass_svc_input
  )

  sklearn_multiclass_svc_model <- sklearn$svm$SVC(kernel = "rbf", gamma = "auto")
  sklearn_multiclass_svc_model$fit(
    as.matrix(unname(iris[, names(iris) != "Species"])),
    as.integer(iris[["Species"]])
  )
  sklearn_multiclass_svc_preds <- sklearn_multiclass_svc_model$predict(
    as.matrix(unname(iris[, names(iris) != "Species"]))
  )

  expect_equal(
    as.integer(cuda_ml_multiclass_svc_preds$.pred_class),
    as.integer(sklearn_multiclass_svc_preds)
  )
})

test_that("cuda_ml_svm() works as expected for regression tasks", {
  cuda_ml_svr_model <- cuda_ml_svm(
    formula = mpg ~ ., data = mtcars, kernel = "rbf"
  )
  cuda_ml_svr_preds <- predict(
    cuda_ml_svr_model, mtcars[, names(mtcars) != "mpg"]
  )

  sklearn_svr_model <- sklearn$svm$SVR(kernel = "rbf", gamma = "auto")
  sklearn_svr_model$fit(
    sklearn_mtcars_dataset$data, sklearn_mtcars_dataset$target
  )
  sklearn_svr_preds <- sklearn_svr_model$predict(sklearn_mtcars_dataset$data)

  expect_equal(
    cuda_ml_svr_preds$.pred, as.numeric(sklearn_svr_preds),
    tolerance = 1e-3, scale = 1
  )
})

test_that("cuda_ml_svm() classification works as expected through parsnip", {
  require("parsnip")

  cuda_ml_multiclass_svc_input <- iris[, names(iris) != "Species"]

  cuda_ml_multiclass_svc_model <- svm_rbf(mode = "classification") %>%
    set_engine("cuda.ml") %>%
    fit(Species ~ ., data = iris)
  cuda_ml_multiclass_svc_preds <- predict(
    cuda_ml_multiclass_svc_model, cuda_ml_multiclass_svc_input
  )

  sklearn_multiclass_svc_model <- sklearn$svm$SVC(kernel = "rbf", gamma = "auto")
  sklearn_multiclass_svc_model$fit(
    as.matrix(unname(iris[, names(iris) != "Species"])),
    as.integer(iris[["Species"]])
  )
  sklearn_multiclass_svc_preds <- sklearn_multiclass_svc_model$predict(
    as.matrix(unname(iris[, names(iris) != "Species"]))
  )

  expect_equal(
    as.integer(cuda_ml_multiclass_svc_preds$.pred_class),
    as.integer(sklearn_multiclass_svc_preds)
  )
})

test_that("cuda_ml_svm() regression works as expected through parsnip", {
  require("parsnip")

  cuda_ml_svr_model <- cuda_ml_svm(
    formula = mpg ~ ., data = mtcars, kernel = "rbf"
  )
  cuda_ml_svr_model <- svm_rbf(mode = "regression") %>%
    set_engine("cuda.ml") %>%
    fit(mpg ~ ., data = mtcars)
  cuda_ml_svr_preds <- predict(
    cuda_ml_svr_model, mtcars[, names(mtcars) != "mpg"]
  )

  sklearn_svr_model <- sklearn$svm$SVR(kernel = "rbf", gamma = "auto")
  sklearn_svr_model$fit(
    sklearn_mtcars_dataset$data, sklearn_mtcars_dataset$target
  )
  sklearn_svr_preds <- sklearn_svr_model$predict(sklearn_mtcars_dataset$data)

  expect_equal(
    cuda_ml_svr_preds$.pred, as.numeric(sklearn_svr_preds),
    tolerance = 1e-3, scale = 1
  )
})

umap_match_init_mode <- function(init = c("spectral", "random")) {
  init <- match.arg(init)

  switch(init,
    spectral = 1L,
    random = 0L
  )
}

umap_match_metric_type <- function(metric_type = c("categorical", "euclidean")) {
  metric_type <- match.arg(metric_type)

  switch(metric_type,
    categorical = 1L,
    euclidean = 0L
  )
}

new_umap_model <- function(model) {
  class(model) <- c("cuda_ml_umap", "cuda_ml_model", class(model))

  model
}

#' Uniform Manifold Approximation and Projection (UMAP) for dimension reduction.
#'
#' Run the Uniform Manifold Approximation and Projection (UMAP) algorithm to
#' find a low dimensional embedding of the input data that approximates an
#' underlying manifold.
#'
#' @template model-with-numeric-input
#' @template transform-input
#' @template cuML-log-level
#' @param y An optional numeric vector of target values for supervised dimension
#'   reduction. Default: NULL.
#' @param n_components The dimension of the space to embed into. Default: 2.
#' @param n_neighbors The size of local neighborhood (in terms of number of
#'   neighboring sample points) used for manifold approximation. Default: 15.
#' @param n_epochs The number of training epochs to be used in optimizing the
#'   low dimensional embedding. Default: 500.
#' @param learning_rate The initial learning rate for the embedding
#'   optimization. Default: 1.0.
#' @param init Initialization mode of the low dimensional embedding. Must be
#'   one of {"spectral", "random"}. Default: "spectral".
#' @param min_dist The effective minimum distance between embedded points.
#'   Default: 0.1.
#' @param spread The effective scale of embedded points. In combination with
#'   \code{min_dist} this determines how clustered/clumped the embedded points
#'   are. Default: 1.0.
#' @param set_op_mix_ratio Interpolate between (fuzzy) union and intersection as
#'   the set operation used to combine local fuzzy simplicial sets to obtain a
#'   global fuzzy simplicial sets. Both fuzzy set operations use the product
#'   t-norm. The value of this parameter should be between 0.0 and 1.0; a value
#'   of 1.0 will use a pure fuzzy union, while 0.0 will use a pure fuzzy
#'   intersection. Default: 1.0.
#' @param local_connectivity The local connectivity required -- i.e. the number
#'   of nearest neighbors that should be assumed to be connected at a local
#'   level. Default: 1.
#' @param repulsion_strength Weighting applied to negative samples in low
#'   dimensional embedding optimization. Values higher than one will result in
#'   greater weight being given to negative samples. Default: 1.0.
#' @param negative_sample_rate The number of negative samples to select per
#'   positive sample in the optimization process. Default: 5.
#' @param transform_queue_size For transform operations (embedding new points
#'   using a trained model this will control how aggressively to search for
#'   nearest neighbors. Default: 4.0.
#' @param a,b More specific parameters controlling the embedding. If not set,
#'   then these values are set automatically as determined by \code{min_dist}
#'   and \code{spread}. Default: NULL.
#' @param target_n_neighbors The number of nearest neighbors to use to construct
#'   the target simplcial set. Default: n_neighbors.
#' @param target_metric The metric for measuring distance between the actual and
#'   and the target values (\code{y}) if using supervised dimension reduction.
#'   Must be one of {"categorical", "euclidean"}. Default: "categorical".
#' @param target_weight Weighting factor between data topology and target
#'   topology. A value of 0.0 weights entirely on data, a value of 1.0 weights
#'   entirely on target. The default of 0.5 balances the weighting equally
#'   between data and target.
#' @param seed Optional seed for pseudo random number generator.  Default: NULL.
#'   Setting a PRNG seed will enable consistency of trained embeddings, allowing
#'   for reproducible results to 3 digits of precision, but at the expense of
#'   potentially slower training and increased memory usage.
#'   If the PRNG seed is not set, then the trained embeddings will not be
#'   deterministic.
#'
#' @return A UMAP model object that can be used as input to the
#'   \code{cuda_ml_transform()} function.
#'   If \code{transform_input} is set to TRUE, then the model object will
#'   contain a "transformed_data" attribute containing the lower dimensional
#'   embedding of the input data.
#'
#' @examples
#' library(cuda.ml)
#'
#' model <- cuda_ml_umap(
#'   x = iris[1:4],
#'   y = iris[[5]],
#'   n_components = 2,
#'   n_epochs = 200,
#'   transform_input = TRUE
#' )
#'
#' set.seed(0L)
#' print(kmeans(model$transformed, iter.max = 100, centers = 3))
#' @export
cuda_ml_umap <- function(x, y = NULL, n_components = 2L, n_neighbors = 15L,
                         n_epochs = 500L, learning_rate = 1.0,
                         init = c("spectral", "random"), min_dist = 0.1,
                         spread = 1.0, set_op_mix_ratio = 1.0,
                         local_connectivity = 1L, repulsion_strength = 1.0,
                         negative_sample_rate = 5L, transform_queue_size = 4.0,
                         a = NULL, b = NULL, target_n_neighbors = n_neighbors,
                         target_metric = c("categorical", "euclidean"),
                         target_weight = 0.5, transform_input = TRUE, seed = NULL,
                         cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace")) {
  init <- umap_match_init_mode(init)
  target_metric <- umap_match_metric_type(target_metric)
  cuML_log_level <- match_cuML_log_level(cuML_log_level)

  model_obj <- .umap_fit(
    x = as.matrix(x),
    y = if (length(y) > 0) as.numeric(y) else numeric(0),
    n_components = as.integer(n_components),
    n_neighbors = as.integer(n_neighbors),
    n_epochs = as.integer(n_epochs),
    learning_rate = as.numeric(learning_rate),
    init = init,
    min_dist = as.numeric(min_dist),
    spread = as.numeric(spread),
    set_op_mix_ratio = as.numeric(set_op_mix_ratio),
    local_connectivity = as.integer(local_connectivity),
    repulsion_strength = as.numeric(repulsion_strength),
    negative_sample_rate = as.integer(negative_sample_rate),
    transform_queue_size = as.numeric(transform_queue_size),
    a = a %||% NaN,
    b = b %||% NaN,
    target_n_neighbors = as.integer(target_n_neighbors),
    target_metric = target_metric,
    target_weight = as.numeric(target_weight),
    random_state = as.integer(seed %||% 0L),
    deterministic = !is.null(seed),
    verbosity = cuML_log_level
  )
  model <- new_umap_model(model_obj)

  if (transform_input) {
    model$transformed_data <- cuda_ml_transform(model, x)
  }

  model
}

cuda_ml_get_state.cuda_ml_umap <- function(model) {
  model_state <- .umap_get_state(model)

  new_model_state(model_state, "cuda_ml_umap_model_state")
}

cuda_ml_set_state.cuda_ml_umap_model_state <- function(model_state) {
  model_obj <- .umap_set_state(model_state)

  new_umap_model(model_obj)
}

#' @export
cuda_ml_transform.cuda_ml_umap <- function(model, x, ...) {
  .umap_transform(model = model, x = as.matrix(x))
}

context("Uniform Manifold Approximation and Projection")

iris_input <- iris[, names(iris) != "Species"]

test_that("cuda_ml_umap() works as expected", {
  umap_output <- cuda_ml_umap(
    x = iris_input, y = iris$Species, n_components = 2,
    n_epochs = 500, transform_input = TRUE, seed = 0L
  )

  verify_iris_embedding(umap_output$transformed_data)
})

agglomerative_clustering_match_metric <- function(metric = c("euclidean", "l1", "l2", "manhattan", "cosine")) {
  metric <- match.arg(metric)

  switch(metric,
    euclidean = 1L,
    l1 = 3L,
    l2 = 1L,
    manhattan = 3L,
    cosine = 2L
  )
}

#' Perform Single-Linkage Agglomerative Clustering.
#'
#' Recursively merge the pair of clusters that minimally increases a given
#' linkage distance.
#'
#' @template model-with-numeric-input
#' @param n_clusters The number of clusters to find. Default: 2L.
#' @param metric Metric used for linkage computation. Must be one of
#'   {"euclidean", "l1", "l2", "manhattan", "cosine"}. If connectivity is
#'   "knn" then only "euclidean" is accepted. Default: "euclidean".
#' @param connectivity The type of connectivity matrix to compute. Must be one
#'   of {"pairwise", "knn"}. Default: "pairwise".
#'     - 'pairwise' will compute the entire fully-connected graph of pairwise
#'        distances between each set of points. This is the fastest to compute
#'        and can be very fast for smaller datasets but requires O(n^2) space.
#'     - 'knn' will sparsify the fully-connected connectivity matrix to save
#'       memory and enable much larger inputs. "n_neighbors" will control the
#'       amount of memory used and the graph will be connected automatically in
#'       the event "n_neighbors" was not large enough to connect it.
#' @param n_neighbors The number of neighbors to compute when
#'   \code{connectivity} is "knn". Default: 15L.
#'
#' @return A clustering object with the following attributes:
#'   "n_clusters": The number of clusters found by the algorithm.
#'   "children": The children of each non-leaf node. Values less than
#'     \code{nrow(x)} correspond to leaves of the tree which are the original
#'     samples. \code{children[i + 1][1]} and \code{children[i + 1][2]} were
#'     merged to form node \code{(nrow(x) + i)} in the \code{i}-th iteration.
#'   "labels": cluster label of each data point.
#'
#' @examples
#'
#' library(cuda.ml)
#' library(MASS)
#' library(magrittr)
#' library(purrr)
#'
#' set.seed(0L)
#'
#' gen_pts <- function() {
#'   centers <- list(c(1000, 1000), c(-1000, -1000), c(-1000, 1000))
#'   pts <- centers %>%
#'     map(~ mvrnorm(50, mu = .x, Sigma = diag(2)))
#'
#'   rlang::exec(rbind, !!!pts) %>% as.matrix()
#' }
#'
#' clust <- cuda_ml_agglomerative_clustering(
#'   x = gen_pts(),
#'   metric = "euclidean",
#'   n_clusters = 3L
#' )
#'
#' print(clust$labels)
#' @export
cuda_ml_agglomerative_clustering <- function(x, n_clusters = 2L,
                                             metric = c("euclidean", "l1", "l2", "manhattan", "cosine"),
                                             connectivity = c("pairwise", "knn"),
                                             n_neighbors = 15L) {
  metric <- agglomerative_clustering_match_metric(metric)
  connectivity <- match.arg(connectivity)

  .agglomerative_clustering(
    x = as.matrix(x),
    pairwise_conn = identical(connectivity, "pairwise"),
    metric = metric,
    n_neighbors = as.integer(n_neighbors),
    n_clusters = as.integer(n_clusters)
  )
}

context("Agglomerative Clustering")

test_that("cuda_ml_agglomerative_clustering() works as expected", {
  set.seed(0)

  blob_sz <- 50
  blobs <- gen_blobs(blob_sz)
  clusters <- blobs %>%
    cuda_ml_agglomerative_clustering(metric = "euclidean", n_clusters = 3L)

  expect_equal(
    clusters$labels,
    c(1L, 2L, 0L) %>%
      purrr::map(~ rep(.x, blob_sz)) %>%
      purrr::flatten_int()
  )
})

match_cuML_log_level <- function(log_level = c("off", "critical", "error", "warn", "info", "debug", "trace")) {
  log_level <- match.arg(log_level)

  switch(log_level,
    off = 0L,
    critical = 1L,
    error = 2L,
    warn = 3L,
    info = 4L,
    debug = 5L,
    trace = 6L
  )
}

ridge_validate_alpha <- function(alpha) {
  if (alpha <= 0) {
    stop("`alpha` (multiplier of the L2 penalty term) must be positive!")
  }
}

#' Train a linear model using ridge regression.
#'
#' Train a linear model with L2 regularization.
#'
#' @template supervised-model-inputs
#' @template supervised-model-output
#' @template ellipsis-unused
#' @template fit-intercept
#' @template normalize-input
#' @param alpha Multiplier of the L2 penalty term (i.e., the result would become
#'   and Ordinary Least Square model if \code{alpha} were set to 0). Default: 1.
#'
#' @return A ridge regressor that can be used with the 'predict' S3 generic to
#'   make predictions on new data points.
#'
#' @examples
#'
#' library(cuda.ml)
#'
#' model <- cuda_ml_ridge(formula = mpg ~ ., data = mtcars, alpha = 1e-3)
#' cuda_ml_predictions <- predict(model, mtcars[names(mtcars) != "mpg"])
#'
#' # predictions will be comparable to those from a `glmnet` model with `lambda`
#' # set to 2e-3 and `alpha` set to 0
#' # (in `glmnet`, `lambda` is the weight of the penalty term, and `alpha` is
#' #  the elastic mixing parameter between L1 and L2 penalties.
#'
#' library(glmnet)
#'
#' glmnet_model <- glmnet(
#'   x = as.matrix(mtcars[names(mtcars) != "mpg"]), y = mtcars$mpg,
#'   alpha = 0, lambda = 2e-3, nlambda = 1, standardize = FALSE
#' )
#'
#' glmnet_predictions <- predict(
#'   glmnet_model, as.matrix(mtcars[names(mtcars) != "mpg"]),
#'   s = 0
#' )
#'
#' print(
#'   all.equal(
#'     as.numeric(glmnet_predictions),
#'     cuda_ml_predictions$.pred,
#'     tolerance = 1e-3
#'   )
#' )
#' @importFrom ellipsis check_dots_used
#' @export
cuda_ml_ridge <- function(x, ...) {
  check_dots_used()
  UseMethod("cuda_ml_ridge")
}

#' @rdname cuda_ml_ridge
#' @export
cuda_ml_ridge.default <- function(x, ...) {
  report_undefined_fn("cuda_ml_ridge", x)
}

#' @rdname cuda_ml_ridge
#' @export
cuda_ml_ridge.data.frame <- function(x, y,
                                     alpha = 1,
                                     fit_intercept = TRUE,
                                     normalize_input = FALSE,
                                     ...) {
  processed <- hardhat::mold(x, y)

  cuda_ml_ridge_bridge(
    processed = processed,
    alpha = alpha,
    fit_intercept = fit_intercept,
    normalize_input = normalize_input
  )
}

#' @rdname cuda_ml_ridge
#' @export
cuda_ml_ridge.matrix <- function(x, y,
                                 alpha = 1,
                                 fit_intercept = TRUE,
                                 normalize_input = FALSE,
                                 ...) {
  processed <- hardhat::mold(x, y)

  cuda_ml_ridge_bridge(
    processed = processed,
    alpha = alpha,
    fit_intercept = fit_intercept,
    normalize_input = normalize_input
  )
}

#' @rdname cuda_ml_ridge
#' @export
cuda_ml_ridge.formula <- function(formula, data,
                                  alpha = 1,
                                  fit_intercept = TRUE,
                                  normalize_input = FALSE,
                                  ...) {
  processed <- hardhat::mold(formula, data)

  cuda_ml_ridge_bridge(
    processed = processed,
    alpha = alpha,
    fit_intercept = fit_intercept,
    normalize_input = normalize_input
  )
}

#' @rdname cuda_ml_ridge
#' @export
cuda_ml_ridge.recipe <- function(x, data,
                                 alpha = 1,
                                 fit_intercept = TRUE,
                                 normalize_input = FALSE,
                                 ...) {
  processed <- hardhat::mold(x, data)

  cuda_ml_ridge_bridge(
    processed = processed,
    alpha = alpha,
    fit_intercept = fit_intercept,
    normalize_input = normalize_input
  )
}

cuda_ml_ridge_bridge <- function(processed,
                                 alpha,
                                 fit_intercept,
                                 normalize_input) {
  validate_lm_input(processed)
  ridge_validate_alpha(alpha)

  x <- as.matrix(processed$predictors)
  y <- processed$outcomes[[1]]

  model_xptr <- .ridge_fit(
    x = x,
    y = y,
    fit_intercept = fit_intercept,
    normalize_input = normalize_input,
    alpha = as.numeric(alpha),
    # TODO: future versions of libcuml may support multiple algorithms
    algo = 0L
  )

  new_linear_model(
    cls = "cuda_ml_ridge",
    xptr = model_xptr,
    blueprint = processed$blueprint
  )
}

context("Ridge Regression")

test_that("Ridge regressor works as expected", {
  for (normalize_input in c(FALSE, TRUE)) {
    for (fit_intercept in c(FALSE, TRUE)) {
      if (!fit_intercept && normalize_input) {
        next
      }

      if (!fit_intercept) {
        input <- mtcars
        input[names(mtcars) != "mpg"] <- scale(
          input[names(mtcars) != "mpg"],
          scale = FALSE
        )
        input <- as.matrix(input)
      } else {
        input <- as.matrix(mtcars)
      }

      if (normalize_input) {
        sklearn_scaler <- sklearn$preprocessing$StandardScaler(
          copy = TRUE, with_mean = TRUE, with_std = TRUE
        )
        sklearn_scaler$fit(as.matrix(mtcars[names(mtcars) != "mpg"]))
        sklearn_predictors <- sklearn_scaler$transform(
          as.matrix(mtcars[names(mtcars) != "mpg"])
        )
      } else {
        sklearn_predictors <- as.matrix(input[, which(names(mtcars) != "mpg")])
      }

      sklearn_ridge_regressor <- sklearn$linear_model$Ridge(
        alpha = 1e-3, fit_intercept = fit_intercept
      )
      sklearn_ridge_regressor$fit(
        X = sklearn_predictors,
        y = mtcars$mpg
      )
      sklearn_ridge_regressor_preds <- sklearn_ridge_regressor$predict(
        sklearn_predictors
      )

      cuda_ml_ridge_regressor <- cuda_ml_ridge(
        mpg ~ ., input,
        alpha = 1e-3,
        fit_intercept = fit_intercept,
        normalize_input = normalize_input
      )
      cuda_ml_ridge_regressor_preds <- predict(
        cuda_ml_ridge_regressor, input
      )

      expect_equal(
        cuda_ml_ridge_regressor_preds$.pred,
        as.numeric(sklearn_ridge_regressor_preds),
        tol = 0.05,
        scale = 1
      )
    }
  }
})

library(testthat)
library(cuda.ml)

if (identical(Sys.getenv("NOT_CRAN"), "true") && has_cuML()) {
  filter <- Sys.getenv("TESTTHAT_FILTER", unset = "")
  if (identical(filter, "")) filter <- NULL

  reporter <- MultiReporter$new(reporters = list(
    CheckReporter$new(),
    LocationReporter$new(),
    SummaryReporter$new(show_praise = FALSE)
  ))

  test_check("cuda.ml", filter = filter, reporter = reporter)
}

ols_match_method <- function(method = c("svd", "eig", "qr")) {
  method <- match.arg(method)

  switch(method,
    "svd" = 0L,
    "eig" = 1L,
    "qr" = 2L
  )
}

#' Train a OLS model.
#'
#' Train an Ordinary Least Square (OLS) model for regression tasks.
#'
#' @template supervised-model-inputs
#' @template supervised-model-output
#' @template ellipsis-unused
#' @template fit-intercept
#' @template normalize-input
#' @param method Must be one of {"svd", "eig", "qr"}.
#'
#'   - "svd": compute SVD decomposition using Jacobi iterations.
#'   - "eig": use an eigendecomposition of the covariance matrix.
#'   - "qr": use the QR decomposition algorithm and solve `Rx = Q^T y`.
#'
#'   If the number of features is larger than the sample size, then the
#'   "svd" algorithm will be force-selected because it is the only
#'    algorithm that can support this type of scenario.
#'
#'   Default: "svd".
#'
#' @return A OLS regressor that can be used with the 'predict' S3 generic to
#'   make predictions on new data points.
#'
#' @examples
#'
#' library(cuda.ml)
#'
#' model <- cuda_ml_ols(formula = mpg ~ ., data = mtcars, method = "qr")
#' predictions <- predict(model, mtcars[names(mtcars) != "mpg"])
#'
#' # predictions will be comparable to those from a `stats::lm` model
#' lm_model <- stats::lm(formula = mpg ~ ., data = mtcars, method = "qr")
#' lm_predictions <- predict(lm_model, mtcars[names(mtcars) != "mpg"])
#'
#' print(
#'   all.equal(
#'     as.numeric(lm_predictions),
#'     predictions$.pred,
#'     tolerance = 1e-3
#'   )
#' )
#' @importFrom ellipsis check_dots_used
#' @export
cuda_ml_ols <- function(x, ...) {
  check_dots_used()
  UseMethod("cuda_ml_ols")
}

#' @rdname cuda_ml_ols
#' @export
cuda_ml_ols.default <- function(x, ...) {
  report_undefined_fn("cuda_ml_ols", x)
}

#' @rdname cuda_ml_ols
#' @export
cuda_ml_ols.data.frame <- function(x, y,
                                   method = c("svd", "eig", "qr"),
                                   fit_intercept = TRUE,
                                   normalize_input = FALSE,
                                   ...) {
  processed <- hardhat::mold(x, y)

  cuda_ml_ols_bridge(
    processed = processed,
    method = method,
    fit_intercept = fit_intercept,
    normalize_input = normalize_input
  )
}

#' @rdname cuda_ml_ols
#' @export
cuda_ml_ols.matrix <- function(x, y,
                               method = c("svd", "eig", "qr"),
                               fit_intercept = TRUE,
                               normalize_input = FALSE,
                               ...) {
  processed <- hardhat::mold(x, y)

  cuda_ml_ols_bridge(
    processed = processed,
    method = method,
    fit_intercept = fit_intercept,
    normalize_input = normalize_input
  )
}

#' @rdname cuda_ml_ols
#' @export
cuda_ml_ols.formula <- function(formula, data,
                                method = c("svd", "eig", "qr"),
                                fit_intercept = TRUE,
                                normalize_input = FALSE,
                                ...) {
  processed <- hardhat::mold(formula, data)

  cuda_ml_ols_bridge(
    processed = processed,
    method = method,
    fit_intercept = fit_intercept,
    normalize_input = normalize_input
  )
}

#' @rdname cuda_ml_ols
#' @export
cuda_ml_ols.recipe <- function(x, data,
                               method = c("svd", "eig", "qr"),
                               fit_intercept = TRUE,
                               normalize_input = FALSE,
                               ...) {
  processed <- hardhat::mold(x, data)

  cuda_ml_ols_bridge(
    processed = processed,
    method = method,
    fit_intercept = fit_intercept,
    normalize_input = normalize_input
  )
}

cuda_ml_ols_bridge <- function(processed,
                               method,
                               fit_intercept,
                               normalize_input) {
  validate_lm_input(processed)

  x <- as.matrix(processed$predictors)
  y <- processed$outcomes[[1]]

  method <- ols_match_method(method)

  model_xptr <- .ols_fit(
    x = x,
    y = y,
    fit_intercept = fit_intercept,
    normalize_input = normalize_input,
    algo = method
  )

  new_linear_model(
    cls = "cuda_ml_ols",
    xptr = model_xptr,
    blueprint = processed$blueprint
  )
}

context("Ordinary Least Square")

test_that("OLS regressor works as expected", {
  for (normalize_input in c(FALSE, TRUE)) {
    for (fit_intercept in c(FALSE, TRUE)) {
      if (!fit_intercept) {
        input <- scale(mtcars, scale = FALSE)
      } else {
        input <- as.matrix(mtcars)
      }

      if (normalize_input) {
        sklearn_scaler <- sklearn$preprocessing$StandardScaler(
          copy = TRUE, with_mean = TRUE, with_std = TRUE
        )
        sklearn_scaler$fit(as.matrix(input))
        sklearn_input <- sklearn_scaler$transform(as.matrix(input))
      } else {
        sklearn_input <- as.matrix(input)
      }

      sklearn_ols_regressor <- sklearn$linear_model$LinearRegression(
        fit_intercept = fit_intercept
      )
      sklearn_predictors <- sklearn_input[, which(names(mtcars) != "mpg")]
      sklearn_ols_regressor$fit(
        X = sklearn_predictors,
        y = mtcars$mpg
      )
      sklearn_ols_regressor_preds <- sklearn_ols_regressor$predict(
        sklearn_input[, which(names(mtcars) != "mpg")]
      )

      for (algo in c("svd", "eig", "qr")) {
        cuda_ml_ols_regressor <- cuda_ml_ols(
          mpg ~ ., input, method = algo,
          fit_intercept = fit_intercept,
          normalize_input = normalize_input
        )
        cuda_ml_ols_regressor_preds <- predict(
          cuda_ml_ols_regressor, input[, which(names(mtcars) != "mpg")]
        )

        expect_equal(
          cuda_ml_ols_regressor_preds$.pred,
          as.numeric(sklearn_ols_regressor_preds)
        )
      }
    }
  }
})

tsne_match_method <- function(method = c("barnes_hut", "fft", "exact")) {
  method <- match.arg(method)

  switch(method,
    exact = 0L,
    barnes_hut = 1L,
    fft = 2L
  )
}

new_tsne_model <- function(embedding) {
  class(embedding) <- c("cuda_ml_tsne_model", "cuda_ml_model")

  embedding
}

#' t-distributed Stochastic Neighbor Embedding.
#'
#' t-distributed Stochastic Neighbor Embedding (TSNE) for visualizing high-
#' dimensional data.
#'
#' @template model-with-numeric-input
#' @template cuML-log-level
#' @param n_components Dimension of the embedded space.
#' @param n_neighbors The number of datapoints to use in the attractive forces.
#'   Default: ceiling(3 * perplexity).
#' @param method T-SNE method, must be one of {"barnes_hut", "fft", "exact"}.
#'   The "exact" method will be more accurate but slower. Both "barnes_hut" and
#'   "fft" methods are fast approximations.
#' @param angle Valid values are between 0.0 and 1.0, which trade off speed and
#'   accuracy, respectively. Generally, these values are set between 0.2 and
#'   0.8. (Barnes-Hut only.)
#' @param n_iter Maximum number of iterations for the optimization. Should be
#'   at least 250. Default: 1000L.
#' @param learning_rate Learning rate of the t-SNE algorithm, usually between
#'   (10, 1000). If the learning rate is too high, then t-SNE result could look
#'   like a cloud / ball of points.
#' @param learning_rate_method Must be one of {"adaptive", "none"}. If
#'   "adaptive", then learning rate, early exaggeration, and perplexity are
#'   automatically tuned based on input size. Default: "adaptive".
#' @param perplexity The target value of the conditional distribution's
#'   perplexity (see
#'   https://en.wikipedia.org/wiki/T-distributed_stochastic_neighbor_embedding
#'   for details).
#' @param perplexity_max_iter The number of epochs the best Gaussian bands are
#'   found for. Default: 100L.
#' @param perplexity_tol Stop optimizing the Gaussian bands when the conditional
#'   distribution's perplexity is within this desired tolerance compared to its
#'   taget value. Default: 1e-5.
#' @param early_exaggeration Controls the space between clusters. Not critical
#'   to tune this. Default: 12.0.
#' @param late_exaggeration Controls the space between clusters. It may be
#'   beneficial to increase this slightly to improve cluster separation. This
#'   will be applied after `exaggeration_iter` iterations (FFT only).
#' @param exaggeration_iter Number of exaggeration iterations. Default: 250L.
#' @param min_grad_norm If the gradient norm is below this threshold, the
#'   optimization will be stopped. Default: 1e-7.
#' @param pre_momentum During the exaggeration iteration, more forcefully apply
#'   gradients. Default: 0.5.
#' @param post_momentum During the late phases, less forcefully apply gradients.
#'   Default: 0.8.
#' @param square_distances Whether TSNE should square the distance values.
#' @param seed Seed to the psuedorandom number generator. Setting this can make
#'   repeated runs look more similar. Note, however, that this highly
#'   parallelized t-SNE implementation is not completely deterministic between
#'   runs, even with the same \code{seed} being used for each run.
#'   Default: NULL.
#'
#' @return A matrix containing the embedding of the input data in a low-
#'   dimensional space, with each row representing an embedded data point.
#'
#' @examples
#' library(cuda.ml)
#'
#' embedding <- cuda_ml_tsne(iris[1:4], method = "exact")
#'
#' set.seed(0L)
#' print(kmeans(embedding, centers = 3))
#' @export
cuda_ml_tsne <- function(x, n_components = 2L,
                         n_neighbors = ceiling(3 * perplexity),
                         method = c("barnes_hut", "fft", "exact"), angle = 0.5,
                         n_iter = 1000L, learning_rate = 200.0,
                         learning_rate_method = c("adaptive", "none"),
                         perplexity = 30.0, perplexity_max_iter = 100L,
                         perplexity_tol = 1e-5, early_exaggeration = 12.0,
                         late_exaggeration = 1.0, exaggeration_iter = 250L,
                         min_grad_norm = 1e-7, pre_momentum = 0.5,
                         post_momentum = 0.8, square_distances = TRUE, seed = NULL,
                         cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace")) {
  learning_rate_method <- match.arg(learning_rate_method)

  if (identical(learning_rate_method, "adaptive") &&
    method %in% c("barnes_hut", "fft")) {
    if (nrow(x) <= 2000L) {
      n_neighbors <- min(max(n_neighbors, 90L), nrow(x))
    } else {
      n_neighbors <- max(as.integer(102 - 0.0012 * nrow(x)), 30L)
    }

    pre_learning_rate <- max(nrow(x) / 3.0, 1.0)
    post_learning_rate <- pre_learning_rate
    early_exaggeration <- ifelse(nrow(x) > 10000L, 24.0, 12.0)
  } else {
    pre_learning_rate <- learning_rate
    post_learning_rate <- learning_rate * 2
  }
  algo <- tsne_match_method(method)
  cuML_log_level <- match_cuML_log_level(cuML_log_level)

  model_obj <- .tsne_fit(
    x = as.matrix(x),
    dim = as.integer(n_components),
    n_neighbors = as.integer(n_neighbors),
    algo = algo,
    initialize_embeddings = TRUE,
    square_distances = square_distances,
    theta = as.numeric(angle),
    epssq = 0.0025,
    perplexity = as.numeric(perplexity),
    perplexity_max_iter = as.integer(perplexity_max_iter),
    perplexity_tol = as.numeric(perplexity_tol),
    early_exaggeration = as.numeric(early_exaggeration),
    late_exaggeration = as.numeric(late_exaggeration),
    exaggeration_iter = as.integer(exaggeration_iter),
    min_gain = 1e-2,
    pre_learning_rate = as.numeric(pre_learning_rate),
    post_learning_rate = as.numeric(post_learning_rate),
    max_iter = as.integer(n_iter),
    min_grad_norm = as.numeric(min_grad_norm),
    pre_momentum = as.numeric(pre_momentum),
    post_momentum = as.numeric(post_momentum),
    random_state = as.integer(seed %||% -1L),
    verbosity = cuML_log_level
  )

  new_tsne_model(model_obj)
}

context("t-distributed Stochastic Neighbor Embedding")

iris_input <- iris[, names(iris) != "Species"]

test_that("cuda_ml_tsne() works as expected with 'exact' method", {
  verify_iris_embedding(cuda_ml_tsne(iris_input, method = "exact"))
})

test_that("cuda_ml_tsne() works as expected with 'fft' method", {
  verify_iris_embedding(
    cuda_ml_tsne(iris_input, method = "fft", n_iter = 50000L)
  )
})

format_msg <- function(...) {
  msg <- c(...)

  if (require(cli, quietly = TRUE)) {
    msg <- cli::ansi_strwrap(msg)
    msg <- cli::boxx(msg, border_style = "double")
  } else {
    msg <- strwrap(msg)
    msg <- paste("*\t", c("", msg, ""))
    msg <- paste(msg, collapse = "\n")
    starline <- paste(rep("*", 0.9 * getOption("width")), collapse = "")
    msg <- paste(c(starline, msg, starline), collapse = "\n")
  }

  msg
}

stop2 <- function(...) {
  stop("\n", format_msg(...), call. = FALSE)
}

warning2 <- function(...) {
  warning("\n", format_msg(...), immediate. = TRUE, call. = FALSE)
}

#' Truncated SVD.
#'
#' Dimensionality reduction using Truncated Singular Value Decomposition.
#'
#' @template model-with-numeric-input
#' @template eigen-decomposition
#' @template transform-input
#' @template cuML-log-level
#' @param n_components Desired dimensionality of output data. Must be strictly
#'   less than \code{ncol(x)} (i.e., the number of features in input data).
#'   Default: 2.
#'
#' @return A TSVD model object with the following attributes:
#'   - "components": a matrix of \code{n_components} rows to be used for
#'      dimensionalitiy reduction on new data points.
#'   - "explained_variance": (only present if "transform_input" is set to TRUE)
#'     amount of variance within the input data explained by each component.
#'   - "explained_variance_ratio": (only present if "transform_input" is set to
#'     TRUE) fraction of variance within the input data explained by each
#'     component.
#'    - "singular_values": The singular values corresponding to each component.
#'     The singular values are equal to the 2-norms of the \code{n_components}
#'     variables in the lower-dimensional space.
#'    - "tsvd_params": opaque pointer to TSVD parameters which will be used for
#'      performing inverse transforms.
#'
#' @examples
#' library(cuda.ml)
#'
#' iris.tsvd <- cuda_ml_tsvd(iris[1:4], n_components = 2)
#' print(iris.tsvd)
#' @export
cuda_ml_tsvd <- function(x,
                         n_components = 2L,
                         eig_algo = c("dq", "jacobi"),
                         tol = 1e-7, n_iters = 15L,
                         transform_input = TRUE,
                         cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace")) {
  eig_algo <- match_eig_algo(eig_algo)
  cuML_log_level <- match_cuML_log_level(cuML_log_level)

  model <- .tsvd_fit_transform(
    x = as.matrix(x),
    n_components = as.integer(n_components),
    algo = eig_algo,
    tol = as.numeric(tol),
    n_iters = as.integer(n_iters),
    transform_input = transform_input,
    verbosity = cuML_log_level
  )
  class(model) <- c("cuda_ml_tsvd", class(model))

  model
}

#' @export
cuda_ml_transform.cuda_ml_tsvd <- function(model, x, ...) {
  .tsvd_transform(model = model, x = as.matrix(x))
}

#' @export
cuda_ml_inverse_transform.cuda_ml_tsvd <- function(model, x, ...) {
  .tsvd_inverse_transform(model = model, x = as.matrix(x))
}

context("Truncated SVD")

tsvd_model <- sklearn$decomposition$TruncatedSVD(
  n_components = 2L, algorithm = "arpack"
)
sklearn_tsvd_model <- tsvd_model$fit(sklearn_iris_dataset$data)

cuda_ml_tsvd_model <- cuda_ml_tsvd(iris[1:4], n_components = 2)

# SVD components are only defined up to sign — align signs before comparing.
# For each component row, flip the cuML sign to match sklearn if the first
# non-negligible element disagrees.
align_svd_signs <- function(a, b) {
  for (i in seq_len(nrow(a))) {
    if (sign(a[i, 1]) != sign(b[i, 1])) {
      a[i, ] <- -a[i, ]
    }
  }
  a
}

test_that("cuda_ml_tsvd() works as expected", {
  sklearn_components <- sklearn_tsvd_model$components_
  aligned_components <- align_svd_signs(cuda_ml_tsvd_model$components, sklearn_components)

  expect_equal(
    aligned_components, sklearn_components,
    tolerance = 1e-8, scale = 1
  )
  expect_equal(
    cuda_ml_tsvd_model$explained_variance,
    as.numeric(sklearn_tsvd_model$explained_variance_),
    tolerance = 1e-8, scale = 1
  )
  expect_equal(
    cuda_ml_tsvd_model$explained_variance_ratio,
    as.numeric(sklearn_tsvd_model$explained_variance_ratio_),
    tolerance = 1e-8, scale = 1
  )
  expect_equal(
    cuda_ml_tsvd_model$singular_values,
    as.numeric(sklearn_tsvd_model$singular_values_),
    tolerance = 1e-8, scale = 1
  )

  # Transformed data columns also have sign ambiguity matching the components
  sklearn_transformed <- sklearn_tsvd_model$transform(sklearn_iris_dataset$data)
  cuda_transformed <- cuda_ml_tsvd_model$transformed_data
  for (j in seq_len(ncol(cuda_transformed))) {
    if (sign(cuda_transformed[1, j]) != sign(sklearn_transformed[1, j])) {
      cuda_transformed[, j] <- -cuda_transformed[, j]
    }
  }
  expect_equal(cuda_transformed, sklearn_transformed, tolerance = 1e-8, scale = 1)
})

test_that("cuda_ml_inverse_transform() works as expected for TSVD models", {
  # inverse_transform recovers the original data regardless of sign convention
  cuda_ml_reconstructed <- cuda_ml_inverse_transform(
    cuda_ml_tsvd_model, cuda_ml_tsvd_model$transformed_data
  )
  sklearn_reconstructed <- sklearn_tsvd_model$inverse_transform(
    sklearn_tsvd_model$transform(sklearn_iris_dataset$data)
  )
  expect_equal(cuda_ml_reconstructed, sklearn_reconstructed, tolerance = 1e-2, scale = 1)
})

kmeans_match_init_method <- function(m = c("kmeans++", "random")) {
  if (is.matrix(m)) {
    if (!is.numeric(m)) {
      stop(
        "Initial value of centroids must be specified using a numeric ",
        "matrix."
      )
    } else {
      2L # ML::kmeans::KMeansParams::Array
    }
  } else {
    m <- match.arg(m)
    switch(m,
      `kmeans++` = 0L,
      random = 1L
    )
  }
}

#' Run the K means clustering algorithm.
#'
#' Run the K means clustering algorithm.
#'
#' @template model-with-numeric-input
#' @template cuML-log-level
#' @param k The number of clusters.
#' @param max_iters Maximum number of iterations. Default: 300.
#' @param tol Relative tolerance with regards to inertia to declare convergence.
#'   Default: 0 (i.e., do not use inertia-based stopping criterion).
#' @param init_method Method for initializing the centroids. Valid methods
#'   include "kmeans++", "random", or a matrix of k rows, each row specifying
#'   the initial value of a centroid. Default: "kmeans++".
#' @param seed Seed to the random number generator. Default: 0.
#'
#' @return A list containing the cluster assignments and the centroid of each
#'   cluster. Each centroid will be a column within the `centroids` matrix.
#'
#' @examples
#'
#' library(cuda.ml)
#'
#' kclust <- cuda_ml_kmeans(
#'   iris[names(iris) != "Species"],
#'   k = 3, max_iters = 100
#' )
#'
#' print(kclust)
#' @export
cuda_ml_kmeans <- function(x, k, max_iters = 300, tol = 0,
                           init_method = c("kmeans++", "random"),
                           seed = 0L,
                           cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace")) {
  init_method_enum <- kmeans_match_init_method(init_method)
  centroids <- matrix(numeric(0))
  if (is.matrix(init_method)) {
    if (nrow(init_method) != k || ncol(init_method) != ncol(x)) {
      stop(
        "Initial value of centroids must be specified with a matrix of k ",
        "rows and (num features) columns, with each row containing the ",
        "initial value of a centroid."
      )
    } else {
      centroids <- init_method
    }
  }
  cuML_log_level <- match_cuML_log_level(cuML_log_level)

  .kmeans(
    x = as.matrix(x),
    k = as.integer(k),
    max_iters = as.integer(max_iters),
    tol = as.numeric(tol),
    init_method = init_method_enum,
    centroids = centroids,
    seed = as.integer(seed),
    verbosity = cuML_log_level
  )
}

context("K-Means")

sklearn_kmeans_model <- sklearn$cluster$KMeans(
  n_clusters = 3L, max_iter = 100L
)
sklearn_kclust <- sklearn_kmeans_model$fit(sklearn_iris_dataset$data)

verify_cluster_centers <- function(centers) {
  expect_equal(
    sort_mat(centers),
    sort_mat(sklearn_kclust$cluster_centers_),
    tolerance = 0.01,
    scale = 1
  )
}

test_that("cuda_ml_kmeans() works as expected with 'kmeans++' initialization method", {
  cuda_ml_kclust <- cuda_ml_kmeans(
    iris[, which(names(iris) != "Species")],
    k = 3,
    max_iters = 100,
    init_method = "kmeans++"
  )

  verify_cluster_centers(cuda_ml_kclust$centroids)
})

test_that("cuda_ml_kmeans() works as expected with 'random' initialization method", {
  cuda_ml_kclust <- cuda_ml_kmeans(
    iris[, which(names(iris) != "Species")],
    k = 3,
    max_iters = 100,
    init_method = "random"
  )

  verify_cluster_centers(cuda_ml_kclust$centroids)
})

test_that("cuda_ml_kmeans() works as expected with user-specified initial cluster centers", {
  cuda_ml_kclust <- cuda_ml_kmeans(
    iris[, which(names(iris) != "Species")],
    k = 3,
    max_iters = 100,
    init_method = sklearn_kclust$cluster_centers_
  )

  verify_cluster_centers(cuda_ml_kclust$centroids)
  expect_equal(cuda_ml_kclust$n_iter, 1)
})

find_nvcc <- function(stop_if_missing = TRUE) {
  # Check if nvcc from path is available
  nvcc_path <- "nvcc"
  cuda_version <- nvcc_version_from_path(nvcc_path)

  # Check if nvcc from CUDA_HOME is available
  cuda_home <- Sys.getenv("CUDA_HOME")
  if (nzchar(cuda_home) && is.null(cuda_version)) {
    nvcc_path <- file.path(cuda_home, "bin", "nvcc")
    cuda_version <- nvcc_version_from_path(nvcc_path)
  }

  # Check nvcc from default install location.
  if (is.null(cuda_version)) {
    nvcc_path <- "/usr/local/cuda/bin/nvcc"
    cuda_version <- nvcc_version_from_path(nvcc_path)
  }

  # No nvcc found! Error!
  if (is.null(cuda_version)) {
    if (stop_if_missing) {
      stop2(
        "Unable to locate a CUDA compiler (nvcc).",
        "Please ensure it is present in PATH (e.g., run",
        "`export PATH=\"${PATH}:/usr/local/cuda/bin\"` or",
        "similar) and try again."
      )
    } else {
      return(NULL)
    }
  }

  # Nvcc found but wrong cuda version.
  minimum_supported <- package_version("11.0")
  if (cuda_version < minimum_supported) {
    stop2(
      paste0("Found nvcc '", nvcc_path, "'"),
      paste0("CUDA version '", cuda_version, "' is not supported."),
      paste0("The minimum required version is '", minimum_supported, "'")
    )
  }

  # return nvcc path.
  return(list(path = nvcc_path, version = cuda_version))
}

nvcc_version_from_path <- function(nvcc) {
  suppressWarnings(
    nvcc <- tryCatch(system2(nvcc, "--version", stdout = TRUE, stderr = TRUE), error = function(e) NULL)
  )

  if (is.null(nvcc) || !any(grepl("release", nvcc))) {
    return(NULL)
  }

  version <- gsub(".*release |, V.*", "", nvcc[grepl("release", nvcc)])
  package_version(version)
}

#' Run the DBSCAN clustering algorithm.
#'
#' Run the DBSCAN (Density-based spatial clustering of applications with noise)
#' clustering algorithm.
#'
#' @template model-with-numeric-input
#' @template cuML-log-level
#' @param min_pts,eps A point `p` is a core point if at least `min_pts` are
#'   within distance `eps` from it.
#'
#' @return A list containing the cluster assignments of all data points. A data
#'  point not belonging to any cluster (i.e., "noise") will have NA its cluster
#'  assignment.
#'
#' @examples
#' library(cuda.ml)
#' library(magrittr)
#'
#' gen_pts <- function() {
#'   centroids <- list(c(1000, 1000), c(-1000, -1000), c(-1000, 1000))
#'
#'   pts <- centroids %>%
#'     purrr::map(~ MASS::mvrnorm(10, mu = .x, Sigma = diag(2)))
#'
#'   rlang::exec(rbind, !!!pts)
#' }
#'
#' m <- gen_pts()
#' clusters <- cuda_ml_dbscan(m, min_pts = 5, eps = 3)
#'
#' print(clusters)
#' @export
cuda_ml_dbscan <- function(x,
                           min_pts,
                           eps,
                           cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace")) {
  cuML_log_level <- match_cuML_log_level(cuML_log_level)

  res <- .dbscan(
    x = as.matrix(x),
    min_pts = min_pts,
    eps = eps,
    max_bytes_per_batch = 0L,
    verbosity = cuML_log_level
  )
  res$labels[which(res$labels == -1)] <- NA

  res
}

context("Density-Based Spatial Clustering of Applications with Noise")

test_that("cuda_ml_dbscan() works as expected", {
  set.seed(0)

  blob_sz <- 10
  blobs <- gen_blobs(blob_sz)
  clusters <- cuda_ml_dbscan(blobs, min_pts = 5, eps = 3)

  expect_equal(
    clusters$labels,
    seq(0, 2) %>%
      purrr::map(~ rep(.x, blob_sz)) %>%
      purrr::flatten_int()
  )
})

library(magrittr, warn.conflicts = FALSE)
library(reticulate)
library(rlang, warn.conflicts = FALSE)

expect_libcuml <- function() {
  if (!has_cuML()) {
    stop(
      "The current installation of {cuda.ml} is not linked with a valid copy of",
      " the RAPIDS cuML shared library!\n",
      ".libPaths:\n",
      paste(.libPaths(), collapse = "\n")
    )
  }
}

expect_libcuml()

reticulate::py_require("scikit-learn")
sklearn <- reticulate::import("sklearn")
sklearn_iris_dataset <- list(
  data = iris[, names(iris) != "Species"] %>%
    unname() %>%
    as.matrix(),
  target = as.integer(iris[["Species"]])
)
sklearn_mtcars_dataset <- list(
  data = mtcars[, names(mtcars) != "mpg"] %>%
    data.frame(row.names = NULL) %>%
    unname() %>%
    as.matrix(),
  target = mtcars[["mpg"]]
)

#' Sort matrix rows by all columns or by a subset of columns.
#'
#' @param cols Indices of columns used for sorting.
sort_mat <- function(m, cols = seq(ncol(m))) {
  m[do.call(order, lapply(cols, function(x) m[, x])), ]
}

#' Attempt to unserialize a CuML model within a sub-process and use the
#' unserialized model to make predictions.
predict_in_sub_proc <- function(model_state, data, expected_mode,
                                expected_model_cls = NULL,
                                additional_predict_args = list()) {
  impl <- function(expect_libcuda_ml_impl, model_state, data, expected_mode,
                   expected_model_cls, additional_predict_args) {
    library(cuda.ml)
    expect_libcuda_ml_impl()

    model <- cuda_ml_unserialize(model_state)
    for (cls in expected_model_cls) {
      testthat::expect_s3_class(model, cls)
    }
    stopifnot(identical(model$mode, expected_mode))

    do.call(predict, append(list(model, data), additional_predict_args))
  }

  callr::r(
    impl,
    args = list(
      expect_libcuda_ml_impl = expect_libcuml,
      model_state = model_state,
      data = data,
      expected_mode = expected_mode,
      expected_model_cls = expected_model_cls,
      additional_predict_args = additional_predict_args
    ),
    stdout = "", stderr = ""
  )
}

gen_blobs <- function(blob_sz = 10, centers = NULL) {
  centers <- centers %||% list(c(1000, 1000), c(-1000, -1000), c(-1000, 1000))
  pts <- centers %>%
    purrr::map(~ MASS::mvrnorm(blob_sz, mu = .x, Sigma = diag(length(.))))

  rlang::exec(rbind, !!!pts)
}

verify_iris_embedding <- function(embedding) {
  set.seed(0L)
  k_clust <- kmeans(embedding, centers = embedding[c(1, 51, 101), ])

  # i.e., one should be able to obtain a reasonably good clustering result
  # (as measured by the BSS/TSS ratio) within very few k-means iterations on the
  # embedding.
  expect_lte(k_clust$iter, 3)
  expect_gte(k_clust$betweenss / k_clust$totss, 0.95)

  # Use `iris$Species` to check pairs of data points from the same species are
  # mostly in the same cluster, and those from different species are mostly in
  # different clusters in the resulting clustering.
  expect_gte(
    sklearn$metrics$adjusted_rand_score(
      labels_true = iris$Species, labels_pred = k_clust$cluster
    ),
    0.7
  )
}

context("K-Nearest Neighbors")

set.seed(0)
blob_sz <- 1000
centers <- list(c(3, 3), c(-3, -3), c(-3, 3))
blobs <- gen_blobs(blob_sz, centers)
blobs_df <- blobs %>%
  as.data.frame() %>%
  cbind(
    label = seq_along(centers) %>%
      sapply(function(x) rep(x, blob_sz)) %>%
      factor()
  )

test_blob_sz <- 10

test_that("KNN classifier works as expected", {
  test_blobs_df <- gen_blobs(test_blob_sz, centers) %>%
    as.data.frame()
  for (algo in c("brute", "ivfflat", "ivfpq", "ivfsq")) {
    model <- cuda_ml_knn(
      label ~ ., blobs_df,
      algo = algo, metric = "euclidean"
    )
    preds <- predict(model, test_blobs_df)

    expect_equal(
      as.integer(preds$.pred_class),
      seq(3) %>%
        purrr::map(~ rep(.x, test_blob_sz)) %>%
        purrr::flatten_int(),
      label = algo
    )
  }
})

test_that("KNN regressor works as expected", {
  resps <- seq_along(centers) %>%
    sapply(function(x) rep(exp(-x), blob_sz)) %>%
    c()
  train_df <- blobs %>%
    as.data.frame() %>%
    cbind(y = resps)
  test_blobs <- gen_blobs(test_blob_sz, centers)

  cuda_ml_knn_regressor <- cuda_ml_knn(
    y ~ .,
    data = train_df, algo = "brute", metric = "euclidean", neighbors = 5L
  )
  cuda_ml_knn_regressor_preds <- predict(
    cuda_ml_knn_regressor, as.data.frame(test_blobs)
  )

  sklearn_knn_regressor <- sklearn$neighbors$KNeighborsRegressor(
    n_neighbors = 5L, algorithm = "brute", metric = "euclidean"
  )
  sklearn_knn_regressor$fit(X = blobs, y = resps)
  sklearn_knn_regressor_preds <- sklearn_knn_regressor$predict(
    as.matrix(test_blobs)
  )

  expect_equal(
    cuda_ml_knn_regressor_preds$.pred, as.numeric(sklearn_knn_regressor_preds)
  )
})

test_that("KNN classifier works as expected through parsnip", {
  require("parsnip")

  test_blobs_df <- gen_blobs(test_blob_sz, centers) %>%
    as.data.frame()
  model <- nearest_neighbor(
    mode = "classification", neighbors = 10, dist_power = 2
  ) %>%
    set_engine("cuda.ml") %>%
    fit(label ~ ., blobs_df)
  preds <- predict(model, test_blobs_df)

  expect_equal(
    as.integer(preds$.pred_class),
    seq(3) %>%
      purrr::map(~ rep(.x, test_blob_sz)) %>%
      purrr::flatten_int()
  )
})

test_that("KNN regressor works as expected through parsnip", {
  require("parsnip")

  resps <- seq_along(centers) %>%
    sapply(function(x) rep(exp(-x), blob_sz)) %>%
    c()
  train_df <- blobs %>%
    as.data.frame() %>%
    cbind(y = resps)
  test_blobs <- gen_blobs(test_blob_sz, centers)

  cuda_ml_knn_regressor <- nearest_neighbor(
    mode = "regression", neighbors = 5, dist_power = 2
  ) %>%
    set_engine("cuda.ml") %>%
    fit(y ~ ., data = train_df)
  cuda_ml_knn_regressor_preds <- predict(
    cuda_ml_knn_regressor, as.data.frame(test_blobs)
  )

  sklearn_knn_regressor <- sklearn$neighbors$KNeighborsRegressor(
    n_neighbors = 5L, algorithm = "brute", metric = "euclidean"
  )
  sklearn_knn_regressor$fit(X = blobs, y = resps)
  sklearn_knn_regressor_preds <- sklearn_knn_regressor$predict(
    as.matrix(test_blobs)
  )

  expect_equal(
    cuda_ml_knn_regressor_preds$.pred, as.numeric(sklearn_knn_regressor_preds)
  )
})

context("(de)serialization of Random Forest models")

test_that("random forest classifier can be serialized and unserialized correctly", {
  model <- cuda_ml_rand_forest(formula = Species ~ ., data = iris, trees = 200)
  model_state <- cuda_ml_serialize(model)

  data <- iris[-which(names(iris) == "Species")]

  expected_preds <- predict(model, data)
  actual_preds <- predict_in_sub_proc(
    model_state,
    data = data,
    expected_mode = "classification",
    expected_model_cls = "cuda_ml_rand_forest"
  )

  expect_equal(expected_preds, actual_preds)

  if (as.integer(cuML_minor_version()) >= 8) {
    # class probabilities output was not supported in earlier versions of RAPIDS
    # cuML
    expected_cls_probs <- predict(
      model, data,
      output_class_probabilities = TRUE
    )
    actual_cls_probs <- predict_in_sub_proc(
      model_state,
      data = data,
      expected_mode = "classification",
      expected_model_cls = "cuda_ml_rand_forest",
      additional_predict_args = list(output_class_probabilities = TRUE)
    )
    expect_equal(expected_cls_probs, actual_cls_probs, tolerance = 1e-3, scale = 1)
  }
})

test_that("random forest regressor can be serialized and unserialized correctly", {
  model <- cuda_ml_rand_forest(formula = mpg ~ ., data = mtcars, trees = 200)
  model_state <- cuda_ml_serialize(model)

  data <- mtcars[-which(names(mtcars) == "mpg")]

  expected_preds <- predict(model, data)
  actual_preds <- predict_in_sub_proc(
    model_state,
    data = data,
    expected_mode = "regression",
    expected_model_cls = "cuda_ml_rand_forest"
  )

  expect_equal(expected_preds, actual_preds, tolerance = 1e-4, scale = 1)
})

#' @importFrom stats predict
NULL

match_eig_algo <- function(eig_algo = c("dq", "jacobi")) {
  eig_algo <- match.arg(eig_algo)

  switch(eig_algo,
    dq = 0L,
    jacobi = 1L
  )
}

new_model <- function(cls,
                      mode = c("classification", "regression"),
                      xptr = NULL, ...) {
  mode <- match.arg(mode)
  do.call(
    hardhat::new_model,
    c(
      list(class = c(cls, "cuda_ml_model"), mode = mode, xptr = xptr),
      rlang::dots_list(...)
    )
  )
}

new_model_state <- function(model_state, cls) {
  class(model_state) <- c(cls, "cuda_ml_model_state", class(model_state))

  model_state
}

get_pred_levels <- function(model) {
  levels(model$blueprint$ptypes$outcomes[[1]])
}

postprocess_classification_results <- function(predictions, model) {
  predictions <- as.integer(predictions)
  pred_levels <- get_pred_levels(model)
  predictions <- pred_levels[predictions]
  predictions <- factor(predictions, levels = pred_levels)

  hardhat::spruce_class(predictions)
}

postprocess_class_probabilities <- function(predictions, model) {
  pred_levels <- get_pred_levels(model)

  hardhat::spruce_prob(pred_levels, predictions)
}

postprocess_regression_results <- function(predictions) {
  hardhat::spruce_numeric(predictions)
}

report_undefined_fn <- function(fn_name, x) {
  stop(
    "`", fn_name, "()` is undefined for object of class ",
    paste(class(x), sep = " "),
    call. = FALSE
  )
}

#' Transform data using a trained cuML model.
#'
#' Given a trained cuML model, transform an input dataset using that model.
#'
#' @template cudaml-transform
#'
#' @importFrom ellipsis check_dots_used
#' @export
cuda_ml_transform <- function(model, x, ...) {
  check_dots_used()
  UseMethod("cuda_ml_transform")
}

#' Apply the inverse transformation defined by a trained cuML model.
#'
#' Given a trained cuML model, apply the inverse transformation defined by that
#' model to an input dataset.
#'
#' @template cudaml-transform
#'
#' @importFrom ellipsis check_dots_used
#' @export
cuda_ml_inverse_transform <- function(model, x, ...) {
  check_dots_used()
  UseMethod("cuda_ml_inverse_transform")
}

#' Determine whether a CuML model is a classifier.
#'
#' Given a trained CuML model, return \code{TRUE} if the model is a classifier,
#' otherwise \code{FALSE} (e.g., if the model is a regressor).
#'
#' @param model A trained CuML model.
#'
#' @return A logical value indicating whether the model is a classifier.
#'
#' @export
cuda_ml_is_classifier <- function(model) {
  UseMethod("cuda_ml_is_classifier")
}

#' @export
cuda_ml_is_classifier.default <- function(model) {
  report_undefined_fn("cuda_ml_is_classifier", model)
}

#' @export
cuda_ml_is_classifier.cuda_ml_model <- function(model) {
  identical(model$mode, "classification")
}

#' Determine whether a CuML model can predict class probabilities.
#'
#' Given a trained CuML model, return \code{TRUE} if the model is a classifier
#' and is capable of outputting class probabilities as prediction results (e.g.,
#' if the model is a KNN or an ensemble classifier), otherwise return
#' \code{FALSE}.
#'
#' @param model A trained CuML model.
#'
#' @return A logical value indicating whether the model supports outputting
#'   class probabilities.
#'
#' @export
cuda_ml_can_predict_class_probabilities <- function(model) {
  UseMethod("cuda_ml_can_predict_class_probabilities")
}

#' @export
cuda_ml_can_predict_class_probabilities.default <- function(model) {
  report_undefined_fn("cuda_ml_can_predict_class_probabilities", model)
}

#' @export
cuda_ml_can_predict_class_probabilities.cuda_ml_model <- function(model) {
  FALSE
}

#' @export
cuda_ml_can_predict_class_probabilities.cuda_ml_fil <- function(model) {
  cuda_ml_is_classifier(model)
}

#' @export
cuda_ml_can_predict_class_probabilities.cuda_ml_knn <- function(model) {
  cuda_ml_is_classifier(model)
}

#' @export
cuda_ml_can_predict_class_probabilities.cuda_ml_rand_forest <- function(model) {
  cuda_ml_is_classifier(model)
}

#' Serialize a CuML model
#'
#' Given a CuML model, serialize its state into a connection.
#'
#' @param model The model object.
#' @param connection An open connection or \code{NULL}. If \code{NULL}, then the
#'   model state is serialized to a raw vector. Default: NULL.
#' @param ... Additional arguments to \code{base::serialize()}.
#'
#' @return \code{NULL} unless \code{connection} is \code{NULL}, in which case
#'   the serialized model state is returned as a raw vector.
#'
#' @seealso \code{\link[base]{serialize}}
#'
#' @export
cuda_ml_serialize <- function(model, connection = NULL, ...) {
  UseMethod("cuda_ml_serialize")
}

#' @rdname cuda_ml_serialize
#'
#' @export
cuda_ml_serialise <- cuda_ml_serialize

#' @export
cuda_ml_serialize.default <- function(model, connection = NULL, ...) {
  report_undefined_fn("cuda_ml_serialize", model)
}

#' @export
cuda_ml_serialize.cuda_ml_model <- function(model, connection = NULL, ...) {
  model_state <- cuda_ml_get_state(model)

  serialize(model_state, connection, ...)
}

cuda_ml_get_state <- function(model) {
  UseMethod("cuda_ml_get_state")
}

#' @export
cuda_ml_get_state.default <- function(model) {
  stop(
    "Model of type '", paste(class(model), collapse = " "), "' does not ",
    "support serialization."
  )
}

#' @export
cuda_ml_get_state.cuda_ml_model <- function(model) {
  # Default implementation: assume the entire model object can be serializabled
  # by `base::serialize()`.
  model_state <- list(model = model)

  new_model_state(model_state, cls = NULL)
}

#' @export
cuda_ml_set_state.cuda_ml_model_state <- function(model_state) {
  # Default implementation: assume the entire model state can be unserialized by
  # `base::unserialize()`.
  model_state$model
}

#' Unserialize a CuML model state
#'
#' Unserialize a CuML model state into a CuML model object.
#'
#' @param connection An open connection or a raw vector.
#' @param ... Additional arguments to \code{base::unserialize()}.
#'
#' @return A unserialized CuML model.
#'
#' @seealso \code{\link[base]{unserialize}}
#'
#' @export
cuda_ml_unserialize <- function(connection, ...) {
  model_state <- unserialize(connection, ...)

  cuda_ml_set_state(model_state)
}

#' @rdname cuda_ml_unserialize
#'
#' @export
cuda_ml_unserialise <- cuda_ml_unserialize


cuda_ml_set_state <- function(model_state) {
  UseMethod("cuda_ml_set_state")
}

#' @export
cuda_ml_set_state.default <- function(model_state) {
  stop(
    "No unserialization routine found for model state of type '",
    paste(class(model_state), collapse = " "), "'"
  )
}

context("(de)serialization of TSNE models")

test_that("TSNE models can be serialized and unserialized correctly", {
  embedding <- cuda_ml_tsne(iris[1:4], method = "exact")
  model_state <- cuda_ml_serialize(embedding)

  expect_equal(
    embedding,
    callr::r(
      function(model_state) {
        library(cuda.ml)

        stopifnot(has_cuML())
        cuda_ml_unserialize(model_state)
      },
      args = list(model_state = model_state)
    )
  )
})

context("(de)serialization of Random Projection models")

test_that("random projection model can be serialized and unserialized correctly", {
  has_mlbench <- require("mlbench")
  stopifnot(has_mlbench)

  data(Vehicle)
  data <- Vehicle[, which(names(Vehicle) != "Class")]

  model <- cuda_ml_rand_proj(data, n_components = 4)
  model_state <- cuda_ml_serialize(model)

  actual_transformed_data <- callr::r(
    function(model_state, data) {
      library(cuda.ml)

      model <- cuda_ml_unserialize(model_state)

      cuda_ml_transform(model, data)
    },
    args = list(
      model_state = model_state,
      data = data
    )
  )

  expect_equal(actual_transformed_data, model$transformed_data)
})

context("(de)serialization of PCA models")

test_that("PCA models can be serialized and unserialized correctly", {
  models <- list()

  for (whiten in c(FALSE, TRUE)) {
    for (transform_input in c(FALSE, TRUE)) {
      model <- cuda_ml_pca(
        iris[1:4],
        n_components = 3, whiten = whiten,
        transform_input = transform_input
      )
      models <- append(models, list(model))
    }
  }

  for (model in models) {
    actual_inv_transform <- callr::r(
      function(model_state, expected_components, expected_expl_var,
               expected_expl_var_ratio, expected_sg_vals, expected_m,
               expected_tf_data, whiten) {
        library(cuda.ml)
        library(testthat)

        stopifnot(has_cuML())

        model <- cuda_ml_unserialize(model_state)

        expect_equal(model$components, expected_components)
        expect_equal(model$explained_variance, expected_expl_var)
        expect_equal(
          model$explained_variance_ratio, expected_expl_var_ratio
        )
        expect_equal(model$singular_values, expected_sg_vals)
        expect_equal(model$mean, expected_m)
        expect_equal(model$transformed_data, expected_tf_data)

        if (!is.null(model$transformed_data) && !whiten) {
          # TODO: look into why this may fail when `whiten` is `TRUE`
          cuda_ml_inverse_transform(model, model$transformed_data)
        } else {
          NULL
        }
      },
      args = list(
        model_state = cuda_ml_serialize(model),
        expected_components = model$components,
        expected_expl_var = model$explained_variance,
        expected_expl_var_ratio = model$explained_variance_ratio,
        expected_sg_vals = model$singular_values,
        expected_m = model$mean,
        expected_tf_data = model$transformed_data,
        whiten = whiten
      )
    )

    if (!is.null(model$transformed_data) && !whiten) {
      # TODO: look into why this may fail when `whiten` is `TRUE`
      expected_inv_transform <- cuda_ml_inverse_transform(
        model, model$transformed_data
      )

      expect_equal(expected_inv_transform, actual_inv_transform)
    } else {
      succeed()
    }
  }
})

#' Options:
#'
#' CUML_VERSION: Specifies the version of the `libcuml` library to be downloaded
#'               and installed. This version must be present in the
#'               `libcuml_versions` list below.
#'
#' CUML_URL: Override the URL to download `libcuml` from. If specified, then no
#'           checks for CUDA version will be performed.
#'
#' CUML_PREFIX: If you have a copy of `libcuml` installed already, you can
#'              specify this environment variable to link {cuda.ml} with an
#'              existing installation of `libcuml`.
#'              If a valid copy of `libcuml` is found in '/usr' or in
#'              "${CUML_PREFIX}/", then no pre-built copy of `libcuml` will be
#'              downloaded.
#'
#' DOWNLOAD_CUML: The default is to automatically download a pre-built copy of
#'                `libcuml` if no existing `libcuml` is specified with the
#'                'CUML_PREFIX' env variable. Set DOWNLOAD_CUML=0 to disable
#'                this default behavior.
#'
#' DISABLE_PARALLEL_BUILD: Parallel build using max($(nproc) - 1, 1) cores is
#'                         enabled by default but can be disabled by setting
#'                         this env variable.
#'
#' CUML4R_CMAKE_PARALLEL_LEVEL: If not set and parallel build is enabled, then
#'                              max($(nproc) - 1, 1) cores will be used by the
#'                              build process. If set, then the number of cores
#'                              specified will be used.

pkg_root <- function() {
  # devtools::load_all() might run the config script from the `src` directory.
  for (p in list(".", "..")) {
    if (file.exists(file.path(p, "DESCRIPTION"))) {
      return(normalizePath(p))
    }
  }

  # should never reach here
  pkg_root <- normalizePath(".")
  warning(
    "Unable to locate 'DESCRIPTION' file! Assuming pkg root is '",
    pkg_root, "'."
  )
  return(pkg_root)
}

load_libcuml_versions <- function() {
  wd <- file.path(pkg_root(), "tools", "config")

  source(file.path(wd, "libcuml_versions.R"))
}

load_util_fns <- function() {
  wd <- file.path(pkg_root(), "tools", "config", "utils")

  for (f in c("cuml.R", "cmake.R", "logging.R", "nvcc.R", "platform.R")) {
    source(file.path(wd, f))
  }
}

load_libcuml_versions()
load_util_fns()

run_cmake <- function() {
  wd <- getwd()
  on.exit(setwd(wd))
  setwd(pkg_root())

  define(R_INCLUDE_DIR = R.home("include"))
  define(RCPP_INCLUDE_DIR = system.file("include", package = "Rcpp"))
  configure_file(file.path("src", "CMakeLists.txt.in"))

  cuml_prefix <- get_cuml_prefix()
  bundle_libcuml <- FALSE
  if (is.na(cuml_prefix)) {
    download_libcuml()
    cuml_prefix <- normalizePath(file.path(pkg_root(), "libcuml"))
    dir.create("inst")
    file.rename(file.path("libcuml", "lib"), file.path("inst", "libs"))
    file.symlink(file.path("..", "inst", "libs"), file.path("libcuml", "lib"))
    libs <- c("libtreelite", "libtreelite_runtime", "libcuml++")
    bundle_libcuml <- TRUE
  }
  cmake_prefix_path <- paste0(
    c(Sys.getenv("CMAKE_PREFIX_PATH", unset = ""), cuml_prefix),
    collapse = ":"
  )
  Sys.setenv(CMAKE_PREFIX_PATH = cmake_prefix_path)

  setwd(file.path(pkg_root(), "src"))

  cmake_args <- c(
    ".",
    paste0("-DCMAKE_CUDA_ARCHITECTURES=", Sys.getenv("CMAKE_CUDA_ARCHITECTURES", unset = "NATIVE")),
    paste0("-DCUML_INCLUDE_DIR=", file.path(cuml_prefix, "include")),
    paste0("-DCUML_LIB_DIR=", file.path(cuml_prefix, "lib")),
    paste0(
      "-DCUML_STUB_HEADERS_DIR=", normalizePath(file.path(getwd(), "stubs"))
    ),
    paste0("-DCMAKE_CUDA_COMPILER=", find_nvcc()$path),
    "-DCMAKE_VERBOSE_MAKEFILE:BOOL=TRUE"
  )
  if (bundle_libcuml) {
    cmake_args <- c(
      cmake_args,
      "-DCMAKE_BUILD_WITH_INSTALL_RPATH:BOOL=TRUE",
      "-DCMAKE_INSTALL_RPATH:STRING='$ORIGIN'"
    )
  }
  cmake_bin <- find_or_download_cmake(
    min_version = cuda_ml_min_cmake_version,
    exdir = file.path(pkg_root(), "tools")
  )
  rc <- system2(cmake_bin, args = cmake_args)

  if (rc != 0) {
    stop("Failed to run 'cmake'!")
  }
}

if (is.null(find_nvcc(stop_if_missing = FALSE)) || !has_libcuml()) {
  wd <- getwd()
  on.exit(setwd(wd))
  setwd(pkg_root())
  define(STUBS_HEADERS_DIR = normalizePath(file.path(getwd(), "src", "stubs")))
  define(CUSTOMIZED_MAKEFLAGS = "")
} else {
  define(STUBS_HEADERS_DIR = "")
  n_jobs <- (
    if (!is.na(Sys.getenv("DISABLE_PARALLEL_BUILD", unset = NA))) {
      1L
    } else {
      user_specified_parallel_level <- Sys.getenv("CUML4R_CMAKE_PARALLEL_LEVEL", unset = NA)
      if (!is.na(user_specified_parallel_level)) {
        as.integer(user_specified_parallel_level)
      } else {
        max(nproc() - 1L, 1L)
      }
    })
  define(CUSTOMIZED_MAKEFLAGS = paste0("MAKEFLAGS += '-j", n_jobs, "'"))

  run_cmake()
}

# A list containing libcuml download links for "cuml_versions" and CUDA major versions.
libcuml_versions <- list(
  "21.08" = list(
    "11" = "https://github.com/mlverse/libcuml-builds/releases/download/v21.08-cuda11.2.1/libcuml-21.08-cuda11.2.1.zip"
  ),
  "21.10" = list(
    "11" = "https://github.com/mlverse/libcuml-builds/releases/download/v21.10-cuda11.2.1/libcuml-21.10-cuda11.2.1.zip"
  ),
  "21.12" = list(
    "11" = "https://github.com/mlverse/libcuml-builds/releases/download/v21.12-cuda11.2.1/libcuml-21.12-cuda11.2.1.zip"
  )
)

#' cuda.ml
#'
#' This package provides a R interface for the RAPIDS cuML library.
#'
#' @author Yitao Li <yitao@rstudio.com>
#' @import Rcpp
#' @useDynLib cuda.ml, .registration = TRUE
"_PACKAGE"

.onLoad <- function(libname, pkgname) {
  register_rand_forest_model(pkgname)
  register_svm_model(pkgname)
  register_knn_model(pkgname)
}

.onAttach <- function(libname, pkgname) {
  if (!has_cuML()) {
    packageStartupMessage(
      "
      The current installation of {", pkgname, "} will not function as expected
      because it was not linked with a valid version of the RAPIDS cuML shared
      library.

      To fix this issue, please follow https://rapids.ai/start.html#get-rapids
      to install the RAPIDS cuML shared library from Conda and ensure the
      'CUML_PREFIX' env variable is set to a valid RAPIDS conda env directory
      (e.g., '/home/user/anaconda3/envs/rapids-21.06', '/usr', or similar)
      during the installation of {", pkgname, "} or alternatively, follow
      https://github.com/yitao-li/cuml-installation-notes#build-from-source-without-conda-and-without-multi-gpu-support
      or
      https://github.com/yitao-li/cuml-installation-notes#build-from-source-without-conda-and-with-multi-gpu-support
      or similar to build and install RAPIDS cuML library from source, and
      then re-install {", pkgname, "}.\n\n
      "
    )
  }
}

# NOTE: this version threshold may change depending on which branch of
#       https://github.com/rapidsai/rapids-cmake.git is being used
#       (e.g., see
#        https://github.com/rapidsai/rapids-cmake/blob/d5d59dbfaff7feafdd87700e8c2b347188897af7/CMakeLists.txt#L28
#        where the minimum cmake version is specified in the rapids-cmake repo in branch-21.10,
#        and
#        https://github.com/mlverse/cuda.ml/blob/7bad914c729011bcf05edc1c873609c518d9a77d/src/CMakeLists.txt.in#L13
#        where cuda.ml specifies which branch of the rapids-cmake repo to use)
cuda_ml_min_cmake_version <- numeric_version("3.21.1")

has_cmake <- function() {
  rc <- system2("which", "cmake", stdout = NULL, stderr = NULL)

  rc == 0
}

cmake_version <- function() {
  tryCatch(
    if (has_cmake()) {
      cmake_version_line <- system2("cmake", "--version", stdout = TRUE)[[1]]

      m <- regexec(".*cmake version (\\d+\\.\\d+\\.\\d+)", cmake_version_line)
      numeric_version(regmatches(cmake_version_line, m)[[1]][[2]])
    } else {
      NULL
    },
    error = function(e) NULL
  )
}

download_cmake <- function(cmake_version, exdir) {
  dl_timeout <- options("timeout")
  on.exit(options(timeout = dl_timeout), add = TRUE)

  dest <- paste0(tempfile(paste0("cmake-", cmake_version)), ".tar.gz")
  url <- paste0(
    "https://github.com/Kitware/CMake/releases/download/v", cmake_version,
    "/cmake-", cmake_version, "-linux-x86_64.tar.gz"
  )
  download.file(url = url, destfile = dest)

  untar(tarfile = dest, exdir = exdir)

  cmake_bin <- file.path(
    exdir, paste0("cmake-", cmake_version, "-linux-x86_64"), "bin", "cmake"
  )

  cmake_bin
}

# If 'cmake' is present in PATH and 'cmake --version' shows a version that is
# not lower than `cuda_ml_min_cmake_version`, then return 'cmake', otherwise
# download the required cmake binary (x86_64 only at the moment) to a temporary
# location and return the path to the temporary cmake binary.
#
# NOTE: if we want to support aarch64 in future, then this function needs to be
#       modified to download the correct binary on a non-x86_64 platform
find_or_download_cmake <- function(min_version, exdir) {
  existing_cmake_version <- cmake_version()
  if (is.null(existing_cmake_version) || existing_cmake_version < min_version) {
    download_cmake(cmake_version = min_version, exdir = exdir)
  } else {
    "cmake"
  }
}

new_rproj_model <- function(rproj_ctx) {
  model <- list(rproj_ctx = rproj_ctx)
  class(model) <- c("cuda_ml_rand_proj_model", "cuda_ml_model", class(model))

  model
}

#' Random projection for dimensionality reduction.
#'
#' Generate a random projection matrix for dimensionality reduction, and
#' optionally transform input data to a projection in a lower dimension space
#' using the generated random matrix.
#'
#' @template model-with-numeric-input
#' @param n_components Dimensionality of the target projection space. If NULL,
#'   then the parameter is deducted using the Johnson-Lindenstrauss lemma,
#'   taking into consideration the number of samples and the \code{eps}
#'   parameter. Default: NULL.
#' @param eps Error tolerance during projection. Default: 0.1.
#' @param gaussian_method If TRUE, then use the Gaussian random projection
#'   method. Otherwise, use the sparse random projection method.
#'   See https://en.wikipedia.org/wiki/Random_projection for details.
#'   Default: TRUE.
#' @param density Ratio of non-zero component in the random projection matrix.
#'   If NULL, then the value is set to the minimum density as recommended by
#'   Ping Li et al.: 1 / sqrt(n_features). Default: NULL.
#' @param transform_input Whether to project input data onto a lower dimension
#'   space using the random matrix. Default: TRUE.
#' @param seed Seed for the pseudorandom number generator. Default: 0L.
#'
#' @return A context object containing GPU pointer to a random matrix that can
#'   be used as input to the \code{cuda_ml_transform()} function.
#'   If \code{transform_input} is set to TRUE, then the context object will also
#'   contain a "transformed_data" attribute containing the lower dimensional
#'   projection of the input data.
#'
#' @examples
#' library(cuda.ml)
#' library(mlbench)
#'
#' data(Vehicle)
#' vehicle_data <- Vehicle[order(Vehicle$Class), which(names(Vehicle) != "Class")]
#'
#' model <- cuda_ml_rand_proj(vehicle_data, n_components = 4)
#'
#' set.seed(0L)
#' print(kmeans(model$transformed_data, centers = 4, iter.max = 1000))
#' @export
cuda_ml_rand_proj <- function(x, n_components = NULL, eps = 0.1,
                              gaussian_method = TRUE, density = NULL,
                              transform_input = TRUE, seed = 0L) {
  n_components <- n_components %||%
    .rproj_johnson_lindenstrauss_min_dim(
      n_samples = nrow(x), eps = as.numeric(eps)
    )
  density <- density %||% 1.0 / sqrt(ncol(x))

  rproj_ctx <- .rproj_fit(
    n_samples = nrow(x),
    n_features = ncol(x),
    n_components = as.integer(n_components),
    eps = as.numeric(eps),
    gaussian_method = gaussian_method,
    density = as.numeric(density),
    random_state = as.integer(seed)
  )

  model <- new_rproj_model(rproj_ctx)

  if (transform_input) {
    model$transformed_data <- cuda_ml_transform(model, x)
  }

  model
}

#' @export
cuda_ml_transform.cuda_ml_rand_proj_model <- function(model, x, ...) {
  .rproj_transform(model$rproj_ctx, as.matrix(x))
}

cuda_ml_get_state.cuda_ml_rand_proj_model <- function(model) {
  model_state <- .rproj_get_state(model$rproj_ctx)

  new_model_state(model_state, "cuda_ml_rand_proj_model_state")
}

cuda_ml_set_state.cuda_ml_rand_proj_model_state <- function(model_state) {
  model_obj <- .rproj_set_state(model_state)

  new_rproj_model(model_obj)
}

#' Determine whether \{cuda.ml\} was linked to a valid version of the RAPIDS cuML
#' shared library.
#'
#' @return A logical value indicating whether the current installation \{cuda.ml\}
#'   was linked to a valid version of the RAPIDS cuML shared library.
#'
#' @examples
#'
#' library(cuda.ml)
#'
#' if (!has_cuML()) {
#'   warning(
#'     "Please install the RAPIDS cuML shared library first, and then re-",
#'     "install {cuda.ml}."
#'   )
#' }
#' @export
has_cuML <- .has_cuML

#' Get the major version of the RAPIDS cuML shared library \{cuda.ml\} was linked
#' to.
#'
#' @return The major version of the RAPIDS cuML shared library \{cuda.ml\} was
#' linked to in a character vector, or \code{NA_character_} if \{cuda.ml\} was not
#' linked to any version of RAPIDS cuML.
#'
#' @examples
#'
#' library(cuda.ml)
#'
#' print(cuML_major_version())
#' @export
cuML_major_version <- .cuML_major_version

#' Get the minor version of the RAPIDS cuML shared library \{cuda.ml\} was linked
#' to.
#'
#' @return The minor version of the RAPIDS cuML shared library \{cuda.ml\} was
#' linked to in a character vector, or \code{NA_character_} if \{cuda.ml\} was not
#' linked to any version of RAPIDS cuML.
#'
#' @examples
#'
#' library(cuda.ml)
#'
#' print(cuML_minor_version())
#' @export
cuML_minor_version <- .cuML_minor_version

check_libcuml_path <- function(path) {
  cuml_headers_dir <- file.path(path, "include", "cuml")
  dir.exists(cuml_headers_dir)
}

get_cuml_prefix <- function() {
  cuml_prefix <- Sys.getenv("CUML_PREFIX", unset = NA_character_)
  if (is.na(cuml_prefix)) {
    # Try the 'CUDA_PATH' env variable if it is present.
    cuml_prefix <- Sys.getenv("CUDA_PATH", unset = NA_character_)
  }
  if (is.na(cuml_prefix)) {
    cuml_prefix <- "/usr"
    if (check_libcuml_path(cuml_prefix)) {
      warning2(
        "'CUML_PREFIX' env variable is missing",
        "will boldly assume it is '/usr' !"
      )
      return(cuml_prefix)
    } else {

      # devtools::load_all() might run the config script from the `src` directory.
      cuml_prefix <- file.path(pkg_root(), "libcuml")
      if (check_libcuml_path(cuml_prefix)) {
        return(cuml_prefix)
      }

      # We will download a pre-built copy of `libcuml`
      return(NA_character_)
    }
  }

  return(cuml_prefix)
}

has_libcuml <- function() {
  # this is here to make sure we only proceed to automatically downloading if we
  # find a compatible nvcc version.
  find_nvcc()

  cuml_prefix <- get_cuml_prefix()
  if (is.na(cuml_prefix)) {
    # Skip subsequent checks if we are downloading a pre-built copy of `libcuml`
    TRUE
  } else {
    cuml_headers_dir <- file.path(cuml_prefix, "include", "cuml")

    if (!dir.exists(cuml_headers_dir)) {
      warning2(
        paste0(cuml_headers_dir, " does not exist or is not a directory!"),
        "",
        "{cuda.ml} requires a valid RAPIDS installation.",
        "Please follow https://rapids.ai/start.html to install RAPIDS first"
      )
      warning2(
        "{cuda.ml} must be installed from an environment containing a valid",
        "CUML_PREFIX env variable such that \"${CUML_PREFIX}/include/cuml\"",
        "is the directory of RAPIDS cuML header files and \"${CUML_PREFIX}/lib\"",
        "is the directory of RAPIDS cuML shared library files.)."
      )
      FALSE
    } else {
      TRUE
    }
  }
}

download_libcuml <- function(cuml_version = Sys.getenv("CUML_VERSION", unset = "21.08")) {
  wd <- getwd()
  on.exit(setwd(wd))
  setwd(pkg_root())

  if (Sys.getenv("DOWNLOAD_CUML", unset = 1) == 0) {
    stop2("No `libcuml` installation has been found and downloading has been prevented by `CUML_NO_DOWNLOAD`.")
  }

  old_timeout <- getOption("timeout")
  options(timeout = 1000)
  on.exit(options(timeout = old_timeout), add = TRUE)

  tmp <- tempfile(fileext = ".zip")
  cuda_version <- as.character(find_nvcc()$version$major)

  url <- Sys.getenv("CUML_URL")
  if (!nzchar(url)) {
    url <- libcuml_versions[[cuml_version]][[cuda_version]]
  }

  download.file(url, tmp)
  unzip(tmp, exdir = ".")

  zip_file_name <- basename(url)
  dir_name <- gsub("\\.zip$", "", zip_file_name)
  file.rename(file.path(".", dir_name), file.path(".", "libcuml"))
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

.agglomerative_clustering <- function(x, pairwise_conn, metric, n_neighbors, n_clusters) {
    .Call(`_cuda_ml_agglomerative_clustering`, x, pairwise_conn, metric, n_neighbors, n_clusters)
}

.cd_fit <- function(x, y, fit_intercept, normalize_input, epochs, loss, alpha, l1_ratio, shuffle, tol) {
    .Call(`_cuda_ml_cd_fit`, x, y, fit_intercept, normalize_input, epochs, loss, alpha, l1_ratio, shuffle, tol)
}

.has_cuML <- function() {
    .Call(`_cuda_ml_has_cuML`)
}

.cuML_major_version <- function() {
    .Call(`_cuda_ml_cuML_major_version`)
}

.cuML_minor_version <- function() {
    .Call(`_cuda_ml_cuML_minor_version`)
}

.dbscan <- function(x, min_pts, eps, max_bytes_per_batch, verbosity) {
    .Call(`_cuda_ml_dbscan`, x, min_pts, eps, max_bytes_per_batch, verbosity)
}

.fil_enabled <- function() {
    .Call(`_cuda_ml_fil_enabled`)
}

.fil_load_model <- function(model_type, filename, algo, classification, threshold, storage_type, blocks_per_sm, threads_per_tree, n_items) {
    .Call(`_cuda_ml_fil_load_model`, model_type, filename, algo, classification, threshold, storage_type, blocks_per_sm, threads_per_tree, n_items)
}

.fil_get_num_classes <- function(model) {
    .Call(`_cuda_ml_fil_get_num_classes`, model)
}

.fil_predict <- function(model, x, output_class_probabilities) {
    .Call(`_cuda_ml_fil_predict`, model, x, output_class_probabilities)
}

.kmeans <- function(x, k, max_iters, tol, init_method, centroids, seed, verbosity) {
    .Call(`_cuda_ml_kmeans`, x, k, max_iters, tol, init_method, centroids, seed, verbosity)
}

.knn_classifier_fit <- function(x, y, algo, metric, p, algo_params) {
    .Call(`_cuda_ml_knn_classifier_fit`, x, y, algo, metric, p, algo_params)
}

.knn_classifier_predict <- function(model, x, n_neighbors) {
    .Call(`_cuda_ml_knn_classifier_predict`, model, x, n_neighbors)
}

.knn_classifier_predict_probabilities <- function(model, x, n_neighbors) {
    .Call(`_cuda_ml_knn_classifier_predict_probabilities`, model, x, n_neighbors)
}

.knn_regressor_fit <- function(x, y, algo, metric, p, algo_params) {
    .Call(`_cuda_ml_knn_regressor_fit`, x, y, algo, metric, p, algo_params)
}

.knn_regressor_predict <- function(model, x, n_neighbors) {
    .Call(`_cuda_ml_knn_regressor_predict`, model, x, n_neighbors)
}

.lm_predict <- function(input, coef, intercept) {
    .Call(`_cuda_ml_lm_predict`, input, coef, intercept)
}

.ols_fit <- function(x, y, fit_intercept, normalize_input, algo) {
    .Call(`_cuda_ml_ols_fit`, x, y, fit_intercept, normalize_input, algo)
}

.pca_fit_transform <- function(x, tol, n_iters, verbosity, n_components, algo, whiten, transform_input) {
    .Call(`_cuda_ml_pca_fit_transform`, x, tol, n_iters, verbosity, n_components, algo, whiten, transform_input)
}

.pca_inverse_transform <- function(model, x) {
    .Call(`_cuda_ml_pca_inverse_transform`, model, x)
}

.pca_get_state <- function(model) {
    .Call(`_cuda_ml_pca_get_state`, model)
}

.pca_set_state <- function(model_state) {
    .Call(`_cuda_ml_pca_set_state`, model_state)
}

.qn_fit <- function(X, y, n_classes, loss_type, fit_intercept, l1, l2, max_iters, tol, delta, linesearch_max_iters, lbfgs_memory, sample_weight) {
    .Call(`_cuda_ml_qn_fit`, X, y, n_classes, loss_type, fit_intercept, l1, l2, max_iters, tol, delta, linesearch_max_iters, lbfgs_memory, sample_weight)
}

.qn_predict <- function(X, n_classes, coefs, loss_type, fit_intercept) {
    .Call(`_cuda_ml_qn_predict`, X, n_classes, coefs, loss_type, fit_intercept)
}

.rf_classifier_fit <- function(input, labels, n_trees, bootstrap, max_samples, n_streams, max_depth, max_leaves, max_features, n_bins, min_samples_leaf, min_samples_split, split_criterion, min_impurity_decrease, max_batch_size, verbosity) {
    .Call(`_cuda_ml_rf_classifier_fit`, input, labels, n_trees, bootstrap, max_samples, n_streams, max_depth, max_leaves, max_features, n_bins, min_samples_leaf, min_samples_split, split_criterion, min_impurity_decrease, max_batch_size, verbosity)
}

.rf_classifier_predict <- function(model_xptr, input, verbosity) {
    .Call(`_cuda_ml_rf_classifier_predict`, model_xptr, input, verbosity)
}

.rf_classifier_predict_class_probabilities <- function(model_xptr, input) {
    .Call(`_cuda_ml_rf_classifier_predict_class_probabilities`, model_xptr, input)
}

.rf_classifier_get_state <- function(model) {
    .Call(`_cuda_ml_rf_classifier_get_state`, model)
}

.rf_classifier_set_state <- function(state) {
    .Call(`_cuda_ml_rf_classifier_set_state`, state)
}

.rf_regressor_fit <- function(input, responses, n_trees, bootstrap, max_samples, n_streams, max_depth, max_leaves, max_features, n_bins, min_samples_leaf, min_samples_split, split_criterion, min_impurity_decrease, max_batch_size, verbosity) {
    .Call(`_cuda_ml_rf_regressor_fit`, input, responses, n_trees, bootstrap, max_samples, n_streams, max_depth, max_leaves, max_features, n_bins, min_samples_leaf, min_samples_split, split_criterion, min_impurity_decrease, max_batch_size, verbosity)
}

.rf_regressor_predict <- function(model_xptr, input, verbosity) {
    .Call(`_cuda_ml_rf_regressor_predict`, model_xptr, input, verbosity)
}

.rf_regressor_get_state <- function(model) {
    .Call(`_cuda_ml_rf_regressor_get_state`, model)
}

.rf_regressor_set_state <- function(state) {
    .Call(`_cuda_ml_rf_regressor_set_state`, state)
}

.rproj_johnson_lindenstrauss_min_dim <- function(n_samples, eps) {
    .Call(`_cuda_ml_rproj_johnson_lindenstrauss_min_dim`, n_samples, eps)
}

.rproj_fit <- function(n_samples, n_features, n_components, eps, gaussian_method, density, random_state) {
    .Call(`_cuda_ml_rproj_fit`, n_samples, n_features, n_components, eps, gaussian_method, density, random_state)
}

.rproj_transform <- function(rproj_ctx_xptr, input) {
    .Call(`_cuda_ml_rproj_transform`, rproj_ctx_xptr, input)
}

.rproj_get_state <- function(model) {
    .Call(`_cuda_ml_rproj_get_state`, model)
}

.rproj_set_state <- function(model_state) {
    .Call(`_cuda_ml_rproj_set_state`, model_state)
}

.ridge_fit <- function(x, y, fit_intercept, normalize_input, alpha, algo) {
    .Call(`_cuda_ml_ridge_fit`, x, y, fit_intercept, normalize_input, alpha, algo)
}

.sgd_fit <- function(x, y, fit_intercept, batch_size, epochs, lr_type, eta0, power_t, loss, penalty, alpha, l1_ratio, shuffle, tol, n_iter_no_change) {
    .Call(`_cuda_ml_sgd_fit`, x, y, fit_intercept, batch_size, epochs, lr_type, eta0, power_t, loss, penalty, alpha, l1_ratio, shuffle, tol, n_iter_no_change)
}

.svc_fit <- function(input, labels, cost, kernel, gamma, coef0, degree, tol, max_iter, nochange_steps, cache_size, sample_weights, verbosity) {
    .Call(`_cuda_ml_svc_fit`, input, labels, cost, kernel, gamma, coef0, degree, tol, max_iter, nochange_steps, cache_size, sample_weights, verbosity)
}

.svc_predict <- function(model_xptr, input, predict_class) {
    .Call(`_cuda_ml_svc_predict`, model_xptr, input, predict_class)
}

.svc_get_state <- function(model) {
    .Call(`_cuda_ml_svc_get_state`, model)
}

.svc_set_state <- function(state) {
    .Call(`_cuda_ml_svc_set_state`, state)
}

.svr_fit <- function(X, y, cost, kernel, gamma, coef0, degree, tol, max_iter, nochange_steps, cache_size, epsilon, sample_weights, verbosity) {
    .Call(`_cuda_ml_svr_fit`, X, y, cost, kernel, gamma, coef0, degree, tol, max_iter, nochange_steps, cache_size, epsilon, sample_weights, verbosity)
}

.svr_predict <- function(svr_xptr, X) {
    .Call(`_cuda_ml_svr_predict`, svr_xptr, X)
}

.svr_get_state <- function(model) {
    .Call(`_cuda_ml_svr_get_state`, model)
}

.svr_set_state <- function(state) {
    .Call(`_cuda_ml_svr_set_state`, state)
}

.tsne_fit <- function(x, dim, n_neighbors, theta, epssq, perplexity, perplexity_max_iter, perplexity_tol, early_exaggeration, late_exaggeration, exaggeration_iter, min_gain, pre_learning_rate, post_learning_rate, max_iter, min_grad_norm, pre_momentum, post_momentum, random_state, verbosity, initialize_embeddings, square_distances, algo) {
    .Call(`_cuda_ml_tsne_fit`, x, dim, n_neighbors, theta, epssq, perplexity, perplexity_max_iter, perplexity_tol, early_exaggeration, late_exaggeration, exaggeration_iter, min_gain, pre_learning_rate, post_learning_rate, max_iter, min_grad_norm, pre_momentum, post_momentum, random_state, verbosity, initialize_embeddings, square_distances, algo)
}

.tsvd_fit_transform <- function(x, tol, n_iters, verbosity, n_components, algo, transform_input) {
    .Call(`_cuda_ml_tsvd_fit_transform`, x, tol, n_iters, verbosity, n_components, algo, transform_input)
}

.tsvd_transform <- function(model, x) {
    .Call(`_cuda_ml_tsvd_transform`, model, x)
}

.tsvd_inverse_transform <- function(model, x) {
    .Call(`_cuda_ml_tsvd_inverse_transform`, model, x)
}

.umap_fit <- function(x, y, n_neighbors, n_components, n_epochs, learning_rate, min_dist, spread, set_op_mix_ratio, local_connectivity, repulsion_strength, negative_sample_rate, transform_queue_size, verbosity, a, b, init, target_n_neighbors, target_metric, target_weight, random_state, deterministic) {
    .Call(`_cuda_ml_umap_fit`, x, y, n_neighbors, n_components, n_epochs, learning_rate, min_dist, spread, set_op_mix_ratio, local_connectivity, repulsion_strength, negative_sample_rate, transform_queue_size, verbosity, a, b, init, target_n_neighbors, target_metric, target_weight, random_state, deterministic)
}

.umap_transform <- function(model, x) {
    .Call(`_cuda_ml_umap_transform`, model, x)
}

.umap_get_state <- function(model) {
    .Call(`_cuda_ml_umap_get_state`, model)
}

.umap_set_state <- function(state) {
    .Call(`_cuda_ml_umap_set_state`, state)
}

#' Train a random forest model.
#'
#' Train a random forest model for classification or regression tasks.
#'
#' @template supervised-model-inputs
#' @template supervised-model-output
#' @template ellipsis-unused
#' @template cuML-log-level
#' @param mtry The number of predictors that will be randomly sampled at each
#'   split when creating the tree models. Default: the square root of the total
#'   number of predictors.
#' @param trees An integer for the number of trees contained in the ensemble.
#'   Default: 100L.
#' @param min_n An integer for the minimum number of data points in a node that
#'   are required for the node to be split further. Default: 2L.
#' @param bootstrap Whether to perform bootstrap.
#'   If TRUE, each tree in the forest is built on a bootstrapped sample with
#'   replacement.
#'   If FALSE, the whole dataset is used to build each tree.
#' @param max_depth Maximum tree depth. Default: 16L.
#' @param max_leaves Maximum leaf nodes per tree. Soft constraint. Default: Inf
#'   (unlimited).
#' @param max_predictors_per_note_split Number of predictor to consider per node
#'   split. Default: square root of the total number predictors.
#' @param n_bins Number of bins used by the split algorithm. Default: 128L.
#' @param min_samples_leaf The minimum number of data points in each leaf node.
#'   Default: 1L.
#' @param split_criterion The criterion used to split nodes, can be "gini" or
#'   "entropy" for classifications, and "mse" or "mae" for regressions.
#'   Default: "gini" for classification; "mse" for regression.
#' @param min_impurity_decrease Minimum decrease in impurity requried for node
#'   to be spilt. Default: 0.
#' @param max_batch_size Maximum number of nodes that can be processed in a
#'   given batch. Default: 128L.
#' @param n_streams Number of CUDA streams to use for building trees.
#'   Default: 8L.
#'
#' @return A random forest classifier / regressor object that can be used with
#'   the 'predict' S3 generic to make predictions on new data points.
#'
#' @examples
#' library(cuda.ml)
#'
#' # Classification
#'
#' model <- cuda_ml_rand_forest(
#'   formula = Species ~ .,
#'   data = iris,
#'   trees = 100
#' )
#'
#' predictions <- predict(model, iris[names(iris) != "Species"])
#'
#' # Regression
#'
#' model <- cuda_ml_rand_forest(
#'   formula = mpg ~ .,
#'   data = mtcars,
#'   trees = 100
#' )
#'
#' predictions <- predict(model, mtcars[names(mtcars) != "mpg"])
#' @importFrom ellipsis check_dots_used
#' @export
cuda_ml_rand_forest <- function(x, ...) {
  check_dots_used()
  UseMethod("cuda_ml_rand_forest")
}

#' @rdname cuda_ml_rand_forest
#' @export
cuda_ml_rand_forest.default <- function(x, ...) {
  report_undefined_fn("cuda_ml_rand_forest", x)
}

#' @rdname cuda_ml_rand_forest
#' @export
cuda_ml_rand_forest.data.frame <- function(x, y, mtry = NULL, trees = NULL,
                                           min_n = 2L, bootstrap = TRUE,
                                           max_depth = 16L, max_leaves = Inf,
                                           max_predictors_per_note_split = NULL,
                                           n_bins = 128L, min_samples_leaf = 1L,
                                           split_criterion = NULL,
                                           min_impurity_decrease = 0,
                                           max_batch_size = 128L, n_streams = 8L,
                                           cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace"),
                                           ...) {
  processed <- hardhat::mold(x, y)

  cuda_ml_rand_forest_bridge(
    processed = processed,
    mtry = mtry,
    trees = trees,
    min_n = min_n,
    bootstrap = bootstrap,
    max_depth = max_depth,
    max_leaves = max_leaves,
    max_predictors_per_note_split = max_predictors_per_note_split,
    n_bins = n_bins,
    min_samples_leaf = min_samples_leaf,
    split_criterion = split_criterion,
    min_impurity_decrease = min_impurity_decrease,
    max_batch_size = max_batch_size,
    n_streams = n_streams,
    cuML_log_level = cuML_log_level
  )
}

#' @rdname cuda_ml_rand_forest
#' @export
cuda_ml_rand_forest.matrix <- function(x, y, mtry = NULL, trees = NULL, min_n = 2L,
                                       bootstrap = TRUE, max_depth = 16L,
                                       max_leaves = Inf,
                                       max_predictors_per_note_split = NULL,
                                       n_bins = 128L, min_samples_leaf = 1L,
                                       split_criterion = NULL,
                                       min_impurity_decrease = 0,
                                       max_batch_size = 128L, n_streams = 8L,
                                       cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace"),
                                       ...) {
  processed <- hardhat::mold(x, y)

  cuda_ml_rand_forest_bridge(
    processed = processed,
    mtry = mtry,
    trees = trees,
    min_n = min_n,
    bootstrap = bootstrap,
    max_depth = max_depth,
    max_leaves = max_leaves,
    max_predictors_per_note_split = max_predictors_per_note_split,
    n_bins = n_bins,
    min_samples_leaf = min_samples_leaf,
    split_criterion = split_criterion,
    min_impurity_decrease = min_impurity_decrease,
    max_batch_size = max_batch_size,
    n_streams = n_streams,
    cuML_log_level = cuML_log_level
  )
}

#' @rdname cuda_ml_rand_forest
#' @export
cuda_ml_rand_forest.formula <- function(formula, data, mtry = NULL, trees = NULL,
                                        min_n = 2L, bootstrap = TRUE,
                                        max_depth = 16L, max_leaves = Inf,
                                        max_predictors_per_note_split = NULL,
                                        n_bins = 128L, min_samples_leaf = 1L,
                                        split_criterion = NULL,
                                        min_impurity_decrease = 0,
                                        max_batch_size = 128L, n_streams = 8L,
                                        cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace"),
                                        ...) {
  processed <- hardhat::mold(formula, data)

  cuda_ml_rand_forest_bridge(
    processed = processed,
    mtry = mtry,
    trees = trees,
    min_n = min_n,
    bootstrap = bootstrap,
    max_depth = max_depth,
    max_leaves = max_leaves,
    max_predictors_per_note_split = max_predictors_per_note_split,
    n_bins = n_bins,
    min_samples_leaf = min_samples_leaf,
    split_criterion = split_criterion,
    min_impurity_decrease = min_impurity_decrease,
    max_batch_size = max_batch_size,
    n_streams = n_streams,
    cuML_log_level = cuML_log_level
  )
}

#' @rdname cuda_ml_rand_forest
#' @export
cuda_ml_rand_forest.recipe <- function(x, data, mtry = NULL, trees = NULL,
                                       min_n = 2L, bootstrap = TRUE,
                                       max_depth = 16L, max_leaves = Inf,
                                       max_predictors_per_note_split = NULL,
                                       n_bins = 128L, min_samples_leaf = 1L,
                                       split_criterion = NULL,
                                       min_impurity_decrease = 0,
                                       max_batch_size = 128L, n_streams = 8L,
                                       cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace"),
                                       ...) {
  processed <- hardhat::mold(x, data)

  cuda_ml_rand_forest_bridge(
    processed = processed,
    mtry = mtry,
    trees = trees,
    min_n = min_n,
    bootstrap = bootstrap,
    max_depth = max_depth,
    max_leaves = max_leaves,
    max_predictors_per_note_split = max_predictors_per_note_split,
    n_bins = n_bins,
    min_samples_leaf = min_samples_leaf,
    split_criterion = split_criterion,
    min_impurity_decrease = min_impurity_decrease,
    max_batch_size = max_batch_size,
    n_streams = n_streams,
    cuML_log_level = cuML_log_level
  )
}

cuda_ml_rand_forest_bridge <- function(processed, mtry, trees, min_n, bootstrap,
                                       max_depth, max_leaves,
                                       max_predictors_per_note_split, n_bins,
                                       min_samples_leaf, split_criterion,
                                       min_impurity_decrease, max_batch_size,
                                       n_streams, cuML_log_level) {
  hardhat::validate_predictors_are_numeric(processed$predictors)
  hardhat::validate_outcomes_are_univariate(processed$outcomes)
  x <- as.matrix(processed$predictors)
  y <- processed$outcomes[[1]]
  classification <- is.factor(y)
  if (identical(max_leaves, Inf)) {
    max_leaves <- -1L
  }

  # Default value for 'split_criterion' depends on whether a classification or a
  # regression task is being performed.
  split_criterion <- decision_tree_match_split_criterion(
    split_criterion,
    classification
  )
  cuML_log_level <- match_cuML_log_level(cuML_log_level)

  rand_forest_fit_impl <- ifelse(
    classification,
    cuda_ml_rand_forest_impl_classification,
    cuda_ml_rand_forest_impl_regression
  )

  rand_forest_fit_impl(
    processed = processed,
    mtry = mtry,
    trees = trees,
    min_n = min_n,
    bootstrap = bootstrap,
    max_depth = max_depth,
    max_leaves = max_leaves,
    max_predictors_per_note_split = max_predictors_per_note_split,
    n_bins = n_bins,
    min_samples_leaf = min_samples_leaf,
    split_criterion = split_criterion,
    min_impurity_decrease = min_impurity_decrease,
    max_batch_size = max_batch_size,
    n_streams = n_streams,
    cuML_log_level = cuML_log_level
  )
}

cuda_ml_rand_forest_impl_classification <- function(processed, mtry, trees, min_n,
                                                    bootstrap, max_depth,
                                                    max_leaves,
                                                    max_predictors_per_note_split,
                                                    n_bins, min_samples_leaf,
                                                    split_criterion,
                                                    min_impurity_decrease,
                                                    max_batch_size, n_streams,
                                                    cuML_log_level) {
  x <- as.matrix(processed$predictors)
  y <- processed$outcomes[[1]]

  model_xptr <- .rf_classifier_fit(
    input = as.matrix(x),
    labels = as.integer(y),
    n_trees = as.integer(trees),
    bootstrap = as.logical(bootstrap),
    max_samples = as.numeric(mtry %||% sqrt(ncol(x))) / ncol(x),
    n_streams = as.integer(n_streams),
    max_depth = as.integer(max_depth),
    max_leaves = as.integer(max_leaves),
    max_features = as.numeric(max_predictors_per_note_split %||% sqrt(ncol(x))) / ncol(x),
    n_bins = as.integer(n_bins),
    min_samples_leaf = as.integer(min_samples_leaf),
    min_samples_split = as.integer(min_n %||% 2L),
    split_criterion = split_criterion,
    min_impurity_decrease = as.numeric(min_impurity_decrease),
    max_batch_size = as.integer(max_batch_size),
    verbosity = cuML_log_level
  )

  new_model(
    cls = "cuda_ml_rand_forest",
    mode = "classification",
    xptr = model_xptr,
    blueprint = processed$blueprint
  )
}

cuda_ml_rand_forest_impl_regression <- function(processed, mtry, trees, min_n,
                                                bootstrap, max_depth, max_leaves,
                                                max_predictors_per_note_split,
                                                n_bins, min_samples_leaf,
                                                split_criterion,
                                                min_impurity_decrease,
                                                max_batch_size, n_streams,
                                                cuML_log_level) {
  x <- as.matrix(processed$predictors)
  y <- processed$outcomes[[1]]

  model_xptr <- .rf_regressor_fit(
    input = as.matrix(x),
    responses = as.numeric(y),
    n_trees = as.integer(trees),
    bootstrap = as.logical(bootstrap),
    max_samples = as.numeric(mtry %||% sqrt(ncol(x))) / ncol(x),
    n_streams = as.integer(n_streams),
    max_depth = as.integer(max_depth),
    max_leaves = as.integer(max_leaves),
    max_features = as.numeric(max_predictors_per_note_split %||% sqrt(ncol(x))) / ncol(x),
    n_bins = as.integer(n_bins),
    min_samples_leaf = as.integer(min_samples_leaf),
    min_samples_split = as.integer(min_n %||% 2L),
    split_criterion = split_criterion,
    min_impurity_decrease = as.numeric(min_impurity_decrease),
    max_batch_size = as.integer(max_batch_size),
    verbosity = cuML_log_level
  )
  new_model(
    cls = "cuda_ml_rand_forest",
    mode = "regression",
    xptr = model_xptr,
    blueprint = processed$blueprint
  )
}

#' @export
cuda_ml_get_state.cuda_ml_rand_forest <- function(model) {
  get_state_impl <- switch(model$mode,
    classification = .rf_classifier_get_state,
    regression = .rf_regressor_get_state
  )

  model_state <- list(
    mode = model$mode,
    rf = get_state_impl(model$xptr),
    blueprint = model$blueprint
  )

  new_model_state(model_state, "cuda_ml_rand_forest_model_state")
}

#' @export
cuda_ml_set_state.cuda_ml_rand_forest_model_state <- function(model_state) {
  set_state_impl <- switch(model_state$mode,
    classification = .rf_classifier_set_state,
    regression = .rf_regressor_set_state
  )

  new_model(
    cls = "cuda_ml_rand_forest",
    mode = model_state$mode,
    xptr = set_state_impl(model_state$rf),
    blueprint = model_state$blueprint
  )
}

#' Make predictions on new data points.
#'
#' Make predictions on new data points using a CuML random forest model.
#'
#' @template predict
#' @template output-class-probabilities
#' @template cuML-log-level
#'
#' @importFrom ellipsis check_dots_used
#' @export
predict.cuda_ml_rand_forest <- function(object, x,
                                        output_class_probabilities = NULL,
                                        cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace"),
                                        ...) {
  check_dots_used()

  processed <- hardhat::forge(x, object$blueprint)

  predict_cuda_ml_rand_forest_bridge(
    model = object,
    processed = processed,
    output_class_probabilities = output_class_probabilities,
    cuML_log_level = cuML_log_level
  )
}

predict_cuda_ml_rand_forest_bridge <- function(model,
                                               processed,
                                               output_class_probabilities,
                                               cuML_log_level) {
  cuML_log_level <- match_cuML_log_level(cuML_log_level)

  out <- switch(model$mode,
    classification = {
      predict_cuda_ml_rand_forest_classification_impl(
        model = model,
        processed = processed,
        output_class_probabilities = output_class_probabilities %||% FALSE,
        cuML_log_level = cuML_log_level
      )
    },
    regression = {
      if (!is.null(output_class_probabilities)) {
        stop("'output_class_probabilities' is not applicable for regression tasks!")
      }

      predict_cuda_ml_rand_forest_regression_impl(
        model = model,
        processed = processed,
        cuML_log_level = cuML_log_level
      )
    }
  )
  hardhat::validate_prediction_size(out, processed$predictors)

  out
}

predict_cuda_ml_rand_forest_classification_impl <- function(model,
                                                            processed,
                                                            output_class_probabilities,
                                                            cuML_log_level) {
  if (output_class_probabilities) {
    if (as.integer(cuML_major_version()) == 21 &&
      as.integer(cuML_minor_version()) < 8) {
      stop(
        "Class probabilities output for random forest classifier is only ",
        "supported by RAPIDS cuML 21.08 or above. Current version of ",
        "RAPIDS cuML linked with {cuda.ml} is v",
        paste0(cuML_major_version(), ".", cuML_minor_version()), "."
      )
    }

    preds <- .rf_classifier_predict_class_probabilities(
      model_xptr = model$xptr,
      input = as.matrix(processed$predictors)
    )

    postprocess_class_probabilities(preds, model)
  } else {
    preds <- .rf_classifier_predict(
      model_xptr = model$xptr,
      input = as.matrix(processed$predictors),
      verbosity = cuML_log_level
    )

    postprocess_classification_results(preds, model)
  }
}

predict_cuda_ml_rand_forest_regression_impl <- function(model, processed,
                                                        cuML_log_level) {
  preds <- .rf_regressor_predict(
    model_xptr = model$xptr,
    input = as.matrix(processed$predictors),
    verbosity = cuML_log_level
  )

  postprocess_regression_results(preds)
}

# register the CuML-based rand_forest model for parsnip
register_rand_forest_model <- function(pkgname) {
  for (mode in c("classification", "regression")) {
    parsnip::set_model_engine(
      model = "rand_forest", mode = mode, eng = pkgname
    )
  }

  parsnip::set_dependency(model = "rand_forest", eng = pkgname, pkg = pkgname)

  parsnip::set_model_arg(
    model = "rand_forest",
    eng = pkgname,
    parsnip = "mtry",
    original = "mtry",
    func = list(pkg = "dials", fun = "mtry"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "rand_forest",
    eng = pkgname,
    parsnip = "trees",
    original = "trees",
    func = list(pkg = "dials", fun = "trees"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "rand_forest",
    eng = pkgname,
    parsnip = "min_n",
    original = "min_n",
    func = list(pkg = "dials", fun = "min_n"),
    has_submodel = FALSE
  )

  for (mode in c("classification", "regression")) {
    parsnip::set_fit(
      model = "rand_forest",
      eng = pkgname,
      mode = mode,
      value = list(
        interface = "formula",
        protect = c("formula", "data"),
        func = c(pkg = pkgname, fun = "cuda_ml_rand_forest"),
        defaults = list(
          bootstrap = TRUE,
          max_depth = 16L,
          max_leaves = Inf,
          max_predictors_per_note_split = NULL,
          n_bins = 128L,
          min_samples_leaf = 1L,
          split_criterion = NULL,
          min_impurity_decrease = 0,
          max_batch_size = 128L,
          n_streams = 8L,
          cuML_log_level = "off"
        )
      )
    )

    parsnip::set_encoding(
      model = "rand_forest",
      eng = pkgname,
      mode = mode,
      options = list(
        predictor_indicators = "none",
        compute_intercept = FALSE,
        remove_intercept = FALSE,
        allow_sparse_x = TRUE
      )
    )
  }

  for (type in c("class", "prob")) {
    parsnip::set_pred(
      model = "rand_forest",
      eng = pkgname,
      mode = "classification",
      type = type,
      value = list(
        pre = NULL,
        post = NULL,
        func = c(fun = "predict"),
        args = list(
          quote(object$fit),
          quote(new_data),
          identical(type, "prob") # output_class_probabilities
        )
      )
    )
  }

  parsnip::set_pred(
    model = "rand_forest",
    eng = pkgname,
    mode = "regression",
    type = "numeric",
    value = list(
      pre = NULL,
      post = NULL,
      func = c(fun = "predict"),
      args = list(
        quote(object$fit),
        quote(new_data)
      )
    )
  )
}

context("Random Forest")

test_that("random forest classifier works as expected", {
  cuda_ml_rf_model <- cuda_ml_rand_forest(
    formula = Species ~ ., data = iris, trees = 200, bootstrap = FALSE
  )
  sklearn_rf_model <- sklearn$ensemble$RandomForestClassifier(
    n_estimators = 200L, bootstrap = FALSE
  )
  sklearn_rf_model$fit(
    X = as.matrix(iris[which(names(iris) != "Species")]),
    y = as.integer(iris$Species)
  )

  cuda_ml_preds <- predict(
    cuda_ml_rf_model, iris[which(names(iris) != "Species")]
  )
  sklearn_preds <- sklearn_rf_model$predict(
    as.matrix(iris[which(names(iris) != "Species")])
  )

  expect_equal(
    as.integer(cuda_ml_preds$.pred_class), as.integer(sklearn_preds)
  )
})

test_that("random forest regressor works as expected", {
  cuda_ml_rf_model <- cuda_ml_rand_forest(
    formula = mpg ~ ., data = mtcars, trees = 100, bootstrap = FALSE
  )
  cuda_ml_preds <- predict(
    cuda_ml_rf_model, mtcars[which(names(mtcars) != "mpg")]
  )

  expect_equal(cuda_ml_preds$.pred, mtcars$mpg, tolerance = 0.2)
})

test_that("random forest classifier works as expected through parsnip", {
  require("parsnip")

  cuda_ml_rf_model <- rand_forest(trees = 200, mode = "classification") %>%
    set_engine("cuda.ml", bootstrap = FALSE) %>%
    fit(Species ~ ., data = iris)
  sklearn_rf_model <- sklearn$ensemble$RandomForestClassifier(
    n_estimators = 200L, bootstrap = FALSE
  )
  sklearn_rf_model$fit(
    X = as.matrix(iris[which(names(iris) != "Species")]),
    y = as.integer(iris$Species)
  )

  cuda_ml_preds <- predict(
    cuda_ml_rf_model, iris[which(names(iris) != "Species")]
  )
  sklearn_preds <- sklearn_rf_model$predict(
    as.matrix(iris[which(names(iris) != "Species")])
  )

  expect_equal(
    as.integer(cuda_ml_preds$.pred_class), as.integer(sklearn_preds)
  )
})

test_that("random forest regressor works as expected through parsnip", {
  require("parsnip")

  cuda_ml_rf_model <- rand_forest(trees = 200, mode = "regression") %>%
    set_engine("cuda.ml", bootstrap = FALSE) %>%
    fit(mpg ~ ., data = mtcars)
  cuda_ml_preds <- predict(
    cuda_ml_rf_model, mtcars[which(names(mtcars) != "mpg")]
  )

  expect_equal(cuda_ml_preds$.pred, mtcars$mpg, tolerance = 0.2)
})

knn_match_algo <- function(algo = c("brute", "ivfflat", "ivfpq", "ivfsq")) {
  algo <- match.arg(algo)

  switch(algo,
    brute = 0L,
    ivfflat = 1L,
    ivfpq = 2L,
    ivfsq = 3L
  )
}

knn_match_metric <- function(metric = c("euclidean", "l2", "l1", "cityblock", "taxicab", "manhattan", "braycurtis", "canberra", "minkowski", "lp", "chebyshev", "linf", "jensenshannon", "cosine", "correlation")) {
  metric <- match.arg(metric)

  switch(metric,
    euclidean = 1L,
    l2 = 1L,
    l1 = 3L,
    cityblock = 3L,
    taxicab = 3L,
    manhattan = 3L,
    braycurtis = 14L,
    canberra = 8L,
    minkowski = 9L,
    lp = 9L,
    chebyshev = 7L,
    linf = 7L,
    jensenshannon = 15L,
    cosine = 2L,
    correlation = 10L
  )
}

#' Build a specification for the "ivfflat" KNN query algorithm.
#'
#' Build a specification of the flat-inverted-file KNN query algorithm, with all
#' required parameters specified explicitly.
#'
#' @template knn-algo-common
#'
#' @return An object encapsulating all required parameters of the "ivfflat" KNN
#'   query algorithm.
#'
#' @export
cuda_ml_knn_algo_ivfflat <- function(nlist, nprobe) {
  list(
    type = 1L,
    params = list(
      nlist = as.integer(nlist),
      nprobe = as.integer(nprobe)
    )
  )
}

#' Build a specification for the "ivfpq" KNN query algorithm.
#'
#' Build a specification of the inverted-file-product-quantization KNN query
#' algorithm, with all required parameters specified explicitly.
#'
#' @template knn-algo-common
#' @template knn-algo-ivfpq
#'
#' @return An object encapsulating all required parameters of the "ivfpq" KNN
#'   query algorithm.
#'
#' @export
cuda_ml_knn_algo_ivfpq <- function(nlist, nprobe, m, n_bits,
                                   use_precomputed_tables = FALSE) {
  list(
    type = 2L,
    params = list(
      nlist = as.integer(nlist),
      nprobe = as.integer(nprobe),
      M = as.integer(m),
      usePrecomputedTables = as.logical(use_precomputed_tables)
    )
  )
}

#' Build a specification for the "ivfsq" KNN query algorithm.
#'
#' Build a specification of the inverted-file-scalar-quantization KNN query
#' algorithm, with all required parameters specified explicitly.
#'
#' @template knn-algo-common
#' @template knn-algo-ivfsq
#'
#' @return An object encapsulating all required parameters of the "ivfsq" KNN
#'   query algorithm.
#'
#' @export
cuda_ml_knn_algo_ivfsq <- function(nlist, nprobe,
                                   qtype = c("QT_8bit", "QT_4bit", "QT_8bit_uniform", "QT_4bit_uniform", "QT_fp16", "QT_8bit_direct", "QT_6bit"),
                                   encode_residual = FALSE) {
  list(
    type = 3L,
    params = list(
      nlist = as.integer(nlist),
      nprobe = as.integer(nprobe),
      qtype = match.arg(qtype),
      encodeResidual = as.logical(encode_residual)
    )
  )
}

#' Build a KNN model.
#'
#' Build a k-nearest-model for classification or regression tasks.
#'
#' @template supervised-model-inputs
#' @template supervised-model-output
#' @template ellipsis-unused
#' @param algo The query algorithm to use. Must be one of
#'   \{"brute", "ivfflat", "ivfpq", "ivfsq"\} or a KNN algorithm specification
#'   constructed using the \code{cuda_ml_knn_algo_*} family of functions.
#'   If the algorithm is specified by one of the \code{cuda_ml_knn_algo_*}
#'   functions, then values of all required parameters of the algorithm will
#'   need to be specified explicitly.
#'   If the algorithm is specified by a character vector, then parameters for
#'   the algorithm are generated automatically.
#'
#'   Descriptions of supported algorithms:
#'     - "brute": for brute-force, slow but produces exact results.
#'     - "ivfflat": for inverted file, divide the dataset in partitions
#'                  and perform search on relevant partitions only.
#'     - "ivfpq": for inverted file and product quantization (vectors
#'                are divided into sub-vectors, and each sub-vector is encoded
#'                using intermediary k-means clusterings to provide partial
#'                information).
#'     - "ivfsq": for inverted file and scalar quantization (vectors components
#'                are quantized into reduced binary representation allowing
#'                faster distances calculations).
#'
#'   Default: "brute".
#' @param metric Distance metric to use. Must be one of \{"euclidean", "l2",
#'   "l1", "cityblock", "taxicab", "manhattan", "braycurtis", "canberra",
#'   "minkowski", "lp", "chebyshev", "linf", "jensenshannon", "cosine",
#'   "correlation"\}.
#'   Default: "euclidean".
#' @param p Parameter for the Minkowski metric. If p = 1, then the metric is
#'   equivalent to manhattan distance (l1). If p = 2, the metric is equivalent
#'   to euclidean distance (l2).
#' @param neighbors Number of nearest neighbors to query. Default: 5L.
#'
#' @return A KNN model that can be used with the 'predict' S3 generic to make
#'   predictions on new data points.
#'   The model object contains the following:
#'     - "knn_index": a GPU pointer to the KNN index.
#'     - "algo": enum value of the algorithm being used for the KNN query.
#'     - "metric": enum value of the distance metric used in KNN computations.
#'     - "p": parameter for the Minkowski metric.
#'     - "n_samples": number of input data points.
#'     - "n_dims": dimension of each input data point.
#'
#' @examples
#'
#' library(cuda.ml)
#' library(MASS)
#' library(magrittr)
#' library(purrr)
#'
#' set.seed(0L)
#'
#' centers <- list(c(3, 3), c(-3, -3), c(-3, 3))
#'
#' gen_pts <- function(cluster_sz) {
#'   pts <- centers %>%
#'     map(~ mvrnorm(cluster_sz, mu = .x, Sigma = diag(2)))
#'
#'   rlang::exec(rbind, !!!pts) %>% as.matrix()
#' }
#'
#' gen_labels <- function(cluster_sz) {
#'   seq_along(centers) %>%
#'     sapply(function(x) rep(x, cluster_sz)) %>%
#'     factor()
#' }
#'
#' sample_cluster_sz <- 1000
#' sample_pts <- cbind(
#'   gen_pts(sample_cluster_sz) %>% as.data.frame(),
#'   label = gen_labels(sample_cluster_sz)
#' )
#'
#' model <- cuda_ml_knn(label ~ ., sample_pts, algo = "ivfflat", metric = "euclidean")
#'
#' test_cluster_sz <- 10
#' test_pts <- gen_pts(test_cluster_sz) %>% as.data.frame()
#'
#' predictions <- predict(model, test_pts)
#' print(predictions, n = 30)
#' @importFrom ellipsis check_dots_used
#' @export
cuda_ml_knn <- function(x, ...) {
  check_dots_used()
  UseMethod("cuda_ml_knn")
}

#' @rdname cuda_ml_knn
#' @export
cuda_ml_knn.default <- function(x, ...) {
  report_undefined_fn("cuda_ml_knn", x)
}

#' @rdname cuda_ml_knn
#' @export
cuda_ml_knn.data.frame <- function(x, y,
                                   algo = c("brute", "ivfflat", "ivfpq", "ivfsq"),
                                   metric = c("euclidean", "l2", "l1", "cityblock", "taxicab", "manhattan", "braycurtis", "canberra", "minkowski", "chebyshev", "jensenshannon", "cosine", "correlation"),
                                   p = 2.0,
                                   neighbors = 5L,
                                   ...) {
  processed <- hardhat::mold(x, y)

  cuda_ml_knn_bridge(
    processed = processed,
    algo = algo,
    metric = metric,
    p = p,
    neighbors = neighbors
  )
}

#' @rdname cuda_ml_knn
#' @export
cuda_ml_knn.matrix <- function(x, y,
                               algo = c("brute", "ivfflat", "ivfpq", "ivfsq"),
                               metric = c("euclidean", "l2", "l1", "cityblock", "taxicab", "manhattan", "braycurtis", "canberra", "minkowski", "chebyshev", "jensenshannon", "cosine", "correlation"),
                               p = 2.0,
                               neighbors = 5L,
                               ...) {
  processed <- hardhat::mold(x, y)

  cuda_ml_knn_bridge(
    processed = processed,
    algo = algo,
    metric = metric,
    p = p,
    neighbors = neighbors
  )
}

#' @rdname cuda_ml_knn
#' @export
cuda_ml_knn.formula <- function(formula, data,
                                algo = c("brute", "ivfflat", "ivfpq", "ivfsq"),
                                metric = c("euclidean", "l2", "l1", "cityblock", "taxicab", "manhattan", "braycurtis", "canberra", "minkowski", "chebyshev", "jensenshannon", "cosine", "correlation"),
                                p = 2.0,
                                neighbors = 5L,
                                ...) {
  processed <- hardhat::mold(formula, data)

  cuda_ml_knn_bridge(
    processed = processed,
    algo = algo,
    metric = metric,
    p = p,
    neighbors = neighbors
  )
}

#' @rdname cuda_ml_knn
#' @export
cuda_ml_knn.recipe <- function(x, data,
                               algo = c("brute", "ivfflat", "ivfpq", "ivfsq"),
                               metric = c("euclidean", "l2", "l1", "cityblock", "taxicab", "manhattan", "braycurtis", "canberra", "minkowski", "chebyshev", "jensenshannon", "cosine", "correlation"),
                               p = 2.0,
                               neighbors = 5L,
                               ...) {
  processed <- hardhat::mold(x, data)

  cuda_ml_knn_bridge(
    processed = processed,
    algo = algo,
    metric = metric,
    p = p,
    neighbors = neighbors
  )
}

cuda_ml_knn_bridge <- function(processed, algo, metric, p, neighbors) {
  hardhat::validate_predictors_are_numeric(processed$predictors)
  hardhat::validate_outcomes_are_univariate(processed$outcomes)
  x <- as.matrix(processed$predictors)
  y <- processed$outcomes[[1]]

  if (is.character(algo)) {
    algo_type <- knn_match_algo(algo)
    algo_params <- list()
  } else {
    algo_type <- algo$type
    algo_params <- algo$params
  }
  metric <- knn_match_metric(metric)

  if (is.factor(y)) {
    # classification
    prediction_mode <- "classification"
    model_xptr <- .knn_classifier_fit(
      x = x,
      y = as.integer(y),
      algo = algo_type,
      metric = metric,
      p = as.numeric(p),
      algo_params = algo_params
    )
  } else {
    prediction_mode <- "regression"
    model_xptr <- .knn_regressor_fit(
      x = x,
      y = as.numeric(y),
      algo = algo_type,
      metric = metric,
      p = as.numeric(p),
      algo_params = algo_params
    )
  }

  new_model(
    cls = "cuda_ml_knn",
    mode = prediction_mode,
    xptr = model_xptr,
    neighbors = as.integer(neighbors),
    blueprint = processed$blueprint
  )
}

#' Make predictions on new data points.
#'
#' Make predictions on new data points using a CuML KNN model.
#'
#' @template predict
#' @template output-class-probabilities
#'
#' @importFrom ellipsis check_dots_used
#' @export
predict.cuda_ml_knn <- function(object, x, output_class_probabilities = NULL, ...) {
  check_dots_used()

  processed <- hardhat::forge(x, object$blueprint)

  predict_cuda_ml_knn_bridge(
    model = object,
    processed = processed,
    output_class_probabilities = output_class_probabilities
  )
}

predict_cuda_ml_knn_bridge <- function(model, processed, output_class_probabilities) {
  out <- switch(model$mode,
    classification = {
      predict_cuda_ml_knn_classification_impl(
        model = model,
        processed = processed,
        output_class_probabilities = output_class_probabilities %||% FALSE
      )
    },
    regression = {
      if (!is.null(output_class_probabilities)) {
        stop("'output_class_probabilities' is not applicable for regression tasks!")
      }

      predict_cuda_ml_knn_regression_impl(
        model = model, processed = processed
      )
    }
  )
  hardhat::validate_prediction_size(out, processed$predictors)

  out
}

predict_cuda_ml_knn_classification_impl <- function(model, processed, output_class_probabilities) {
  if (output_class_probabilities) {
    preds <- .knn_classifier_predict_probabilities(
      model = model$xptr,
      x = as.matrix(processed$predictors),
      n_neighbors = model$neighbors
    )

    postprocess_class_probabilities(preds, model)
  } else {
    preds <- .knn_classifier_predict(
      model = model$xptr,
      x = as.matrix(processed$predictors),
      n_neighbors = model$neighbors
    )

    postprocess_classification_results(preds, model)
  }
}

predict_cuda_ml_knn_regression_impl <- function(model, processed) {
  preds <- .knn_regressor_predict(
    model = model$xptr,
    x = as.matrix(processed$predictors),
    n_neighbors = model$neighbors
  )

  postprocess_regression_results(preds)
}

# register the CuML-based knn model for parsnip
register_knn_model <- function(pkgname) {
  for (mode in c("classification", "regression")) {
    parsnip::set_model_engine(
      model = "nearest_neighbor", mode = mode, eng = pkgname
    )
  }

  parsnip::set_dependency(model = "nearest_neighbor", eng = pkgname, pkg = pkgname)

  parsnip::set_model_arg(
    model = "nearest_neighbor",
    eng = pkgname,
    parsnip = "neighbors",
    original = "neighbors",
    func = list(pkg = "dials", fun = "neighbors", range = c(1, 15)),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "nearest_neighbor",
    eng = pkgname,
    parsnip = "dist_power",
    original = "p",
    func = list(pkg = "dials", fun = "dist_power", range = c(1 / 10, 2)),
    has_submodel = FALSE
  )

  for (mode in c("classification", "regression")) {
    parsnip::set_fit(
      model = "nearest_neighbor",
      eng = pkgname,
      mode = mode,
      value = list(
        interface = "formula",
        protect = c("formula", "data"),
        func = c(pkg = pkgname, fun = "cuda_ml_knn"),
        defaults = list(algo = "ivfflat", metric = "euclidean")
      )
    )

    parsnip::set_encoding(
      model = "nearest_neighbor",
      eng = pkgname,
      mode = mode,
      options = list(
        predictor_indicators = "none",
        compute_intercept = FALSE,
        remove_intercept = FALSE,
        allow_sparse_x = TRUE
      )
    )
  }

  for (type in c("class", "prob")) {
    parsnip::set_pred(
      model = "nearest_neighbor",
      eng = pkgname,
      mode = "classification",
      type = type,
      value = list(
        pre = NULL,
        post = NULL,
        func = c(fun = "predict"),
        args = list(
          quote(object$fit),
          quote(new_data),
          identical(type, "prob") # output_class_probabilities
        )
      )
    )
  }

  parsnip::set_pred(
    model = "nearest_neighbor",
    eng = pkgname,
    mode = "regression",
    type = "numeric",
    value = list(
      pre = NULL,
      post = NULL,
      func = c(fun = "predict"),
      args = list(
        quote(object$fit),
        quote(new_data)
      )
    )
  )
}

logistic_reg_loss_type <- list(sigmoid = 0L, softmax = 2L)

logistic_reg_build_qn_params <- function(C,
                                         l1_ratio,
                                         penalty = c("l2", "l1", "elasticnet", "none")) {
  penalty <- match.arg(penalty)

  switch(penalty,
    "none" = {
      list(l1 = 0, l2 = 0)
    },
    "l1" = {
      list(l1 = 1 / C, l2 = 0)
    },
    "l2" = {
      list(l1 = 0, l2 = 1 / C)
    },
    "elasticnet" = {
      strength <- 1 / C
      if (is.null(l1_ratio) || !is.numeric(l1_ratio) || l1_ratio < 0 ||
          l1_ratio > 1) {
        stop(
          "`l1_ratio` must be non-NULL and be within [0, 1] for elasticnet ",
          "regularization."
        )
      }
      list(l1 = l1_ratio * strength, l2 = (1 - l1_ratio) * strength)
    }
  )
}

logistic_reg_validate_class_weight <- function(class_weight, processed) {
  if (identical(class_weight, "balanced")) {
    return()
  }
  lvls <- levels(processed$outcomes[[1]])
  n_lvls <- length(lvls)
  if (!setequal(names(class_weight), lvls) || length(class_weight) != n_lvls) {
    stop(
      "Expected `class_weight` to specify weights for the following possible ",
      "outcomes: {", paste(lvls, collapse = ", "), "}."
    )
  }

  for (v in class_weight) {
    if (!is.numeric(v) || v < 0 || !is.finite(v)) {
      stop(
        "Expected `class_weight` to be a numeric vector of finite, non-",
        "negative and non-NaN values."
      )
    }
  }
}

logistic_reg_validate_sample_weight <- function(sample_weight, processed) {
  n_samples <- length(processed$outcomes[[1]])
  if (!is.numeric(sample_weight) || length(sample_weight) != n_samples) {
    stop(
      "Expected `sample_weight` to be a numeric vector with length equal to ",
      "the number of training samples (", length(sample_weight), ")."
    )
  }
  if (any(sample_weight < 0) || any(!is.finite(sample_weight))) {
    stop(
      "Expected `sample_weight` to only contain finite, non-negative, and ",
      "non-NaN numeric values."
    )
  }
}

logistic_reg_build_sample_weight <- function(sample_weight,
                                             class_weight,
                                             processed) {
  outcomes <- processed$outcomes[[1]]
  n_samples <- length(outcomes)
  if (identical(class_weight, "balanced")) {
    class_weight <- n_samples / table(outcomes)
  }
  if (!is.null(class_weight)) {
    # update each sample weight value by multiplying it with its corresponding
    # class weight
    sample_weight <- sample_weight *
      as.numeric(class_weight[levels(outcomes)[outcomes]])
  }

  sample_weight
}

#' Train a logistic regression model.
#'
#' Train a logistic regression model using Quasi-Newton (QN) algorithms (i.e.,
#' Orthant-Wise Limited Memory Quasi-Newton (OWL-QN) if there is L1
#' regularization, Limited Memory BFGS (L-BFGS) otherwise).
#'
#' @template supervised-model-inputs
#' @template supervised-model-output
#' @template ellipsis-unused
#' @template fit-intercept
#' @param penalty The penalty type, must be one of
#'   \{"none", "l1", "l2", "elasticnet"\}.
#'   If "none" or "l2" is selected, then L-BFGS solver will be used.
#'   If "l1" is selected, solver OWL-QN will be used.
#'   If "elasticnet" is selected, OWL-QN will be used if l1_ratio > 0, otherwise
#'   L-BFGS will be used. Default: "l2".
#' @param tol Tolerance for stopping criteria. Default: 1e-4.
#' @param C Inverse of regularization strength; must be a positive float.
#'   Default: 1.0.
#' @param class_weight If \code{NULL}, then each class has equal weight of
#'   \code{1}.
#'   If \code{class_weight} is set to \code{"balanced"}, then weights will be
#'   inversely proportional to class frequencies in the input data.
#'   If otherwise, then \code{class_weight} must be a named numeric vector of
#'   weight values, with names being class labels.
#'   If \code{class_weight} is not \code{NULL}, then each entry in
#'   \code{sample_weight} will be adjusted by multiplying its original value
#'   with the class weight of the corresponding sample's class.
#'   Default: NULL.
#' @param sample_weight Array of weights assigned to individual samples.
#'   If \code{NULL}, then each sample has an equal weight of 1. Default: NULL.
#' @param max_iters Maximum number of solver iterations. Default: 1000L.
#' @param linesearch_max_iters Max number of linesearch iterations per outer
#'   iteration used in the LBFGS- and OWL- QN solvers. Default: 50L.
#' @param l1_ratio The Elastic-Net mixing parameter, must \code{NULL} or be
#'   within the range of [0, 1]. Default: NULL.
#'
#' @examples
#' library(cuda.ml)
#'
#' X <- scale(as.matrix(iris[names(iris) != "Species"]))
#' y <- iris$Species
#'
#' model <- cuda_ml_logistic_reg(X, y, max_iters = 100)
#' predictions <- predict(model, X)
#'
#' # NOTE: if we were only performing binary classifications (e.g., by having
#' # `iris_data <- iris %>% mutate(Species = (Species == "setosa"))`), then the
#' # above would be conceptually equivalent to the following:
#' #
#' # iris_data <- iris %>% mutate(Species = (Species == "setosa"))
#' # model <- glm(
#' #   Species ~ ., data = iris_data, family = binomial(link = "logit"),
#' #   control = glm.control(epsilon = 1e-8, maxit = 100)
#' # )
#' #
#' # predict(model, iris_data, type = "response")
#' @importFrom ellipsis check_dots_used
#' @export
cuda_ml_logistic_reg <- function(x, ...) {
  check_dots_used()
  UseMethod("cuda_ml_logistic_reg")
}

#' @rdname cuda_ml_logistic_reg
#' @export
cuda_ml_logistic_reg.default <- function(x, ...) {
  report_undefined_fn("cuda_ml_logistic_reg", x)
}

#' @rdname cuda_ml_logistic_reg
#' @export
cuda_ml_logistic_reg.data.frame <- function(x, y,
                                            fit_intercept = TRUE,
                                            penalty = c("l2", "l1", "elasticnet", "none"),
                                            tol = 1e-4,
                                            C = 1.0,
                                            class_weight = NULL,
                                            sample_weight = NULL,
                                            max_iters = 1000L,
                                            linesearch_max_iters = 50L,
                                            l1_ratio = NULL,
                                            ...) {
  processed <- hardhat::mold(x, y)

  cuda_ml_logistic_reg_bridge(
    processed = processed,
    fit_intercept = fit_intercept,
    penalty = penalty,
    tol = tol,
    C = C,
    class_weight = class_weight,
    sample_weight = sample_weight,
    max_iters = max_iters,
    linesearch_max_iters = linesearch_max_iters,
    l1_ratio = l1_ratio
  )
}

#' @rdname cuda_ml_logistic_reg
#' @export
cuda_ml_logistic_reg.matrix <- function(x, y,
                                        fit_intercept = TRUE,
                                        penalty = c("l2", "l1", "elasticnet", "none"),
                                        tol = 1e-4,
                                        C = 1.0,
                                        class_weight = NULL,
                                        sample_weight = NULL,
                                        max_iters = 1000L,
                                        linesearch_max_iters = 50L,
                                        l1_ratio = NULL,
                                        ...) {
  processed <- hardhat::mold(x, y)

  cuda_ml_logistic_reg_bridge(
    processed = processed,
    fit_intercept = fit_intercept,
    penalty = penalty,
    tol = tol,
    C = C,
    class_weight = class_weight,
    sample_weight = sample_weight,
    max_iters = max_iters,
    linesearch_max_iters = linesearch_max_iters,
    l1_ratio = l1_ratio
  )
}

#' @rdname cuda_ml_logistic_reg
#' @export
cuda_ml_logistic_reg.formula <- function(formula, data,
                                         fit_intercept = TRUE,
                                         penalty = c("l2", "l1", "elasticnet", "none"),
                                         tol = 1e-4,
                                         C = 1.0,
                                         class_weight = NULL,
                                         sample_weight = NULL,
                                         max_iters = 1000L,
                                         linesearch_max_iters = 50L,
                                         l1_ratio = NULL,
                                         ...) {
  processed <- hardhat::mold(formula, data)

  cuda_ml_logistic_reg_bridge(
    processed = processed,
    fit_intercept = fit_intercept,
    penalty = penalty,
    tol = tol,
    C = C,
    class_weight = class_weight,
    sample_weight = sample_weight,
    max_iters = max_iters,
    linesearch_max_iters = linesearch_max_iters,
    l1_ratio = l1_ratio
  )
}

#' @rdname cuda_ml_logistic_reg
#' @export
cuda_ml_logistic_reg.recipe <- function(x, data,
                                        fit_intercept = TRUE,
                                        penalty = c("l2", "l1", "elasticnet", "none"),
                                        tol = 1e-4,
                                        C = 1.0,
                                        class_weight = NULL,
                                        sample_weight = NULL,
                                        max_iters = 1000L,
                                        linesearch_max_iters = 50L,
                                        l1_ratio = NULL,
                                        ...) {
  processed <- hardhat::mold(x, data)

  cuda_ml_logistic_reg_bridge(
    processed = processed,
    fit_intercept = fit_intercept,
    penalty = penalty,
    tol = tol,
    C = C,
    class_weight = class_weight,
    sample_weight = sample_weight,
    max_iters = max_iters,
    linesearch_max_iters = linesearch_max_iters,
    l1_ratio = l1_ratio
  )
}

cuda_ml_logistic_reg_bridge <- function(processed,
                                        fit_intercept,
                                        penalty, tol, C,
                                        class_weight, sample_weight,
                                        max_iters, linesearch_max_iters,
                                        l1_ratio) {
  hardhat::validate_predictors_are_numeric(processed$predictors)
  hardhat::validate_outcomes_are_univariate(processed$outcomes)
  hardhat::validate_outcomes_are_factors(processed$outcomes)

  x <- as.matrix(processed$predictors)
  # convert outcomes to 0-based enum values
  y <- as.integer(processed$outcomes[[1]]) - 1L

  n_classes <- nlevels(processed$outcomes[[1]])
  loss_type <- ifelse(
    n_classes > 2,
    logistic_reg_loss_type$softmax,
    logistic_reg_loss_type$sigmoid
  )

  qn_params <- logistic_reg_build_qn_params(C, l1_ratio, penalty)

  if (!is.null(class_weight)) {
    logistic_reg_validate_class_weight(class_weight, processed)
  }

  if (!is.null(sample_weight)) {
    logistic_reg_validate_sample_weight(sample_weight, processed)
  } else {
    sample_weight <- rep(1, nrow(x))
  }
  sample_weight <- logistic_reg_build_sample_weight(
    sample_weight, class_weight, processed
  )

  model_xptr <- .qn_fit(
    X = x, y = y,
    n_classes = n_classes,
    loss_type = loss_type,
    fit_intercept = fit_intercept,
    l1 = qn_params$l1, l2 = qn_params$l2,
    max_iters = as.integer(max_iters), tol = as.numeric(tol),
    delta = as.numeric(tol) * 0.01,
    linesearch_max_iters = as.integer(linesearch_max_iters),
    lbfgs_memory = 5L,
    sample_weight = as.numeric(sample_weight)
  )

  new_model(
    cls = "cuda_ml_logistic_reg",
    xptr = model_xptr,
    blueprint = processed$blueprint
  )
}

#' Make predictions on new data points.
#'
#' Make predictions on new data points using a CuML logistic regression model.
#'
#' @template predict
#'
#' @importFrom ellipsis check_dots_used
#' @export
predict.cuda_ml_logistic_reg <- function(object, x, ...) {
  check_dots_used()

  processed <- hardhat::forge(x, object$blueprint)
  model <- object$xptr

  preds <- .qn_predict(
    X = as.matrix(processed$predictors),
    n_classes = model$n_classes,
    coefs = model$coefs,
    loss_type = model$loss_type,
    fit_intercept = model$fit_intercept
  )
  # convert from 0-based internal representation to 1-based representation which
  # will correspond to the factor levels of all possible outcomes in R
  preds <- preds + 1L

  postprocess_classification_results(preds, object)
}

context("Logistic Regression")

iris_scaled <- scale(as.matrix(iris[names(iris) != "Species"]))
y <- iris$Species
subset <- c(1:50, 51:60, 141:150)
iris_subset <- iris_scaled[subset, ]
y_subset <- y[subset]

test_that("logistic regression works as expected", {
  model <- cuda_ml_logistic_reg(iris_scaled, y, max_iters = 100)
  preds <- predict(model, iris_scaled)

  expect_gte(sum(preds$.pred_class == iris$Species), 140)
})

test_that("logistic regression works as expected with custom sample weight", {
  sample_weight <- c(rep(1, 50), rep(5, 20))

  model <- cuda_ml_logistic_reg(
    iris_subset, y_subset, max_iters = 100, sample_weight = sample_weight
  )
  preds <- predict(model, iris_scaled)

  expect_gte(sum(preds$.pred_class == iris$Species), 140)
})

test_that("logistic regression works as expected with custom class weight", {
  class_weight <- c(setosa = 1, versicolor = 5, virginica = 5)

  model <- cuda_ml_logistic_reg(
    iris_subset, y_subset, max_iters = 100, class_weight = class_weight
  )
  preds <- predict(model, iris_scaled)

  expect_gte(sum(preds$.pred_class == iris$Species), 140)
})

test_that("logistic regression works as expected with \"balanced\" class weight", {
  model <- cuda_ml_logistic_reg(
    iris_subset, y_subset, max_iters = 100, class_weight = "balanced"
  )
  preds <- predict(model, iris_scaled)

  expect_gte(sum(preds$.pred_class == iris$Species), 140)
})

test_that("logistic regression works as expected with custom sample weight and class weight", {
  sample_weight <- c(rep(1, 50), rep(2, 10), rep(3, 10))
  class_weight <- c(setosa = 1, versicolor = 5 / 2, virginica = 5 / 3)

  model <- cuda_ml_logistic_reg(
    iris_subset, y_subset, max_iters = 100, sample_weight = sample_weight
  )
  preds <- predict(model, iris_scaled)

  expect_gte(sum(preds$.pred_class == iris$Species), 140)
})

test_that("logistic regression works as expected with L1 regularization", {
  model <- cuda_ml_logistic_reg(
    iris_scaled, y, max_iters = 100, penalty = "l1", C = 0.5
  )
  preds <- predict(model, iris_scaled)

  expect_gte(sum(preds$.pred_class == iris$Species), 140)
})

test_that("logistic regression works as expected with L2 regularization", {
  model <- cuda_ml_logistic_reg(
    iris_scaled, y, max_iters = 100, penalty = "l2", C = 0.5
  )
  preds <- predict(model, iris_scaled)

  expect_gte(sum(preds$.pred_class == iris$Species), 140)
})

test_that("logistic regression works as expected with elasticnet regularization", {
  model <- cuda_ml_logistic_reg(
    iris_scaled, y, max_iters = 100, penalty = "elasticnet",
    C = 0.5, l1_ratio = 0.5
  )
  preds <- predict(model, iris_scaled)

  expect_gte(sum(preds$.pred_class == iris$Species), 140)
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/agglomerative.R
\name{cuda_ml_agglomerative_clustering}
\alias{cuda_ml_agglomerative_clustering}
\title{Perform Single-Linkage Agglomerative Clustering.}
\usage{
cuda_ml_agglomerative_clustering(
  x,
  n_clusters = 2L,
  metric = c("euclidean", "l1", "l2", "manhattan", "cosine"),
  connectivity = c("pairwise", "knn"),
  n_neighbors = 15L
)
}
\arguments{
\item{x}{The input matrix or dataframe. Each data point should be a row
and should consist of numeric values only.}

\item{n_clusters}{The number of clusters to find. Default: 2L.}

\item{metric}{Metric used for linkage computation. Must be one of
\{"euclidean", "l1", "l2", "manhattan", "cosine"\}. If connectivity is
"knn" then only "euclidean" is accepted. Default: "euclidean".}

\item{connectivity}{The type of connectivity matrix to compute. Must be one
of \{"pairwise", "knn"\}. Default: "pairwise".
  - 'pairwise' will compute the entire fully-connected graph of pairwise
     distances between each set of points. This is the fastest to compute
     and can be very fast for smaller datasets but requires O(n^2) space.
  - 'knn' will sparsify the fully-connected connectivity matrix to save
    memory and enable much larger inputs. "n_neighbors" will control the
    amount of memory used and the graph will be connected automatically in
    the event "n_neighbors" was not large enough to connect it.}

\item{n_neighbors}{The number of neighbors to compute when
\code{connectivity} is "knn". Default: 15L.}
}
\value{
A clustering object with the following attributes:
  "n_clusters": The number of clusters found by the algorithm.
  "children": The children of each non-leaf node. Values less than
    \code{nrow(x)} correspond to leaves of the tree which are the original
    samples. \code{children[i + 1][1]} and \code{children[i + 1][2]} were
    merged to form node \code{(nrow(x) + i)} in the \code{i}-th iteration.
  "labels": cluster label of each data point.
}
\description{
Recursively merge the pair of clusters that minimally increases a given
linkage distance.
}
\examples{

library(cuda.ml)
library(MASS)
library(magrittr)
library(purrr)

set.seed(0L)

gen_pts <- function() {
  centers <- list(c(1000, 1000), c(-1000, -1000), c(-1000, 1000))
  pts <- centers \%>\%
    map(~ mvrnorm(50, mu = .x, Sigma = diag(2)))

  rlang::exec(rbind, !!!pts) \%>\% as.matrix()
}

clust <- cuda_ml_agglomerative_clustering(
  x = gen_pts(),
  metric = "euclidean",
  n_clusters = 3L
)

print(clust$labels)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model.R
\name{cuda_ml_inverse_transform}
\alias{cuda_ml_inverse_transform}
\title{Apply the inverse transformation defined by a trained cuML model.}
\usage{
cuda_ml_inverse_transform(model, x, ...)
}
\arguments{
\item{model}{A model object.}

\item{x}{The dataset to be transformed.}

\item{...}{Additional model-specific parameters (if any).}
}
\value{
The transformed data points.
}
\description{
Given a trained cuML model, apply the inverse transformation defined by that
model to an input dataset.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rand_proj.R
\name{cuda_ml_rand_proj}
\alias{cuda_ml_rand_proj}
\title{Random projection for dimensionality reduction.}
\usage{
cuda_ml_rand_proj(
  x,
  n_components = NULL,
  eps = 0.1,
  gaussian_method = TRUE,
  density = NULL,
  transform_input = TRUE,
  seed = 0L
)
}
\arguments{
\item{x}{The input matrix or dataframe. Each data point should be a row
and should consist of numeric values only.}

\item{n_components}{Dimensionality of the target projection space. If NULL,
then the parameter is deducted using the Johnson-Lindenstrauss lemma,
taking into consideration the number of samples and the \code{eps}
parameter. Default: NULL.}

\item{eps}{Error tolerance during projection. Default: 0.1.}

\item{gaussian_method}{If TRUE, then use the Gaussian random projection
method. Otherwise, use the sparse random projection method.
See https://en.wikipedia.org/wiki/Random_projection for details.
Default: TRUE.}

\item{density}{Ratio of non-zero component in the random projection matrix.
If NULL, then the value is set to the minimum density as recommended by
Ping Li et al.: 1 / sqrt(n_features). Default: NULL.}

\item{transform_input}{Whether to project input data onto a lower dimension
space using the random matrix. Default: TRUE.}

\item{seed}{Seed for the pseudorandom number generator. Default: 0L.}
}
\value{
A context object containing GPU pointer to a random matrix that can
  be used as input to the \code{cuda_ml_transform()} function.
  If \code{transform_input} is set to TRUE, then the context object will also
  contain a "transformed_data" attribute containing the lower dimensional
  projection of the input data.
}
\description{
Generate a random projection matrix for dimensionality reduction, and
optionally transform input data to a projection in a lower dimension space
using the generated random matrix.
}
\examples{
library(cuda.ml)
library(mlbench)

data(Vehicle)
vehicle_data <- Vehicle[order(Vehicle$Class), which(names(Vehicle) != "Class")]

model <- cuda_ml_rand_proj(vehicle_data, n_components = 4)

set.seed(0L)
print(kmeans(model$transformed_data, centers = 4, iter.max = 1000))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kmeans.R
\name{cuda_ml_kmeans}
\alias{cuda_ml_kmeans}
\title{Run the K means clustering algorithm.}
\usage{
cuda_ml_kmeans(
  x,
  k,
  max_iters = 300,
  tol = 0,
  init_method = c("kmeans++", "random"),
  seed = 0L,
  cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace")
)
}
\arguments{
\item{x}{The input matrix or dataframe. Each data point should be a row
and should consist of numeric values only.}

\item{k}{The number of clusters.}

\item{max_iters}{Maximum number of iterations. Default: 300.}

\item{tol}{Relative tolerance with regards to inertia to declare convergence.
Default: 0 (i.e., do not use inertia-based stopping criterion).}

\item{init_method}{Method for initializing the centroids. Valid methods
include "kmeans++", "random", or a matrix of k rows, each row specifying
the initial value of a centroid. Default: "kmeans++".}

\item{seed}{Seed to the random number generator. Default: 0.}

\item{cuML_log_level}{Log level within cuML library functions. Must be one of
{"off", "critical", "error", "warn", "info", "debug", "trace"}.
Default: off.}
}
\value{
A list containing the cluster assignments and the centroid of each
  cluster. Each centroid will be a column within the `centroids` matrix.
}
\description{
Run the K means clustering algorithm.
}
\examples{

library(cuda.ml)

kclust <- cuda_ml_kmeans(
  iris[names(iris) != "Species"],
  k = 3, max_iters = 100
)

print(kclust)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fil.R
\name{cuda_ml_fil_enabled}
\alias{cuda_ml_fil_enabled}
\title{Determine whether Forest Inference Library (FIL) functionalities are enabled
in the current installation of {cuda.ml}.}
\usage{
cuda_ml_fil_enabled()
}
\value{
A logical value indicating whether the Forest Inference Library (FIL)
  functionalities are enabled.
}
\description{
CuML Forest Inference Library (FIL) functionalities (see
https://github.com/rapidsai/cuml/tree/main/python/cuml/fil#readme) will
require Treelite C API. If you need FIL to run tree-based model ensemble on
GPU, and \code{fil_enabled()} returns FALSE, then please consider installing
Treelite and then re-installing {cuda.ml}.
}
\examples{
if (cuda_ml_fil_enabled()) {
  # run GPU-accelerated Forest Inference Library (FIL) functionalities
} else {
  message(
    "FIL functionalities are disabled in the current installation of ",
    "{cuda.ml}. Please reinstall Treelite C library first, and then re-install",
    " {cuda.ml} to enable FIL."
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cuml_utils.R
\name{cuML_minor_version}
\alias{cuML_minor_version}
\title{Get the minor version of the RAPIDS cuML shared library \{cuda.ml\} was linked
to.}
\usage{
cuML_minor_version()
}
\value{
The minor version of the RAPIDS cuML shared library \{cuda.ml\} was
linked to in a character vector, or \code{NA_character_} if \{cuda.ml\} was not
linked to any version of RAPIDS cuML.
}
\description{
Get the minor version of the RAPIDS cuML shared library \{cuda.ml\} was linked
to.
}
\examples{

library(cuda.ml)

print(cuML_minor_version())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model.R
\name{cuda_ml_is_classifier}
\alias{cuda_ml_is_classifier}
\title{Determine whether a CuML model is a classifier.}
\usage{
cuda_ml_is_classifier(model)
}
\arguments{
\item{model}{A trained CuML model.}
}
\value{
A logical value indicating whether the model is a classifier.
}
\description{
Given a trained CuML model, return \code{TRUE} if the model is a classifier,
otherwise \code{FALSE} (e.g., if the model is a regressor).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package.R
\docType{package}
\name{cuda.ml-package}
\alias{cuda.ml}
\alias{cuda.ml-package}
\title{cuda.ml}
\description{
This package provides a R interface for the RAPIDS cuML library.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://mlverse.github.io/cuda.ml/}
  \item Report bugs at \url{https://github.com/mlverse/cuda.ml/issues}
}

}
\author{
Yitao Li <yitao@rstudio.com>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dbscan.R
\name{cuda_ml_dbscan}
\alias{cuda_ml_dbscan}
\title{Run the DBSCAN clustering algorithm.}
\usage{
cuda_ml_dbscan(
  x,
  min_pts,
  eps,
  cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace")
)
}
\arguments{
\item{x}{The input matrix or dataframe. Each data point should be a row
and should consist of numeric values only.}

\item{min_pts, eps}{A point `p` is a core point if at least `min_pts` are
within distance `eps` from it.}

\item{cuML_log_level}{Log level within cuML library functions. Must be one of
{"off", "critical", "error", "warn", "info", "debug", "trace"}.
Default: off.}
}
\value{
A list containing the cluster assignments of all data points. A data
 point not belonging to any cluster (i.e., "noise") will have NA its cluster
 assignment.
}
\description{
Run the DBSCAN (Density-based spatial clustering of applications with noise)
clustering algorithm.
}
\examples{
library(cuda.ml)
library(magrittr)

gen_pts <- function() {
  centroids <- list(c(1000, 1000), c(-1000, -1000), c(-1000, 1000))

  pts <- centroids \%>\%
    purrr::map(~ MASS::mvrnorm(10, mu = .x, Sigma = diag(2)))

  rlang::exec(rbind, !!!pts)
}

m <- gen_pts()
clusters <- cuda_ml_dbscan(m, min_pts = 5, eps = 3)

print(clusters)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sgd.R
\name{cuda_ml_sgd}
\alias{cuda_ml_sgd}
\alias{cuda_ml_sgd.default}
\alias{cuda_ml_sgd.data.frame}
\alias{cuda_ml_sgd.matrix}
\alias{cuda_ml_sgd.formula}
\alias{cuda_ml_sgd.recipe}
\title{Train a MBSGD linear model.}
\usage{
cuda_ml_sgd(x, ...)

\method{cuda_ml_sgd}{default}(x, ...)

\method{cuda_ml_sgd}{data.frame}(
  x,
  y,
  fit_intercept = TRUE,
  loss = c("squared_loss", "log", "hinge"),
  penalty = c("none", "l1", "l2", "elasticnet"),
  alpha = 1e-04,
  l1_ratio = 0.5,
  epochs = 1000L,
  tol = 0.001,
  shuffle = TRUE,
  learning_rate = c("constant", "invscaling", "adaptive"),
  eta0 = 0.001,
  power_t = 0.5,
  batch_size = 32L,
  n_iters_no_change = 5L,
  ...
)

\method{cuda_ml_sgd}{matrix}(
  x,
  y,
  fit_intercept = TRUE,
  loss = c("squared_loss", "log", "hinge"),
  penalty = c("none", "l1", "l2", "elasticnet"),
  alpha = 1e-04,
  l1_ratio = 0.5,
  epochs = 1000L,
  tol = 0.001,
  shuffle = TRUE,
  learning_rate = c("constant", "invscaling", "adaptive"),
  eta0 = 0.001,
  power_t = 0.5,
  batch_size = 32L,
  n_iters_no_change = 5L,
  ...
)

\method{cuda_ml_sgd}{formula}(
  formula,
  data,
  fit_intercept = TRUE,
  loss = c("squared_loss", "log", "hinge"),
  penalty = c("none", "l1", "l2", "elasticnet"),
  alpha = 1e-04,
  l1_ratio = 0.5,
  epochs = 1000L,
  tol = 0.001,
  shuffle = TRUE,
  learning_rate = c("constant", "invscaling", "adaptive"),
  eta0 = 0.001,
  power_t = 0.5,
  batch_size = 32L,
  n_iters_no_change = 5L,
  ...
)

\method{cuda_ml_sgd}{recipe}(
  x,
  data,
  fit_intercept = TRUE,
  loss = c("squared_loss", "log", "hinge"),
  penalty = c("none", "l1", "l2", "elasticnet"),
  alpha = 1e-04,
  l1_ratio = 0.5,
  epochs = 1000L,
  tol = 0.001,
  shuffle = TRUE,
  learning_rate = c("constant", "invscaling", "adaptive"),
  eta0 = 0.001,
  power_t = 0.5,
  batch_size = 32L,
  n_iters_no_change = 5L,
  ...
)
}
\arguments{
\item{x}{Depending on the context:

  * A __data frame__ of predictors.
  * A __matrix__ of predictors.
  * A __recipe__ specifying a set of preprocessing steps
  * created from [recipes::recipe()].
  * A __formula__ specifying the predictors and the outcome.}

\item{...}{Optional arguments; currently unused.}

\item{y}{A numeric vector (for regression) or factor (for classification) of
desired responses.}

\item{fit_intercept}{If TRUE, then the model tries to correct for the global
mean of the response variable. If FALSE, then the model expects data to be
centered. Default: TRUE.}

\item{loss}{Loss function, must be one of \{"squared_loss", "log", "hinge"\}.}

\item{penalty}{Type of regularization to perform, must be one of
  \{"none", "l1", "l2", "elasticnet"\}.

  - "none": no regularization.
  - "l1": perform regularization based on the L1-norm (LASSO) which tries to
          minimize the sum of the absolute values of the coefficients.
  - "l2": perform regularization based on the L2 norm (Ridge) which tries to
          minimize the sum of the square of the coefficients.
  - "elasticnet": perform the Elastic Net regularization which is based on
                  the weighted averable of L1 and L2 norms.
  Default: "none".}

\item{alpha}{Multiplier of the penalty term. Default: 1e-4.}

\item{l1_ratio}{The ElasticNet mixing parameter, with 0 <= l1_ratio <= 1.
For l1_ratio = 0 the penalty is an L2 penalty. For l1_ratio = 1 it is an L1
penalty.
For 0 < l1_ratio < 1, the penalty is a combination of L1 and L2.
The penalty term is computed using the following formula:
  penalty = \code{alpha} * \code{l1_ratio} * ||w||_1 +
            0.5 * \code{alpha} * (1 - \code{l1_ratio}) * ||w||^2_2
where ||w||_1 is the L1 norm of the coefficients, and ||w||_2 is the L2
norm of the coefficients.}

\item{epochs}{The number of times the model should iterate through the entire
dataset during training. Default: 1000L.}

\item{tol}{Threshold for stopping training. Training will stop if
(loss in current epoch) > (loss in previous epoch) - \code{tol}.
Default: 1e-3.}

\item{shuffle}{Whether to shuffles the training data after each epoch.
Default: True.}

\item{learning_rate}{Must be one of \{"constant", "invscaling", "adaptive"\}.

  - "constant": the learning rate will be kept constant.
  - "invscaling": (learning rate) = (initial learning rate) / pow(t, power_t)
                  where \code{t} is the number of epochs and
                  \code{power_t} is a tunable parameter of this model.
  - "adaptive": (learning rate) = (initial learning rate) as long as the
                training loss keeps decreasing. Each time the last
                \code{n_iter_no_change} consecutive epochs fail to decrease
                the training loss by \code{tol}, the current learning rate is
                divided by 5.
  Default: "constant".}

\item{eta0}{The initial learning rate. Default: 1e-3.}

\item{power_t}{The exponent used in the invscaling learning rate
calculations.}

\item{batch_size}{The number of samples that will be included in each batch.
Default: 32L.}

\item{n_iters_no_change}{The maximum number of epochs to train if there is no
imporvement in the model. Default: 5.}

\item{formula}{A formula specifying the outcome terms on the left-hand side,
and the predictor terms on the right-hand side.}

\item{data}{When a __recipe__ or __formula__ is used, \code{data} is
specified as a  __data frame__ containing the predictors and (if
applicable) the outcome.}
}
\value{
A linear model that can be used with the 'predict' S3 generic to make
  predictions on new data points.
}
\description{
Train a linear model using mini-batch stochastic gradient descent.
}
\examples{

library(cuda.ml)

model <- cuda_ml_sgd(
  mpg ~ ., mtcars,
  batch_size = 4L, epochs = 50000L,
  learning_rate = "adaptive", eta0 = 1e-5,
  penalty = "l2", alpha = 1e-5, tol = 1e-6,
  n_iters_no_change = 10L
)

preds <- predict(model, mtcars[names(mtcars) != "mpg"])
print(all.equal(preds$.pred, mtcars$mpg, tolerance = 0.09))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model.R
\name{cuda_ml_unserialize}
\alias{cuda_ml_unserialize}
\alias{cuda_ml_unserialise}
\title{Unserialize a CuML model state}
\usage{
cuda_ml_unserialize(connection, ...)

cuda_ml_unserialise(connection, ...)
}
\arguments{
\item{connection}{An open connection or a raw vector.}

\item{...}{Additional arguments to \code{base::unserialize()}.}
}
\value{
A unserialized CuML model.
}
\description{
Unserialize a CuML model state into a CuML model object.
}
\seealso{
\code{\link[base]{unserialize}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model.R
\name{cuda_ml_can_predict_class_probabilities}
\alias{cuda_ml_can_predict_class_probabilities}
\title{Determine whether a CuML model can predict class probabilities.}
\usage{
cuda_ml_can_predict_class_probabilities(model)
}
\arguments{
\item{model}{A trained CuML model.}
}
\value{
A logical value indicating whether the model supports outputting
  class probabilities.
}
\description{
Given a trained CuML model, return \code{TRUE} if the model is a classifier
and is capable of outputting class probabilities as prediction results (e.g.,
if the model is a KNN or an ensemble classifier), otherwise return
\code{FALSE}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/svm.R
\name{cuda_ml_svm}
\alias{cuda_ml_svm}
\alias{cuda_ml_svm.default}
\alias{cuda_ml_svm.data.frame}
\alias{cuda_ml_svm.matrix}
\alias{cuda_ml_svm.formula}
\alias{cuda_ml_svm.recipe}
\title{Train a SVM model.}
\usage{
cuda_ml_svm(x, ...)

\method{cuda_ml_svm}{default}(x, ...)

\method{cuda_ml_svm}{data.frame}(
  x,
  y,
  cost = 1,
  kernel = c("rbf", "tanh", "polynomial", "linear"),
  gamma = NULL,
  coef0 = 0,
  degree = 3L,
  tol = 0.001,
  max_iter = NULL,
  nochange_steps = 1000L,
  cache_size = 1024,
  epsilon = 0.1,
  sample_weights = NULL,
  cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace"),
  ...
)

\method{cuda_ml_svm}{matrix}(
  x,
  y,
  cost = 1,
  kernel = c("rbf", "tanh", "polynomial", "linear"),
  gamma = NULL,
  coef0 = 0,
  degree = 3L,
  tol = 0.001,
  max_iter = NULL,
  nochange_steps = 1000L,
  cache_size = 1024,
  epsilon = 0.1,
  sample_weights = NULL,
  cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace"),
  ...
)

\method{cuda_ml_svm}{formula}(
  formula,
  data,
  cost = 1,
  kernel = c("rbf", "tanh", "polynomial", "linear"),
  gamma = NULL,
  coef0 = 0,
  degree = 3L,
  tol = 0.001,
  max_iter = NULL,
  nochange_steps = 1000L,
  cache_size = 1024,
  epsilon = 0.1,
  sample_weights = NULL,
  cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace"),
  ...
)

\method{cuda_ml_svm}{recipe}(
  x,
  data,
  cost = 1,
  kernel = c("rbf", "tanh", "polynomial", "linear"),
  gamma = NULL,
  coef0 = 0,
  degree = 3L,
  tol = 0.001,
  max_iter = NULL,
  nochange_steps = 1000L,
  cache_size = 1024,
  epsilon = 0.1,
  sample_weights = NULL,
  cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace"),
  ...
)
}
\arguments{
\item{x}{Depending on the context:

  * A __data frame__ of predictors.
  * A __matrix__ of predictors.
  * A __recipe__ specifying a set of preprocessing steps
  * created from [recipes::recipe()].
  * A __formula__ specifying the predictors and the outcome.}

\item{...}{Optional arguments; currently unused.}

\item{y}{A numeric vector (for regression) or factor (for classification) of
desired responses.}

\item{cost}{A positive number for the cost of predicting a sample within or
on the wrong side of the margin. Default: 1.}

\item{kernel}{Type of the SVM kernel function (must be one of "rbf", "tanh",
"polynomial", or "linear"). Default: "rbf".}

\item{gamma}{The gamma coefficient (only relevant to polynomial, RBF, and
  tanh kernel functions, see explanations below).
  Default: 1 / (num features).

  The following kernels are implemented:
    - RBF K(x_1, x_2) = exp(-gamma |x_1-x_2|^2)
    - TANH K(x_1, x_2) = tanh(gamma <x_1,x_2> + coef0)
    - POLYNOMIAL K(x_1, x_2) = (gamma <x_1,x_2> + coef0)^degree
    - LINEAR K(x_1,x_2) = <x_1,x_2>,
  where < , > denotes the dot product.}

\item{coef0}{The 0th coefficient (only applicable to polynomial and tanh
  kernel functions, see explanations below). Default: 0.

  The following kernels are implemented:
    - RBF K(x_1, x_2) = exp(-gamma |x_1-x_2|^2)
    - TANH K(x_1, x_2) = tanh(gamma <x_1,x_2> + coef0)
    - POLYNOMIAL K(x_1, x_2) = (gamma <x_1,x_2> + coef0)^degree
    - LINEAR K(x_1,x_2) = <x_1,x_2>,
  where < , > denotes the dot product.}

\item{degree}{Degree of the polynomial kernel function (note: not applicable
  to other kernel types, see explanations below). Default: 3.

  The following kernels are implemented:
    - RBF K(x_1, x_2) = exp(-gamma |x_1-x_2|^2)
    - TANH K(x_1, x_2) = tanh(gamma <x_1,x_2> + coef0)
    - POLYNOMIAL K(x_1, x_2) = (gamma <x_1,x_2> + coef0)^degree
    - LINEAR K(x_1,x_2) = <x_1,x_2>,
  where < , > denotes the dot product.}

\item{tol}{Tolerance to stop fitting. Default: 1e-3.}

\item{max_iter}{Maximum number of outer iterations in SmoSolver.
Default: 100 * (num samples).}

\item{nochange_steps}{Number of steps with no change w.r.t convergence.
Default: 1000.}

\item{cache_size}{Size of kernel cache (MiB) in device memory. Default: 1024.}

\item{epsilon}{Espsilon parameter of the epsilon-SVR model. There is no
penalty for points that are predicted within the epsilon-tube around the
target values. Please note this parameter is only relevant for regression
tasks. Default: 0.1.}

\item{sample_weights}{Optional weight assigned to each input data point.}

\item{cuML_log_level}{Log level within cuML library functions. Must be one of
{"off", "critical", "error", "warn", "info", "debug", "trace"}.
Default: off.}

\item{formula}{A formula specifying the outcome terms on the left-hand side,
and the predictor terms on the right-hand side.}

\item{data}{When a __recipe__ or __formula__ is used, \code{data} is
specified as a  __data frame__ containing the predictors and (if
applicable) the outcome.}
}
\value{
A SVM classifier / regressor object that can be used with the
  'predict' S3 generic to make predictions on new data points.
}
\description{
Train a Support Vector Machine model for classification or regression tasks.
}
\examples{

library(cuda.ml)

# Classification

model <- cuda_ml_svm(
  formula = Species ~ .,
  data = iris,
  kernel = "rbf"
)

predictions <- predict(model, iris[names(iris) != "Species"])

# Regression

model <- cuda_ml_svm(
  formula = mpg ~ .,
  data = mtcars,
  kernel = "rbf"
)

predictions <- predict(model, mtcars)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cuml_utils.R
\name{has_cuML}
\alias{has_cuML}
\title{Determine whether \{cuda.ml\} was linked to a valid version of the RAPIDS cuML
shared library.}
\usage{
has_cuML()
}
\value{
A logical value indicating whether the current installation \{cuda.ml\}
  was linked to a valid version of the RAPIDS cuML shared library.
}
\description{
Determine whether \{cuda.ml\} was linked to a valid version of the RAPIDS cuML
shared library.
}
\examples{

library(cuda.ml)

if (!has_cuML()) {
  warning(
    "Please install the RAPIDS cuML shared library first, and then re-",
    "install {cuda.ml}."
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pca.R
\name{cuda_ml_pca}
\alias{cuda_ml_pca}
\title{Perform principal component analysis.}
\usage{
cuda_ml_pca(
  x,
  n_components = NULL,
  eig_algo = c("dq", "jacobi"),
  tol = 1e-07,
  n_iters = 15L,
  whiten = FALSE,
  transform_input = TRUE,
  cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace")
)
}
\arguments{
\item{x}{The input matrix or dataframe. Each data point should be a row
and should consist of numeric values only.}

\item{n_components}{Number of principal component(s) to keep. Default:
min(nrow(x), ncol(x)).}

\item{eig_algo}{Eigen decomposition algorithm to be applied to the covariance
matrix. Valid choices are "dq" (divid-and-conquer method for symmetric
matrices) and "jacobi" (the Jacobi method for symmetric matrices).
Default: "dq".}

\item{tol}{Tolerance for singular values computed by the Jacobi method.
Default: 1e-7.}

\item{n_iters}{Maximum number of iterations for the Jacobi method.
Default: 15.}

\item{whiten}{If TRUE, then de-correlate all components, making each
component have unit variance  and removing multi-collinearity.
Default: FALSE.}

\item{transform_input}{If TRUE, then compute an approximate representation
of the input data. Default: TRUE.}

\item{cuML_log_level}{Log level within cuML library functions. Must be one of
{"off", "critical", "error", "warn", "info", "debug", "trace"}.
Default: off.}
}
\value{
A PCA model object with the following attributes:
   - "components": a matrix of \code{n_components} rows containing the top
     principal components.
   - "explained_variance": amount of variance within the input data explained
     by each component.
   - "explained_variance_ratio": fraction of variance within the input data
     explained by each component.
   - "singular_values": singular values (non-negative) corresponding to the
     top principal components.
   - "mean": the column wise mean of \code{x} which was used to mean-center
     \code{x} first.
   - "transformed_data": (only present if "transform_input" is set to TRUE)
     an approximate representation of input data based on principal
     components.
   - "pca_params": opaque pointer to PCA parameters which will be used for
     performing inverse transforms.

 The model object can be used as input to the inverse_transform() function to
 map a representation based on principal components back to the original
 feature space.
}
\description{
Compute principal component(s) of the input data. Each feature from the input
will be mean-centered (but not scaled) before the SVD computation takes
place.
}
\examples{

library(cuda.ml)

iris.pca <- cuda_ml_pca(iris[1:4], n_components = 3)
print(iris.pca)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/knn.R
\name{cuda_ml_knn}
\alias{cuda_ml_knn}
\alias{cuda_ml_knn.default}
\alias{cuda_ml_knn.data.frame}
\alias{cuda_ml_knn.matrix}
\alias{cuda_ml_knn.formula}
\alias{cuda_ml_knn.recipe}
\title{Build a KNN model.}
\usage{
cuda_ml_knn(x, ...)

\method{cuda_ml_knn}{default}(x, ...)

\method{cuda_ml_knn}{data.frame}(
  x,
  y,
  algo = c("brute", "ivfflat", "ivfpq", "ivfsq"),
  metric = c("euclidean", "l2", "l1", "cityblock", "taxicab", "manhattan", "braycurtis",
    "canberra", "minkowski", "chebyshev", "jensenshannon", "cosine", "correlation"),
  p = 2,
  neighbors = 5L,
  ...
)

\method{cuda_ml_knn}{matrix}(
  x,
  y,
  algo = c("brute", "ivfflat", "ivfpq", "ivfsq"),
  metric = c("euclidean", "l2", "l1", "cityblock", "taxicab", "manhattan", "braycurtis",
    "canberra", "minkowski", "chebyshev", "jensenshannon", "cosine", "correlation"),
  p = 2,
  neighbors = 5L,
  ...
)

\method{cuda_ml_knn}{formula}(
  formula,
  data,
  algo = c("brute", "ivfflat", "ivfpq", "ivfsq"),
  metric = c("euclidean", "l2", "l1", "cityblock", "taxicab", "manhattan", "braycurtis",
    "canberra", "minkowski", "chebyshev", "jensenshannon", "cosine", "correlation"),
  p = 2,
  neighbors = 5L,
  ...
)

\method{cuda_ml_knn}{recipe}(
  x,
  data,
  algo = c("brute", "ivfflat", "ivfpq", "ivfsq"),
  metric = c("euclidean", "l2", "l1", "cityblock", "taxicab", "manhattan", "braycurtis",
    "canberra", "minkowski", "chebyshev", "jensenshannon", "cosine", "correlation"),
  p = 2,
  neighbors = 5L,
  ...
)
}
\arguments{
\item{x}{Depending on the context:

  * A __data frame__ of predictors.
  * A __matrix__ of predictors.
  * A __recipe__ specifying a set of preprocessing steps
  * created from [recipes::recipe()].
  * A __formula__ specifying the predictors and the outcome.}

\item{...}{Optional arguments; currently unused.}

\item{y}{A numeric vector (for regression) or factor (for classification) of
desired responses.}

\item{algo}{The query algorithm to use. Must be one of
  \{"brute", "ivfflat", "ivfpq", "ivfsq"\} or a KNN algorithm specification
  constructed using the \code{cuda_ml_knn_algo_*} family of functions.
  If the algorithm is specified by one of the \code{cuda_ml_knn_algo_*}
  functions, then values of all required parameters of the algorithm will
  need to be specified explicitly.
  If the algorithm is specified by a character vector, then parameters for
  the algorithm are generated automatically.

  Descriptions of supported algorithms:
    - "brute": for brute-force, slow but produces exact results.
    - "ivfflat": for inverted file, divide the dataset in partitions
                 and perform search on relevant partitions only.
    - "ivfpq": for inverted file and product quantization (vectors
               are divided into sub-vectors, and each sub-vector is encoded
               using intermediary k-means clusterings to provide partial
               information).
    - "ivfsq": for inverted file and scalar quantization (vectors components
               are quantized into reduced binary representation allowing
               faster distances calculations).

  Default: "brute".}

\item{metric}{Distance metric to use. Must be one of \{"euclidean", "l2",
"l1", "cityblock", "taxicab", "manhattan", "braycurtis", "canberra",
"minkowski", "lp", "chebyshev", "linf", "jensenshannon", "cosine",
"correlation"\}.
Default: "euclidean".}

\item{p}{Parameter for the Minkowski metric. If p = 1, then the metric is
equivalent to manhattan distance (l1). If p = 2, the metric is equivalent
to euclidean distance (l2).}

\item{neighbors}{Number of nearest neighbors to query. Default: 5L.}

\item{formula}{A formula specifying the outcome terms on the left-hand side,
and the predictor terms on the right-hand side.}

\item{data}{When a __recipe__ or __formula__ is used, \code{data} is
specified as a  __data frame__ containing the predictors and (if
applicable) the outcome.}
}
\value{
A KNN model that can be used with the 'predict' S3 generic to make
  predictions on new data points.
  The model object contains the following:
    - "knn_index": a GPU pointer to the KNN index.
    - "algo": enum value of the algorithm being used for the KNN query.
    - "metric": enum value of the distance metric used in KNN computations.
    - "p": parameter for the Minkowski metric.
    - "n_samples": number of input data points.
    - "n_dims": dimension of each input data point.
}
\description{
Build a k-nearest-model for classification or regression tasks.
}
\examples{

library(cuda.ml)
library(MASS)
library(magrittr)
library(purrr)

set.seed(0L)

centers <- list(c(3, 3), c(-3, -3), c(-3, 3))

gen_pts <- function(cluster_sz) {
  pts <- centers \%>\%
    map(~ mvrnorm(cluster_sz, mu = .x, Sigma = diag(2)))

  rlang::exec(rbind, !!!pts) \%>\% as.matrix()
}

gen_labels <- function(cluster_sz) {
  seq_along(centers) \%>\%
    sapply(function(x) rep(x, cluster_sz)) \%>\%
    factor()
}

sample_cluster_sz <- 1000
sample_pts <- cbind(
  gen_pts(sample_cluster_sz) \%>\% as.data.frame(),
  label = gen_labels(sample_cluster_sz)
)

model <- cuda_ml_knn(label ~ ., sample_pts, algo = "ivfflat", metric = "euclidean")

test_cluster_sz <- 10
test_pts <- gen_pts(test_cluster_sz) \%>\% as.data.frame()

predictions <- predict(model, test_pts)
print(predictions, n = 30)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/elastic_net.R
\name{cuda_ml_elastic_net}
\alias{cuda_ml_elastic_net}
\alias{cuda_ml_elastic_net.default}
\alias{cuda_ml_elastic_net.data.frame}
\alias{cuda_ml_elastic_net.matrix}
\alias{cuda_ml_elastic_net.formula}
\alias{cuda_ml_elastic_net.recipe}
\title{Train a linear model using elastic regression.}
\usage{
cuda_ml_elastic_net(x, ...)

\method{cuda_ml_elastic_net}{default}(x, ...)

\method{cuda_ml_elastic_net}{data.frame}(
  x,
  y,
  alpha = 1,
  l1_ratio = 0.5,
  max_iter = 1000L,
  tol = 0.001,
  fit_intercept = TRUE,
  normalize_input = FALSE,
  selection = c("cyclic", "random"),
  ...
)

\method{cuda_ml_elastic_net}{matrix}(
  x,
  y,
  alpha = 1,
  l1_ratio = 0.5,
  max_iter = 1000L,
  tol = 0.001,
  fit_intercept = TRUE,
  normalize_input = FALSE,
  selection = c("cyclic", "random"),
  ...
)

\method{cuda_ml_elastic_net}{formula}(
  formula,
  data,
  alpha = 1,
  l1_ratio = 0.5,
  max_iter = 1000L,
  tol = 0.001,
  fit_intercept = TRUE,
  normalize_input = FALSE,
  selection = c("cyclic", "random"),
  ...
)

\method{cuda_ml_elastic_net}{recipe}(
  x,
  data,
  alpha = 1,
  l1_ratio = 0.5,
  max_iter = 1000L,
  tol = 0.001,
  fit_intercept = TRUE,
  normalize_input = FALSE,
  selection = c("cyclic", "random"),
  ...
)
}
\arguments{
\item{x}{Depending on the context:

  * A __data frame__ of predictors.
  * A __matrix__ of predictors.
  * A __recipe__ specifying a set of preprocessing steps
  * created from [recipes::recipe()].
  * A __formula__ specifying the predictors and the outcome.}

\item{...}{Optional arguments; currently unused.}

\item{y}{A numeric vector (for regression) or factor (for classification) of
desired responses.}

\item{alpha}{Multiplier of the penalty term (i.e., the result would become
and Ordinary Least Square model if \code{alpha} were set to 0). Default: 1.
For numerical reasons, running elastic regression with \code{alpha} set to
0 is not advised. For the \code{alpha}-equals-to-0 scenario, one should use
\code{cuda_ml_ols} to train an OLS model instead.
Default: 1.}

\item{l1_ratio}{The ElasticNet mixing parameter, with 0 <= l1_ratio <= 1.
For l1_ratio = 0 the penalty is an L2 penalty. For l1_ratio = 1 it is an L1
penalty.
For 0 < l1_ratio < 1, the penalty is a combination of L1 and L2.
The penalty term is computed using the following formula:
  penalty = \code{alpha} * \code{l1_ratio} * ||w||_1 +
            0.5 * \code{alpha} * (1 - \code{l1_ratio}) * ||w||^2_2
where ||w||_1 is the L1 norm of the coefficients, and ||w||_2 is the L2
norm of the coefficients.}

\item{max_iter}{The maximum number of coordinate descent iterations.
Default: 1000L.}

\item{tol}{Stop the coordinate descent when the duality gap is below this
threshold. Default: 1e-3.}

\item{fit_intercept}{If TRUE, then the model tries to correct for the global
mean of the response variable. If FALSE, then the model expects data to be
centered. Default: TRUE.}

\item{normalize_input}{Ignored when \code{fit_intercept} is FALSE. If TRUE,
then the predictors will be normalized to have a L2 norm of 1.
Default: FALSE.}

\item{selection}{If "random", then instead of updating coefficients in cyclic
order, a random coefficient is updated in each iteration. Default: "cyclic".}

\item{formula}{A formula specifying the outcome terms on the left-hand side,
and the predictor terms on the right-hand side.}

\item{data}{When a __recipe__ or __formula__ is used, \code{data} is
specified as a  __data frame__ containing the predictors and (if
applicable) the outcome.}
}
\value{
An elastic net regressor that can be used with the 'predict' S3
  generic to make predictions on new data points.
}
\description{
Train a linear model with combined L1 and L2 priors as the regularizer.
}
\examples{

library(cuda.ml)

model <- cuda_ml_elastic_net(
  formula = mpg ~ ., data = mtcars, alpha = 1e-3, l1_ratio = 0.6
)
cuda_ml_predictions <- predict(model, mtcars)

# predictions will be comparable to those from a `glmnet` model with `lambda`
# set to 1e-3 and `alpha` set to 0.6
# (in `glmnet`, `lambda` is the weight of the penalty term, and `alpha` is
#  the elastic mixing parameter between L1 and L2 penalties.

library(glmnet)

glmnet_model <- glmnet(
  x = as.matrix(mtcars[names(mtcars) != "mpg"]), y = mtcars$mpg,
  alpha = 0.6, lambda = 1e-3, nlambda = 1, standardize = FALSE
)

glm_predictions <- predict(
  glmnet_model, as.matrix(mtcars[names(mtcars) != "mpg"]),
  s = 0
)

print(
  all.equal(
    as.numeric(glm_predictions),
    cuda_ml_predictions$.pred,
    tolerance = 1e-2
  )
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rand_forest.R
\name{cuda_ml_rand_forest}
\alias{cuda_ml_rand_forest}
\alias{cuda_ml_rand_forest.default}
\alias{cuda_ml_rand_forest.data.frame}
\alias{cuda_ml_rand_forest.matrix}
\alias{cuda_ml_rand_forest.formula}
\alias{cuda_ml_rand_forest.recipe}
\title{Train a random forest model.}
\usage{
cuda_ml_rand_forest(x, ...)

\method{cuda_ml_rand_forest}{default}(x, ...)

\method{cuda_ml_rand_forest}{data.frame}(
  x,
  y,
  mtry = NULL,
  trees = NULL,
  min_n = 2L,
  bootstrap = TRUE,
  max_depth = 16L,
  max_leaves = Inf,
  max_predictors_per_note_split = NULL,
  n_bins = 128L,
  min_samples_leaf = 1L,
  split_criterion = NULL,
  min_impurity_decrease = 0,
  max_batch_size = 128L,
  n_streams = 8L,
  cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace"),
  ...
)

\method{cuda_ml_rand_forest}{matrix}(
  x,
  y,
  mtry = NULL,
  trees = NULL,
  min_n = 2L,
  bootstrap = TRUE,
  max_depth = 16L,
  max_leaves = Inf,
  max_predictors_per_note_split = NULL,
  n_bins = 128L,
  min_samples_leaf = 1L,
  split_criterion = NULL,
  min_impurity_decrease = 0,
  max_batch_size = 128L,
  n_streams = 8L,
  cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace"),
  ...
)

\method{cuda_ml_rand_forest}{formula}(
  formula,
  data,
  mtry = NULL,
  trees = NULL,
  min_n = 2L,
  bootstrap = TRUE,
  max_depth = 16L,
  max_leaves = Inf,
  max_predictors_per_note_split = NULL,
  n_bins = 128L,
  min_samples_leaf = 1L,
  split_criterion = NULL,
  min_impurity_decrease = 0,
  max_batch_size = 128L,
  n_streams = 8L,
  cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace"),
  ...
)

\method{cuda_ml_rand_forest}{recipe}(
  x,
  data,
  mtry = NULL,
  trees = NULL,
  min_n = 2L,
  bootstrap = TRUE,
  max_depth = 16L,
  max_leaves = Inf,
  max_predictors_per_note_split = NULL,
  n_bins = 128L,
  min_samples_leaf = 1L,
  split_criterion = NULL,
  min_impurity_decrease = 0,
  max_batch_size = 128L,
  n_streams = 8L,
  cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace"),
  ...
)
}
\arguments{
\item{x}{Depending on the context:

  * A __data frame__ of predictors.
  * A __matrix__ of predictors.
  * A __recipe__ specifying a set of preprocessing steps
  * created from [recipes::recipe()].
  * A __formula__ specifying the predictors and the outcome.}

\item{...}{Optional arguments; currently unused.}

\item{y}{A numeric vector (for regression) or factor (for classification) of
desired responses.}

\item{mtry}{The number of predictors that will be randomly sampled at each
split when creating the tree models. Default: the square root of the total
number of predictors.}

\item{trees}{An integer for the number of trees contained in the ensemble.
Default: 100L.}

\item{min_n}{An integer for the minimum number of data points in a node that
are required for the node to be split further. Default: 2L.}

\item{bootstrap}{Whether to perform bootstrap.
If TRUE, each tree in the forest is built on a bootstrapped sample with
replacement.
If FALSE, the whole dataset is used to build each tree.}

\item{max_depth}{Maximum tree depth. Default: 16L.}

\item{max_leaves}{Maximum leaf nodes per tree. Soft constraint. Default: Inf
(unlimited).}

\item{max_predictors_per_note_split}{Number of predictor to consider per node
split. Default: square root of the total number predictors.}

\item{n_bins}{Number of bins used by the split algorithm. Default: 128L.}

\item{min_samples_leaf}{The minimum number of data points in each leaf node.
Default: 1L.}

\item{split_criterion}{The criterion used to split nodes, can be "gini" or
"entropy" for classifications, and "mse" or "mae" for regressions.
Default: "gini" for classification; "mse" for regression.}

\item{min_impurity_decrease}{Minimum decrease in impurity requried for node
to be spilt. Default: 0.}

\item{max_batch_size}{Maximum number of nodes that can be processed in a
given batch. Default: 128L.}

\item{n_streams}{Number of CUDA streams to use for building trees.
Default: 8L.}

\item{cuML_log_level}{Log level within cuML library functions. Must be one of
{"off", "critical", "error", "warn", "info", "debug", "trace"}.
Default: off.}

\item{formula}{A formula specifying the outcome terms on the left-hand side,
and the predictor terms on the right-hand side.}

\item{data}{When a __recipe__ or __formula__ is used, \code{data} is
specified as a  __data frame__ containing the predictors and (if
applicable) the outcome.}
}
\value{
A random forest classifier / regressor object that can be used with
  the 'predict' S3 generic to make predictions on new data points.
}
\description{
Train a random forest model for classification or regression tasks.
}
\examples{
library(cuda.ml)

# Classification

model <- cuda_ml_rand_forest(
  formula = Species ~ .,
  data = iris,
  trees = 100
)

predictions <- predict(model, iris[names(iris) != "Species"])

# Regression

model <- cuda_ml_rand_forest(
  formula = mpg ~ .,
  data = mtcars,
  trees = 100
)

predictions <- predict(model, mtcars[names(mtcars) != "mpg"])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fil.R
\name{cuda_ml_fil_load_model}
\alias{cuda_ml_fil_load_model}
\title{Load a XGBoost or LightGBM model file.}
\usage{
cuda_ml_fil_load_model(
  filename,
  mode = c("classification", "regression"),
  model_type = c("xgboost", "lightgbm"),
  algo = c("auto", "naive", "tree_reorg", "batch_tree_reorg"),
  threshold = 0.5,
  storage_type = c("auto", "dense", "sparse"),
  threads_per_tree = 1L,
  n_items = 0L,
  blocks_per_sm = 0L
)
}
\arguments{
\item{filename}{Path to the saved model file.}

\item{mode}{Type of task to be performed by the model. Must be one of
\{"classification", "regression"\}.}

\item{model_type}{Format of the saved model file. Notice if \code{filename}
ends with ".json" and \code{model_type} is "xgboost", then \{cuda.ml\} will
assume the model file is in XGBoost JSON (instead of binary) format.
Default: "xgboost".}

\item{algo}{Type of the algorithm for inference, must be one of the
following.
  - "auto":
      Choose the algorithm automatically. Currently 'batch_tree_reorg' is
      used for dense storage, and 'naive' for sparse storage.
  - "naive":
      Simple inference using shared memory.
  - "tree_reorg":
      Similar to naive but with trees rearranged to be more coalescing-
      friendly.
  - "batch_tree_reorg":
      Similar to 'tree_reorg' but predicting multiple rows per thread
      block.
Default: "auto".}

\item{threshold}{Class probability threshold for classification. Ignored for
regression tasks. Default: 0.5.}

\item{storage_type}{In-memory storage format of the FIL model. Must be one of
the following.
- "auto":
    Choose the storage type automatically,
- "dense":
    Create a dense forest,
- "sparse":
    Create a sparse forest. Requires \code{algo} to be 'naive' or 'auto'.}

\item{threads_per_tree}{If >1, then have multiple (neighboring) threads infer
on the same tree within a block, which will improve memory bandwith near
tree root (but consuming more shared memory). Default: 1L.}

\item{n_items}{Number of input samples each thread processes. If 0, then
choose (up to 4) that fit into shared memory. Default: 0L.}

\item{blocks_per_sm}{Indicates how CuML should determine the number of thread
blocks to lauch for the inference kernel.
- 0:
  Launches the number of blocks proportional to the number of data points.
- >= 1:
  Attempts to lauch \code{blocks_per_sm} blocks for each streaming
  multiprocessor.
  This will fail if \code{blocks_per_sm} blocks result in more threads than
  the maximum supported number of threads per GPU. Even if successful, it
  is not guaranteed that \code{blocks_per_sm} blocks will run on an SM
  concurrently.}
}
\value{
A GPU-accelerated FIL model that can be used with the 'predict' S3
  generic to make predictions on new data points.
}
\description{
Load a XGBoost or LightGBM model file using Treelite. The resulting model
object can be used to perform high-throughput batch inference on new data
points using the GPU acceleration functionality from the CuML Forest
Inference Library (FIL).
}
\examples{

library(cuda.ml)
library(xgboost)

model_path <- file.path(tempdir(), "xgboost.model")

model <- xgboost(
  data = as.matrix(mtcars[names(mtcars) != "mpg"]),
  label = as.matrix(mtcars["mpg"]),
  max.depth = 6,
  eta = 1,
  nthread = 2,
  nrounds = 20,
  objective = "reg:squarederror"
)

xgb.save(model, model_path)

model <- cuda_ml_fil_load_model(
  model_path,
  mode = "regression",
  model_type = "xgboost"
)

preds <- predict(model, mtcars[names(mtcars) != "mpg"])

print(preds)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/knn.R
\name{cuda_ml_knn_algo_ivfsq}
\alias{cuda_ml_knn_algo_ivfsq}
\title{Build a specification for the "ivfsq" KNN query algorithm.}
\usage{
cuda_ml_knn_algo_ivfsq(
  nlist,
  nprobe,
  qtype = c("QT_8bit", "QT_4bit", "QT_8bit_uniform", "QT_4bit_uniform", "QT_fp16",
    "QT_8bit_direct", "QT_6bit"),
  encode_residual = FALSE
)
}
\arguments{
\item{nlist}{Number of cells to partition dataset into.}

\item{nprobe}{At query time, the number of cells used for approximate nearest
neighbor search.}

\item{qtype}{Quantizer type. Must be one of {"QT_8bit", "QT_4bit",
"QT_8bit_uniform", "QT_4bit_uniform", "QT_fp16", "QT_8bit_direct",
"QT_6bit"}.}

\item{encode_residual}{Whether to encode residuals.}
}
\value{
An object encapsulating all required parameters of the "ivfsq" KNN
  query algorithm.
}
\description{
Build a specification of the inverted-file-scalar-quantization KNN query
algorithm, with all required parameters specified explicitly.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsne.R
\name{cuda_ml_tsne}
\alias{cuda_ml_tsne}
\title{t-distributed Stochastic Neighbor Embedding.}
\usage{
cuda_ml_tsne(
  x,
  n_components = 2L,
  n_neighbors = ceiling(3 * perplexity),
  method = c("barnes_hut", "fft", "exact"),
  angle = 0.5,
  n_iter = 1000L,
  learning_rate = 200,
  learning_rate_method = c("adaptive", "none"),
  perplexity = 30,
  perplexity_max_iter = 100L,
  perplexity_tol = 1e-05,
  early_exaggeration = 12,
  late_exaggeration = 1,
  exaggeration_iter = 250L,
  min_grad_norm = 1e-07,
  pre_momentum = 0.5,
  post_momentum = 0.8,
  square_distances = TRUE,
  seed = NULL,
  cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace")
)
}
\arguments{
\item{x}{The input matrix or dataframe. Each data point should be a row
and should consist of numeric values only.}

\item{n_components}{Dimension of the embedded space.}

\item{n_neighbors}{The number of datapoints to use in the attractive forces.
Default: ceiling(3 * perplexity).}

\item{method}{T-SNE method, must be one of {"barnes_hut", "fft", "exact"}.
The "exact" method will be more accurate but slower. Both "barnes_hut" and
"fft" methods are fast approximations.}

\item{angle}{Valid values are between 0.0 and 1.0, which trade off speed and
accuracy, respectively. Generally, these values are set between 0.2 and
0.8. (Barnes-Hut only.)}

\item{n_iter}{Maximum number of iterations for the optimization. Should be
at least 250. Default: 1000L.}

\item{learning_rate}{Learning rate of the t-SNE algorithm, usually between
(10, 1000). If the learning rate is too high, then t-SNE result could look
like a cloud / ball of points.}

\item{learning_rate_method}{Must be one of {"adaptive", "none"}. If
"adaptive", then learning rate, early exaggeration, and perplexity are
automatically tuned based on input size. Default: "adaptive".}

\item{perplexity}{The target value of the conditional distribution's
perplexity (see
https://en.wikipedia.org/wiki/T-distributed_stochastic_neighbor_embedding
for details).}

\item{perplexity_max_iter}{The number of epochs the best Gaussian bands are
found for. Default: 100L.}

\item{perplexity_tol}{Stop optimizing the Gaussian bands when the conditional
distribution's perplexity is within this desired tolerance compared to its
taget value. Default: 1e-5.}

\item{early_exaggeration}{Controls the space between clusters. Not critical
to tune this. Default: 12.0.}

\item{late_exaggeration}{Controls the space between clusters. It may be
beneficial to increase this slightly to improve cluster separation. This
will be applied after `exaggeration_iter` iterations (FFT only).}

\item{exaggeration_iter}{Number of exaggeration iterations. Default: 250L.}

\item{min_grad_norm}{If the gradient norm is below this threshold, the
optimization will be stopped. Default: 1e-7.}

\item{pre_momentum}{During the exaggeration iteration, more forcefully apply
gradients. Default: 0.5.}

\item{post_momentum}{During the late phases, less forcefully apply gradients.
Default: 0.8.}

\item{square_distances}{Whether TSNE should square the distance values.}

\item{seed}{Seed to the psuedorandom number generator. Setting this can make
repeated runs look more similar. Note, however, that this highly
parallelized t-SNE implementation is not completely deterministic between
runs, even with the same \code{seed} being used for each run.
Default: NULL.}

\item{cuML_log_level}{Log level within cuML library functions. Must be one of
{"off", "critical", "error", "warn", "info", "debug", "trace"}.
Default: off.}
}
\value{
A matrix containing the embedding of the input data in a low-
  dimensional space, with each row representing an embedded data point.
}
\description{
t-distributed Stochastic Neighbor Embedding (TSNE) for visualizing high-
dimensional data.
}
\examples{
library(cuda.ml)

embedding <- cuda_ml_tsne(iris[1:4], method = "exact")

set.seed(0L)
print(kmeans(embedding, centers = 3))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rand_forest.R
\name{predict.cuda_ml_rand_forest}
\alias{predict.cuda_ml_rand_forest}
\title{Make predictions on new data points.}
\usage{
\method{predict}{cuda_ml_rand_forest}(
  object,
  x,
  output_class_probabilities = NULL,
  cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace"),
  ...
)
}
\arguments{
\item{object}{A trained CuML model.}

\item{x}{A matrix or dataframe containing new data points.}

\item{output_class_probabilities}{Whether to output class probabilities.
NOTE: setting \code{output_class_probabilities} to \code{TRUE} is only
valid when the model being applied is a classification model and supports
class probabilities output. CuML classification models supporting class
probabilities include \code{knn}, \code{fil}, and \code{rand_forest}.
A warning message will be emitted if \code{output_class_probabilities}
is set to \code{TRUE} or \code{FALSE} but the model being applied does
not support class probabilities output.}

\item{cuML_log_level}{Log level within cuML library functions. Must be one of
{"off", "critical", "error", "warn", "info", "debug", "trace"}.
Default: off.}

\item{...}{Additional arguments to \code{predict()}. Currently unused.}
}
\value{
Predictions on new data points.
}
\description{
Make predictions on new data points using a CuML random forest model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logistic_reg.R
\name{cuda_ml_logistic_reg}
\alias{cuda_ml_logistic_reg}
\alias{cuda_ml_logistic_reg.default}
\alias{cuda_ml_logistic_reg.data.frame}
\alias{cuda_ml_logistic_reg.matrix}
\alias{cuda_ml_logistic_reg.formula}
\alias{cuda_ml_logistic_reg.recipe}
\title{Train a logistic regression model.}
\usage{
cuda_ml_logistic_reg(x, ...)

\method{cuda_ml_logistic_reg}{default}(x, ...)

\method{cuda_ml_logistic_reg}{data.frame}(
  x,
  y,
  fit_intercept = TRUE,
  penalty = c("l2", "l1", "elasticnet", "none"),
  tol = 1e-04,
  C = 1,
  class_weight = NULL,
  sample_weight = NULL,
  max_iters = 1000L,
  linesearch_max_iters = 50L,
  l1_ratio = NULL,
  ...
)

\method{cuda_ml_logistic_reg}{matrix}(
  x,
  y,
  fit_intercept = TRUE,
  penalty = c("l2", "l1", "elasticnet", "none"),
  tol = 1e-04,
  C = 1,
  class_weight = NULL,
  sample_weight = NULL,
  max_iters = 1000L,
  linesearch_max_iters = 50L,
  l1_ratio = NULL,
  ...
)

\method{cuda_ml_logistic_reg}{formula}(
  formula,
  data,
  fit_intercept = TRUE,
  penalty = c("l2", "l1", "elasticnet", "none"),
  tol = 1e-04,
  C = 1,
  class_weight = NULL,
  sample_weight = NULL,
  max_iters = 1000L,
  linesearch_max_iters = 50L,
  l1_ratio = NULL,
  ...
)

\method{cuda_ml_logistic_reg}{recipe}(
  x,
  data,
  fit_intercept = TRUE,
  penalty = c("l2", "l1", "elasticnet", "none"),
  tol = 1e-04,
  C = 1,
  class_weight = NULL,
  sample_weight = NULL,
  max_iters = 1000L,
  linesearch_max_iters = 50L,
  l1_ratio = NULL,
  ...
)
}
\arguments{
\item{x}{Depending on the context:

  * A __data frame__ of predictors.
  * A __matrix__ of predictors.
  * A __recipe__ specifying a set of preprocessing steps
  * created from [recipes::recipe()].
  * A __formula__ specifying the predictors and the outcome.}

\item{...}{Optional arguments; currently unused.}

\item{y}{A numeric vector (for regression) or factor (for classification) of
desired responses.}

\item{fit_intercept}{If TRUE, then the model tries to correct for the global
mean of the response variable. If FALSE, then the model expects data to be
centered. Default: TRUE.}

\item{penalty}{The penalty type, must be one of
{"none", "l1", "l2", "elasticnet"}.
If "none" or "l2" is selected, then L-BFGS solver will be used.
If "l1" is selected, solver OWL-QN will be used.
If "elasticnet" is selected, OWL-QN will be used if l1_ratio > 0, otherwise
L-BFGS will be used. Default: "l2".}

\item{tol}{Tolerance for stopping criteria. Default: 1e-4.}

\item{C}{Inverse of regularization strength; must be a positive float.
Default: 1.0.}

\item{class_weight}{If \code{NULL}, then each class has equal weight of
\code{1}.
If \code{class_weight} is set to \code{"balanced"}, then weights will be
inversely proportional to class frequencies in the input data.
If otherwise, then \code{class_weight} must be a named numeric vector of
weight values, with names being class labels.
If \code{class_weight} is not \code{NULL}, then each entry in
\code{sample_weight} will be adjusted by multiplying its original value
with the class weight of the corresponding sample's class.
Default: NULL.}

\item{sample_weight}{Array of weights assigned to individual samples.
If \code{NULL}, then each sample has an equal weight of 1. Default: NULL.}

\item{max_iters}{Maximum number of solver iterations. Default: 1000L.}

\item{linesearch_max_iters}{Max number of linesearch iterations per outer
iteration used in the LBFGS- and OWL- QN solvers. Default: 50L.}

\item{l1_ratio}{The Elastic-Net mixing parameter, must \code{NULL} or be
within the range of [0, 1]. Default: NULL.}

\item{formula}{A formula specifying the outcome terms on the left-hand side,
and the predictor terms on the right-hand side.}

\item{data}{When a __recipe__ or __formula__ is used, \code{data} is
specified as a  __data frame__ containing the predictors and (if
applicable) the outcome.}
}
\description{
Train a logistic regression model using Quasi-Newton (QN) algorithms (i.e.,
Orthant-Wise Limited Memory Quasi-Newton (OWL-QN) if there is L1
regularization, Limited Memory BFGS (L-BFGS) otherwise).
}
\examples{
library(cuda.ml)

X <- scale(as.matrix(iris[names(iris) != "Species"]))
y <- iris$Species

model <- cuda_ml_logistic_reg(X, y, max_iters = 100)
predictions <- predict(model, X)

# NOTE: if we were only performing binary classifications (e.g., by having
# `iris_data <- iris \%>\% mutate(Species = (Species == "setosa"))`), then the
# above would be conceptually equivalent to the following:
#
# iris_data <- iris \%>\% mutate(Species = (Species == "setosa"))
# model <- glm(
#   Species ~ ., data = iris_data, family = binomial(link = "logit"),
#   control = glm.control(epsilon = 1e-8, maxit = 100)
# )
#
# predict(model, iris_data, type = "response")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/knn.R
\name{cuda_ml_knn_algo_ivfflat}
\alias{cuda_ml_knn_algo_ivfflat}
\title{Build a specification for the "ivfflat" KNN query algorithm.}
\usage{
cuda_ml_knn_algo_ivfflat(nlist, nprobe)
}
\arguments{
\item{nlist}{Number of cells to partition dataset into.}

\item{nprobe}{At query time, the number of cells used for approximate nearest
neighbor search.}
}
\value{
An object encapsulating all required parameters of the "ivfflat" KNN
  query algorithm.
}
\description{
Build a specification of the flat-inverted-file KNN query algorithm, with all
required parameters specified explicitly.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model.R
\name{cuda_ml_serialize}
\alias{cuda_ml_serialize}
\alias{cuda_ml_serialise}
\title{Serialize a CuML model}
\usage{
cuda_ml_serialize(model, connection = NULL, ...)

cuda_ml_serialise(model, connection = NULL, ...)
}
\arguments{
\item{model}{The model object.}

\item{connection}{An open connection or \code{NULL}. If \code{NULL}, then the
model state is serialized to a raw vector. Default: NULL.}

\item{...}{Additional arguments to \code{base::serialize()}.}
}
\value{
\code{NULL} unless \code{connection} is \code{NULL}, in which case
  the serialized model state is returned as a raw vector.
}
\description{
Given a CuML model, serialize its state into a connection.
}
\seealso{
\code{\link[base]{serialize}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tsvd.R
\name{cuda_ml_tsvd}
\alias{cuda_ml_tsvd}
\title{Truncated SVD.}
\usage{
cuda_ml_tsvd(
  x,
  n_components = 2L,
  eig_algo = c("dq", "jacobi"),
  tol = 1e-07,
  n_iters = 15L,
  transform_input = TRUE,
  cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace")
)
}
\arguments{
\item{x}{The input matrix or dataframe. Each data point should be a row
and should consist of numeric values only.}

\item{n_components}{Desired dimensionality of output data. Must be strictly
less than \code{ncol(x)} (i.e., the number of features in input data).
Default: 2.}

\item{eig_algo}{Eigen decomposition algorithm to be applied to the covariance
matrix. Valid choices are "dq" (divid-and-conquer method for symmetric
matrices) and "jacobi" (the Jacobi method for symmetric matrices).
Default: "dq".}

\item{tol}{Tolerance for singular values computed by the Jacobi method.
Default: 1e-7.}

\item{n_iters}{Maximum number of iterations for the Jacobi method.
Default: 15.}

\item{transform_input}{If TRUE, then compute an approximate representation
of the input data. Default: TRUE.}

\item{cuML_log_level}{Log level within cuML library functions. Must be one of
{"off", "critical", "error", "warn", "info", "debug", "trace"}.
Default: off.}
}
\value{
A TSVD model object with the following attributes:
  - "components": a matrix of \code{n_components} rows to be used for
     dimensionalitiy reduction on new data points.
  - "explained_variance": (only present if "transform_input" is set to TRUE)
    amount of variance within the input data explained by each component.
  - "explained_variance_ratio": (only present if "transform_input" is set to
    TRUE) fraction of variance within the input data explained by each
    component.
   - "singular_values": The singular values corresponding to each component.
    The singular values are equal to the 2-norms of the \code{n_components}
    variables in the lower-dimensional space.
   - "tsvd_params": opaque pointer to TSVD parameters which will be used for
     performing inverse transforms.
}
\description{
Dimensionality reduction using Truncated Singular Value Decomposition.
}
\examples{
library(cuda.ml)

iris.tsvd <- cuda_ml_tsvd(iris[1:4], n_components = 2)
print(iris.tsvd)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ols.R
\name{cuda_ml_ols}
\alias{cuda_ml_ols}
\alias{cuda_ml_ols.default}
\alias{cuda_ml_ols.data.frame}
\alias{cuda_ml_ols.matrix}
\alias{cuda_ml_ols.formula}
\alias{cuda_ml_ols.recipe}
\title{Train a OLS model.}
\usage{
cuda_ml_ols(x, ...)

\method{cuda_ml_ols}{default}(x, ...)

\method{cuda_ml_ols}{data.frame}(
  x,
  y,
  method = c("svd", "eig", "qr"),
  fit_intercept = TRUE,
  normalize_input = FALSE,
  ...
)

\method{cuda_ml_ols}{matrix}(
  x,
  y,
  method = c("svd", "eig", "qr"),
  fit_intercept = TRUE,
  normalize_input = FALSE,
  ...
)

\method{cuda_ml_ols}{formula}(
  formula,
  data,
  method = c("svd", "eig", "qr"),
  fit_intercept = TRUE,
  normalize_input = FALSE,
  ...
)

\method{cuda_ml_ols}{recipe}(
  x,
  data,
  method = c("svd", "eig", "qr"),
  fit_intercept = TRUE,
  normalize_input = FALSE,
  ...
)
}
\arguments{
\item{x}{Depending on the context:

  * A __data frame__ of predictors.
  * A __matrix__ of predictors.
  * A __recipe__ specifying a set of preprocessing steps
  * created from [recipes::recipe()].
  * A __formula__ specifying the predictors and the outcome.}

\item{...}{Optional arguments; currently unused.}

\item{y}{A numeric vector (for regression) or factor (for classification) of
desired responses.}

\item{method}{Must be one of \{"svd", "eig", "qr"\}.

  - "svd": compute SVD decomposition using Jacobi iterations.
  - "eig": use an eigendecomposition of the covariance matrix.
  - "qr": use the QR decomposition algorithm and solve `Rx = Q^T y`.

  If the number of features is larger than the sample size, then the
  "svd" algorithm will be force-selected because it is the only
   algorithm that can support this type of scenario.

  Default: "svd".}

\item{fit_intercept}{If TRUE, then the model tries to correct for the global
mean of the response variable. If FALSE, then the model expects data to be
centered. Default: TRUE.}

\item{normalize_input}{Ignored when \code{fit_intercept} is FALSE. If TRUE,
then the predictors will be normalized to have a L2 norm of 1.
Default: FALSE.}

\item{formula}{A formula specifying the outcome terms on the left-hand side,
and the predictor terms on the right-hand side.}

\item{data}{When a __recipe__ or __formula__ is used, \code{data} is
specified as a  __data frame__ containing the predictors and (if
applicable) the outcome.}
}
\value{
A OLS regressor that can be used with the 'predict' S3 generic to
  make predictions on new data points.
}
\description{
Train an Ordinary Least Square (OLS) model for regression tasks.
}
\examples{

library(cuda.ml)

model <- cuda_ml_ols(formula = mpg ~ ., data = mtcars, method = "qr")
predictions <- predict(model, mtcars[names(mtcars) != "mpg"])

# predictions will be comparable to those from a `stats::lm` model
lm_model <- stats::lm(formula = mpg ~ ., data = mtcars, method = "qr")
lm_predictions <- predict(lm_model, mtcars[names(mtcars) != "mpg"])

print(
  all.equal(
    as.numeric(lm_predictions),
    predictions$.pred,
    tolerance = 1e-3
  )
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package.R
\docType{package}
\name{cuda.ml}
\alias{cuda.ml}
\title{cuda.ml}
\description{
This package provides a R interface for the RAPIDS cuML library.
}
\author{
Yitao Li <yitao@rstudio.com>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/umap.R
\name{cuda_ml_umap}
\alias{cuda_ml_umap}
\title{Uniform Manifold Approximation and Projection (UMAP) for dimension reduction.}
\usage{
cuda_ml_umap(
  x,
  y = NULL,
  n_components = 2L,
  n_neighbors = 15L,
  n_epochs = 500L,
  learning_rate = 1,
  init = c("spectral", "random"),
  min_dist = 0.1,
  spread = 1,
  set_op_mix_ratio = 1,
  local_connectivity = 1L,
  repulsion_strength = 1,
  negative_sample_rate = 5L,
  transform_queue_size = 4,
  a = NULL,
  b = NULL,
  target_n_neighbors = n_neighbors,
  target_metric = c("categorical", "euclidean"),
  target_weight = 0.5,
  transform_input = TRUE,
  seed = NULL,
  cuML_log_level = c("off", "critical", "error", "warn", "info", "debug", "trace")
)
}
\arguments{
\item{x}{The input matrix or dataframe. Each data point should be a row
and should consist of numeric values only.}

\item{y}{An optional numeric vector of target values for supervised dimension
reduction. Default: NULL.}

\item{n_components}{The dimension of the space to embed into. Default: 2.}

\item{n_neighbors}{The size of local neighborhood (in terms of number of
neighboring sample points) used for manifold approximation. Default: 15.}

\item{n_epochs}{The number of training epochs to be used in optimizing the
low dimensional embedding. Default: 500.}

\item{learning_rate}{The initial learning rate for the embedding
optimization. Default: 1.0.}

\item{init}{Initialization mode of the low dimensional embedding. Must be
one of {"spectral", "random"}. Default: "spectral".}

\item{min_dist}{The effective minimum distance between embedded points.
Default: 0.1.}

\item{spread}{The effective scale of embedded points. In combination with
\code{min_dist} this determines how clustered/clumped the embedded points
are. Default: 1.0.}

\item{set_op_mix_ratio}{Interpolate between (fuzzy) union and intersection as
the set operation used to combine local fuzzy simplicial sets to obtain a
global fuzzy simplicial sets. Both fuzzy set operations use the product
t-norm. The value of this parameter should be between 0.0 and 1.0; a value
of 1.0 will use a pure fuzzy union, while 0.0 will use a pure fuzzy
intersection. Default: 1.0.}

\item{local_connectivity}{The local connectivity required -- i.e. the number
of nearest neighbors that should be assumed to be connected at a local
level. Default: 1.}

\item{repulsion_strength}{Weighting applied to negative samples in low
dimensional embedding optimization. Values higher than one will result in
greater weight being given to negative samples. Default: 1.0.}

\item{negative_sample_rate}{The number of negative samples to select per
positive sample in the optimization process. Default: 5.}

\item{transform_queue_size}{For transform operations (embedding new points
using a trained model this will control how aggressively to search for
nearest neighbors. Default: 4.0.}

\item{a, b}{More specific parameters controlling the embedding. If not set,
then these values are set automatically as determined by \code{min_dist}
and \code{spread}. Default: NULL.}

\item{target_n_neighbors}{The number of nearest neighbors to use to construct
the target simplcial set. Default: n_neighbors.}

\item{target_metric}{The metric for measuring distance between the actual and
and the target values (\code{y}) if using supervised dimension reduction.
Must be one of {"categorical", "euclidean"}. Default: "categorical".}

\item{target_weight}{Weighting factor between data topology and target
topology. A value of 0.0 weights entirely on data, a value of 1.0 weights
entirely on target. The default of 0.5 balances the weighting equally
between data and target.}

\item{transform_input}{If TRUE, then compute an approximate representation
of the input data. Default: TRUE.}

\item{seed}{Optional seed for pseudo random number generator.  Default: NULL.
Setting a PRNG seed will enable consistency of trained embeddings, allowing
for reproducible results to 3 digits of precision, but at the expense of
potentially slower training and increased memory usage.
If the PRNG seed is not set, then the trained embeddings will not be
deterministic.}

\item{cuML_log_level}{Log level within cuML library functions. Must be one of
{"off", "critical", "error", "warn", "info", "debug", "trace"}.
Default: off.}
}
\value{
A UMAP model object that can be used as input to the
  \code{cuda_ml_transform()} function.
  If \code{transform_input} is set to TRUE, then the model object will
  contain a "transformed_data" attribute containing the lower dimensional
  embedding of the input data.
}
\description{
Run the Uniform Manifold Approximation and Projection (UMAP) algorithm to
find a low dimensional embedding of the input data that approximates an
underlying manifold.
}
\examples{
library(cuda.ml)

model <- cuda_ml_umap(
  x = iris[1:4],
  y = iris[[5]],
  n_components = 2,
  n_epochs = 200,
  transform_input = TRUE
)

set.seed(0L)
print(kmeans(model$transformed, iter.max = 100, centers = 3))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/knn.R
\name{cuda_ml_knn_algo_ivfpq}
\alias{cuda_ml_knn_algo_ivfpq}
\title{Build a specification for the "ivfpq" KNN query algorithm.}
\usage{
cuda_ml_knn_algo_ivfpq(
  nlist,
  nprobe,
  m,
  n_bits,
  use_precomputed_tables = FALSE
)
}
\arguments{
\item{nlist}{Number of cells to partition dataset into.}

\item{nprobe}{At query time, the number of cells used for approximate nearest
neighbor search.}

\item{m}{Number of subquantizers.}

\item{n_bits}{Bits allocated per subquantizer.}

\item{use_precomputed_tables}{Whether to use precomputed tables.}
}
\value{
An object encapsulating all required parameters of the "ivfpq" KNN
  query algorithm.
}
\description{
Build a specification of the inverted-file-product-quantization KNN query
algorithm, with all required parameters specified explicitly.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/logistic_reg.R
\name{predict.cuda_ml_logistic_reg}
\alias{predict.cuda_ml_logistic_reg}
\title{Make predictions on new data points.}
\usage{
\method{predict}{cuda_ml_logistic_reg}(object, x, ...)
}
\arguments{
\item{object}{A trained CuML model.}

\item{x}{A matrix or dataframe containing new data points.}

\item{...}{Additional arguments to \code{predict()}. Currently unused.}
}
\value{
Predictions on new data points.
}
\description{
Make predictions on new data points using a CuML logistic regression model.
}
